#include <stdio.h>
#include <stdlib.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360

void printerror( int status);

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,i1,i2,j1,j2,iw,jw,ii,jj,ii1,ii2,jj1,jj2,iiw,jjw,status,nfound,bitpix=-32,k,kk;
  long fpixel,naxis=2,naxes[2]={LINEPIX, LINEPIY};
  static float zi[LINEPIY][LINEPIX],zo[LINEPIY][LINEPIX];
  float zz[5];
  
  if(argc!=5){
    printf("Usage: %s input output xw yw\n",argv[0]);
    exit(1);
  }
  sscanf(argv[3],"%d",&jw);
  sscanf(argv[4],"%d",&iw);
  jw=jw/2;iw=iw/2;

  fprintf(stderr,"Read %s  ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  if((naxes[0]>LINEPIX)||(naxes[1]>LINEPIY)){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  if(fits_read_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  j1--;i1--;
  j2+=j1;i2+=i1;
  fpixel = 1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zo[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  fprintf(stderr,"Effective Area [%d:%d,%d:%d]\n",j1+1,j2,i1+1,i2);
  fprintf(stderr,"Applying 3x3 median smoothing ... ");
  for(i=i1;i<i2;i++){
    for(j=j1;j<j2;j++){
      for(k=0;k<5;k++) zz[k]=-10000.;
      for(ii=i-1;ii<=i+1;ii++){
	if(ii<i1) iiw=i1;
	else if(ii>i2-1) iiw=i2-1;
	else iiw=ii;
	for(jj=j-1;jj<=j+1;jj++){
	  if(jj<j1) jjw=j1;
	  else if(jj>j2-1) jjw=j2-1;
	  else jjw=jj;	  
	  for(k=4;k>=0;k--){
	    if(zz[k]<zo[iiw][jjw]){
	      for(kk=0;kk<k;kk++) zz[kk]=zz[kk+1];
	      zz[k]=zo[iiw][jjw];
	      break;
	    }
	  }
	}
      }
      //fprintf(stderr,"%d %d",j+1,i+1);
      //for(k=0;k<5;k++) fprintf(stderr," %lf",zz[k]);
      //getchar();
      zi[i][j]=zz[0];
    }
  }
  fprintf(stderr,"OK\n");
  fprintf(stderr,"Applying %dx%d boxcar smoothing ",2*jw+1,2*iw+1);
  for(i=0;i<naxes[1];i++){
    if((int)((double)(i+1)/naxes[1]*10)-(int)((double)i/naxes[1]*10)) fprintf(stderr,".");
    for(j=0;j<naxes[0];j++){
      if(i1<=i&&i<i2&&j1<=j&&j<j2){
	if(i-i1<i2-1-i) iiw=i-i1;
	else iiw=i2-1-i;
	if(iiw<=2*iw) iiw=(float)iiw*iiw/iw/4;
	else iiw=iw;
	ii1=i-iiw;
	ii2=i+iiw+1;
	if(j-j1<j2-1-j) jjw=j-j1;
	else jjw=j2-1-j;
	if(jjw<=2*jw) jjw=(float)jjw*jjw/jw/4;
	else jjw=jw;
	jj1=j-jjw;
	jj2=j+jjw+1;
	zo[i][j]=0.;
	for(ii=ii1;ii<ii2;ii++){
	  for(jj=jj1;jj<jj2;jj++){
	    zo[i][j]+=zi[ii][jj];
	  }
	}
	zo[i][j]/=(ii2-ii1)*(jj2-jj1);
      }
      //else zo[i][j]=zi[i][j];
    }
  }
  fprintf(stderr," OK\n");
  remove(argv[2]);
  fprintf(stderr,"\nWrite %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fp, argv[2], &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  j2-=j1;i2-=i1;
  j1++;i1++;
  if(fits_write_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  fpixel=1;
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zo[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
