#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCAL 200
#define NAV 10
#define MAXODR 5
#define WMIN 0.01

void printerror( int status);

double fit(int od1, int od2, float x[], float y[], int n1, int n2, double c[MAXODR], double sig)
{
  double sx[MAXODR*2],xn[MAXODR*2],dy=1.,yf[LINEPIX];
  double p[MAXODR][MAXODR],q[MAXODR][MAXODR],xs[MAXODR],xy[MAXODR],ys[MAXODR],f;
  int    i,j,k,n,ns,lp,rj[LINEPIX];

  od2++;
  for(n=n1;n<n2;n++) yf[n]=y[n];
  if(sig<0.001) lp=0; else lp=10;
  for(;lp>=0;lp--){
    xn[0]=1;
    sx[0]=1;
    ns=0;
    for(n=n1;n<n2;n++) rj[n]=0;
    for(n=n1;n<n2;n++){
      if(fabs(y[n]-yf[n])>sig*(1+lp*0.1)*dy){
	for(i=-1;i<=1;i++){
	  if(0<=n+i && n+i<LINEPIX) rj[n+i]=1;
	}
      }
    }
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]){
	if(lp==0){
	  //fprintf(stderr,"%g %g %g %g\n",x[n],y[n],y[n]-yf[n],sig*dy);
	  //y[n]=0.;
	}
	continue;
      }
      ns++;
    }
    if(ns<od2*2) {
      //fprintf(stderr,"ns(%d)<od2*2(%d)\n",ns,od2*2);
      for(k=0;k<MAXODR;k++) c[k]=0./0.;
      return(0.);
    }
    for(i=1;i<od2*2;i++) sx[i]=0;
    for(i=0;i<od2;i++) xy[i]=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      for(i=1;i<od2*2;i++) {
        xn[i]=xn[i-1]*x[n];
        sx[i]+=xn[i]/ns;}
      for(i=0;i<od2;i++) xy[i]+=xn[i]*y[n]/ns;
    }
    for(j=0;j<od2;j++){
      for(i=0;i<od2;i++){
        p[i][j]=sx[i+j];
        if(i==j) q[i][j]=1;
        else q[i][j]=0;
      }
    }
    for(k=od1;k<od2;k++){
      for(j=k+1;j<od2;j++){
        f=p[k][j]/p[k][k];
        for(i=0;i<od2;i++){
          p[i][j]-=f*p[i][k];
          q[i][j]-=f*q[i][k];
        }
      }
      f=p[k][k];
      for(i=od1;i<od2;i++){
        p[i][k]/=f;
        q[i][k]/=f;
      }
    }
    k=od2-1;
    for(j=od1;j<k+1;j++){
      ys[j]=0;
      for(i=od1;i<k+1;i++) ys[j]+=q[i][j]*xy[i];
    }
    for(j=k;j>=od1;j--){
      xs[j]=0;
      for(i=j+1;i<k+1;i++) xs[j]+=p[i][j]*c[i];
      c[j]=ys[j]-xs[j];
    }
    dy=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      yf[n]=c[0];
      xn[0]=1;
      for(i=1;i<=k;i++){
	xn[i]=xn[i-1]*x[n];
	yf[n]+=c[i]*xn[i];
      }
      dy+=(y[n]-yf[n])*(y[n]-yf[n])/ns;
    }
    dy=sqrt(dy);
  }
  return(dy);
}

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,imin,imax,jmin,jmax,ii,ic,di,dj,status,nfound,bitpix=-32;
  long fpixel,naxis=2,naxes[2]={LINEPIX, LINEPIY};
  static float z[LINEPIY][LINEPIX],zx[LINEPIY][LINEPIX];
  float xc[NCAL][LINEPIY],zc[NCAL][LINEPIY],w[3],wt,x[NCAL],y[LINEPIY],c[NCAL],wc[NCAL];
  int k,kmax,k1,k2,kt,ns,nc,sn,i1,i2,nl[NCAL],m;
  double zn[LINEPIY],xs,zs,z2s,av,sg,znmax,x1s,x2s,dxsg,xcs[NCAL],co[NCAL][MAXODR],ct[MAXODR][MAXODR];

  if(argc!=2){
    printf("Usage: %s input\n",argv[0]);
    exit(0);
  }

  fprintf(stderr,"Read %s ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, z[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");

  for(i=0;i<LINEPIY;i++){
    y[i]=i-LINEPIY/2;
    for(k=0;k<NCAL;k++){
      xc[k][i]=0./0.;
      zc[k][i]=0./0.;
    }
  }

  znmax=0.;
  for(i=0;i<naxes[1];i++){
    zn[i]=0.;
    for(j=naxes[0]/4;j<naxes[0]/4*3;j++) zn[i]+=z[i][j];
    zn[i]/=naxes[0]/2;
    if(znmax<zn[i]) znmax=zn[i];
  }
  imin=-1;imax=-1;
  for(i=0;i<naxes[1];i++){
    if(imin<0 && zn[i]>znmax*0.10) imin=i;
    if(i>naxes[1]/2 && imax<0 && zn[i]<znmax*0.10) imax=i;
    for(j=0;j<naxes[0];j++) z[i][j]/=zn[i];
  }
  //fprintf(stderr,"imin=%d imax=%d ",imin,imax);
  for(j=0;j<naxes[0];j++){
    for(i=imin;i<imax;i++){
      if(i<NAV) di=i;
      else if(i>naxes[1]-1-NAV) di=naxes[1]-1-i;
      else di=NAV;
      zn[i]=0.;
      av=0.;sg=10000.;
      for(k=0;k<3;k++){
	zs=0.;z2s=0.;ns=0;
	for(ii=i-di;ii<=i+di;ii++){
	  if(fabs(z[ii][j]-av)<2*sg+1e-5){
	    zs+=z[ii][j];
	    z2s+=(double)z[ii][j]*z[ii][j];
	    ns++;
	  }
	}
	av=zs/ns;
	sg=sqrt(z2s/ns-av*av);
      }
      zn[i]=av;
    }
    for(i=imin;i<imax;i++) z[i][j]=zn[i];
  }
  i1=imin*3/4+imax/4;
  i2=imin/4+imax*3/4;
  jmin=-1;jmax=-1;
  for(j=40;j<naxes[0]-40;j++){
    zs=0.;
    for(i=i1;i<i2;i++) zs+=z[i][j];
    zs/=i2-i1;
    if(jmin<0 && zs>0.1) jmin=j;
    if(j>naxes[0]/2 && jmax<0 && zs<0.1) jmax=j;
  }
  //fprintf(stderr,"jmin=%d jmax=%d\n",jmin,jmax);

  i=naxes[1]/2-1;k=0;
  for(j=jmin;j<jmax;j++){
    for(dj=-1;dj<2;dj++) w[dj+1]=(z[i][j+dj-1]+z[i][j+dj]+z[i][j+dj+1])/3-(z[i][j+dj-3]+z[i][j+dj-2]+z[i][j+dj+2]+z[i][j+dj+3])/4;
    if(w[1]<0) sn=-1; else sn=1; 
    if(w[0]*sn<w[1]*sn&&w[1]*sn>w[2]*sn && w[1]*sn>WMIN){
      //if(k==48) fprintf(stderr,"%d: %d %d %g %g %g %g %g %g %g %g %g  %g %g %g\n",k,j,dj,z[i][j-4],z[i][j-3],z[i][j-2],z[i][j-1],z[i][j],z[i][j+1],z[i][j+2],z[i][j+3],z[i][j+4],w[0],w[1],w[2]);
      zc[k][i]=w[1];
      w[0]=(z[i][j-3]+z[i][j-2])/2;
      w[1]=(z[i][j-1]+z[i][j]+z[i][j+1])/3;
      w[2]=(z[i][j+2]+z[i][j+3])/2;
      if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.) continue;
      xc[k][i]=j+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5;
      if(z[i][j-1]*sn<z[i][j]*sn && z[i][j]*sn>z[i][j+1]*sn) dj=0;
      else if(z[i][j-1]*sn>z[i][j]*sn && z[i][j]*sn>z[i][j+1]*sn) dj=-1;
      else if(z[i][j-1]*sn<z[i][j]*sn && z[i][j]*sn<z[i][j+1]*sn) dj=1;
      else dj=2;
      if(dj!=2){
	w[0]=z[i][j+dj-1];
	w[1]=z[i][j+dj];
	w[2]=z[i][j+dj+1];
	xc[k][i]=j+dj+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2]);
      }
      //if(k==48) fprintf(stderr,"%d %d: %g  %d %d %g %g %g %g %g\n",k,i,xc[k][i],j,dj,z[i][j-2],z[i][j-1],z[i][j],z[i][j+1],z[i][j+2]);
      for(di=1;di<=20;di++){
	w[0]=(z[i-di][j-3]+z[i-di][j-2])/2;
	w[1]=(z[i-di][j-1]+z[i-di][j]+z[i-di][j+1])/3;
	w[2]=(z[i-di][j+2]+z[i-di][j+3])/2;
	//if(k==48) fprintf(stderr,"%d %d %d: %g %g %g %g %g\n",k,j,i-di,w[0],w[1],w[2],(w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5,w[1]*sn-(w[0]+w[2])/2*sn);
	if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.) break;
	if(w[1]*sn-(w[0]+w[2])/2*sn<WMIN) break;
	w[0]=(z[i+di][j-3]+z[i+di][j-2])/2;
	w[1]=(z[i+di][j-1]+z[i+di][j]+z[i+di][j+1])/3;
	w[2]=(z[i+di][j+2]+z[i+di][j+3])/2;
	//if(k==48) fprintf(stderr,"%d %d %d: %g %g %g %g %g\n",k,j,i+di,w[0],w[1],w[2],(w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5,w[1]*sn-(w[0]+w[2])/2*sn);
	if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.5) break;
	if(w[1]*sn-(w[0]+w[2])/2*sn<WMIN) break;
      }
      if(di!=21) continue;
      k++;
      if(k==NCAL){
	fprintf(stderr,"NCAL is too small.\n");
	exit(0);
      }
    }
  }
  kmax=k;

  dxsg=0.;
  for(k=0;k<kmax;k++){
    nc=0;x1s=0.;x2s=0.;
    for(ic=1;ic<naxes[1];ic++){
      if(ic<naxes[1]/2){
        i=naxes[1]/2-1-ic;
        xs=0.;zs=0.;ns=0;
	for(di=1;di<=10;di++){
          if(isnan(xc[k][i+di])==0){
	    //if(k==48) fprintf(stderr,"%d: %d %d %g %g\n",di,k,i+di,xc[k][i+di],zc[k][i+di]);
            xs+=xc[k][i+di];
            zs+=zc[k][i+di];
            ns++;
          }
        }
        if(ns && i>=imin){xs/=ns;zs/=ns;}
        else{
          ic=naxes[1]/2-1;
          continue;
        } 
      }
      else {
        i=ic;
        xs=0.;zs=0.;ns=0;
        for(di=1;di<=10;di++){
          if(isnan(xc[k][i-di])==0){
	    //if(k==48) fprintf(stderr,"%d: %d %d %g %g\n",di,k,i-di,xc[k][i-di],zc[k][i-di]);
            xs+=xc[k][i-di];
            zs+=zc[k][i-di];
            ns++;
          }
        }
        if(ns && i<imax){xs/=ns;zs/=ns;}
        else break;  
      }
      j=(xs+0.5);
      if(zs<0) sn=-1; else sn=1; 
      for(dj=-1;dj<2;dj++) w[dj+1]=(z[i][j+dj-1]+z[i][j+dj]+z[i][j+dj+1])/3-(z[i][j+dj-3]+z[i][j+dj-2]+z[i][j+dj+2]+z[i][j+dj+3])/4;
      //if(k==48){fprintf(stderr,"%d %d %d %g %g %g %g\n",k,i,j,xs,w[0]*sn,w[1]*sn,w[2]*sn);}
      if(w[0]*sn>w[1]*sn&&w[0]*sn>w[2]*sn){
        if(w[0]*sn>WMIN){
          zc[k][i]=w[0];
          j--;
        }
        else continue;
      }
      else if(w[2]*sn>w[1]*sn&&w[2]*sn>w[0]*sn){
        if(w[2]*sn>WMIN){
          zc[k][i]=w[2];
          j++;
        }
        else continue;
      }
      else{
        if(w[1]*sn>WMIN) zc[k][i]=w[1];
        else continue;
      }
      //if(k==48) fprintf(stderr,"%d: %d %d %g %g %g %g %g %g %g  %g %g %g\n",k,j,dj,z[i][j-3],z[i][j-2],z[i][j-1],z[i][j],z[i][j+1],z[i][j+2],z[i][j+3],w[0],w[1],w[2]);
      w[0]=(z[i][j-3]+z[i][j-2])/2;
      w[1]=(z[i][j-1]+z[i][j]+z[i][j+1])/3;
      w[2]=(z[i][j+2]+z[i][j+3])/2;
      //if(k==48) fprintf(stderr,"%g %g %g %g\n",w[0],w[1],w[2],(w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5);
      if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.5) continue;
      xc[k][i]=j+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5;
      if(z[i][j-1]*sn<z[i][j]*sn && z[i][j]*sn>z[i][j+1]*sn) di=0;
      else if(z[i][j-1]*sn>z[i][j]*sn && z[i][j]*sn>z[i][j+1]*sn) di=-1;
      else if(z[i][j-1]*sn<z[i][j]*sn && z[i][j]*sn<z[i][j+1]*sn) di=1;
      else dj=2;
      if(dj!=2){
        w[0]=z[i][j+dj-1];
        w[1]=z[i][j+dj];
        w[2]=z[i][j+dj+1];
        xc[k][i]=j+dj+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2]);
      }
      //if(k==48){fprintf(stderr,"%d %d %d %g %g",k,i,j,xc[k][i],zc[k][i]);getchar();}
      //if(fabs(xc[k][i]-xs)>1.){xc[k][i]=0./0.;zc[k][i]=0./0.;}
      if(zc[k][i]/zs<0.5){xc[k][i]=0./0.;zc[k][i]=0./0.;}
      if(isnan(xc[k][i])==0){
        x1s+=xc[k][i];
        x2s+=(double)xc[k][i]*xc[k][i];
        nc++;
      }
    }
    if(nc){
      x1s/=nc;x2s/=nc;
      dxsg+=x2s-x1s*x1s;
      //fprintf(stderr,"%2d: %g %g %d %g %g\n",k,xc[k][naxes[1]/2-1],sqrt(x2s-x1s*x1s),nc,x1s,x2s);
    }
  }
  dxsg=sqrt(dxsg/kmax);
  if(isnan(dxsg)) printf("999");
  else printf("%g",dxsg);
  fprintf(stderr,"### dxsg=%g ###\n",dxsg);
  if(isnan(dxsg)) exit(0);
  for(k=0;k<kmax;k++){
    //fprintf(stderr,"%d:",k);
    nc=0;
    ns=0;
    for(i=0;i<naxes[1]/2;i++){
      //if(k==46) fprintf(stderr,"%d %d %g\n",k,i,xc[k][i]);
      if(isnan(xc[k][i])){
        if(ns==0) continue;
        ic=i+1;
        while(isnan(xc[k][ic])) ic++;
        for(di=0;di<ic-i;di++){
          wt=(float)(di+1)/(ic-i+1);
          xc[k][i+di]=xc[k][i-1]*(1-wt)+xc[k][ic]*wt;
	  //if(k==46){fprintf(stderr,"%d %d %d %g %g %g",i-1,i+di,ic,xc[k][i-1],xc[k][i+di],xc[k][ic]);getchar();}
          nc++;
        }
      }
      //if(ns==0) fprintf(stderr," %d,%g",i,xc[k][i]);
      //if(k==46) fprintf(stderr,"%d %d %g\n",k,i,xc[k][i]);
      ns++;
    }
    nl[k]=ns;
    ns=0;
    for(i=naxes[1]-1;i>=naxes[1]/2;i--){
      if(isnan(xc[k][i])){
        if(ns==0) continue;
        ic=i-1;
        while(isnan(xc[k][ic])) ic--;
        for(di=0;di<i-ic;di++){
          wt=(float)(di+1)/(i+1-ic);
          xc[k][i-di]=xc[k][i+1]*(1-wt)+xc[k][ic]*wt;
	  //if(k==46){fprintf(stderr,"%d %d %d %g %g %g",i+1,i-di,ic,xc[k][i+1],xc[k][i-di],xc[k][ic]);getchar();}
          nc++;
        }
      }
      //if(ns==0) fprintf(stderr," - %d,%g",i,xc[k][i]);
      //if(k==46) fprintf(stderr,"%d %d %g\n",k,i,xc[k][i]);
      ns++;
    }
    nl[k]+=ns;
    //fprintf(stderr,"  %d %d\n",nl[k],nc);
  }
  for(k=0;k<kmax;k++) xcs[k]=0.;
  for(ic=0;ic<naxes[1];ic++){
    if(ic<naxes[1]/2){
      i=naxes[1]/2-1-ic;
      di=1;
    }
    else{
      i=ic;
      di=-1;
    }
    for(k1=0;k1<kmax;k1++){
      if(isnan(xc[k1][i])==0) break;
    }
    for(k2=k1+1;k2<kmax;k2++){
      if(isnan(xc[k2][i])==0) break;
    }
    if(k2>=kmax){
      if(ic<naxes[1]/2){
        imin=i+10;
        ic=naxes[1]/2-1;
        for(k=0;k<kmax;k++) xcs[k]=0.;
        continue;
      }
      else{
        imax=i-10;
        break;
      }
    }
    for(k=0;k<kmax;k++){      
      if(k2<k){
	for(kt=k;kt<kmax;kt++){
	  if(isnan(xc[kt][i])==0) break;
	}
	if(kt<kmax){
	  k1=k2;
	  k2=kt;
	}
      }
      if(isnan(xc[k][i])){
        //wt=(xcs[k]-xcs[k1])/(xcs[k2]-xcs[k1]);
        wt=(xc[k][i+di]-xc[k1][i+di])/(xc[k2][i+di]-xc[k1][i+di]);
        xc[k][i]=xc[k1][i]*(1-wt)+xc[k2][i]*wt;
	//if(k==46){fprintf(stderr,"%d: %d %d %d %g %g %g %g %g %g %g %g %g",i,k1,k,k2,xc[k1][i],xc[k][i],xc[k2][i],xcs[k1],xcs[k],xcs[k2],xc[k1][i+di],xc[k][i+di],xc[k2][i+di]);getchar();}
      }
      xcs[k]+=xc[k][i];
    }
  }
  fprintf(stderr,"[%d:%d,%d:%d]\n",jmin+1,jmax+1,imin+1,imax+1);
  //for(i=0;i<naxes[1];i++){
  //printf("%3d:",i);
  //for(k=0;k<kmax;k++) printf(" %5.1lf",xc[k][i]);
  //for(k=48;k<49;k++) printf(" %5.1lf  %5.1lf",xc[k][i],xc[k][i]-xc[49][i]);
  //printf("\n");
  //}
  for(k=0;k<kmax;k++){
    wc[k]=fit(0,MAXODR-1,y,xc[k],0,(int)naxes[1],co[k],0.);
    x[k]=co[k][0];
    //printf("%2d %5.3lf",k,wc[k]);
    //for(m=MAXODR-1;m>=0;m--) printf(" %6.2lf",co[k][m]*pow(500.,(double)m));
    //printf("\n");
    ct[0][k]=0.;
  }
  for(m=MAXODR-1;m>=1;m--){
    for(k=0;k<kmax;k++){
      if(wc[k]<2.) c[k]=co[k][m];
      else c[k]=0./0.;
      //fprintf(stderr,"%d: %g %g\n",k,x[k],c[k]);
    } //getchar();
    zs=fit(0,MAXODR-1,x,c,0,kmax,ct[m],0.);
    //fprintf(stderr,"%d %7.3lf",m,zs*pow(500.,(double)m));
    //for(k=MAXODR-1;k>=0;k--) fprintf(stderr," %7.3lf",ct[m][k]*pow(500.,(double)(m+k)));
    //fprintf(stderr,"\n");
  }
  for(j=jmin;j<=jmax;j++){
    for(m=MAXODR-1;m>=0;m--){
      c[m]=0.;
      for(k=MAXODR-1;k>=0;k--) c[m]=c[m]*j+ct[m][k];
      //if(j==964) fprintf(stderr," %6.2lf",c[m]*pow(500.,(double)m));
    }
    //if(j==964) fprintf(stderr,"\n");
    for(i=imin;i<=imax;i++){
      zx[i][j]=0.;
      for(m=MAXODR-1;m>=0;m--) zx[i][j]+=zx[i][j]*y[i]+c[m];
      zx[i][j]*=-1.;
    }
    for(i=imin-1;i>=0;i--) zx[i][j]=zx[imin][j];
    for(i=imax;i<naxes[1];i++) zx[i][j]=zx[imax][j];
  }
  for(j=jmin-1;j>=0;j--){
    for(i=0;i<naxes[1];i++) zx[i][j]=zx[i][jmin];
  }
  for(j=jmax;j<naxes[0];j++){
    for(i=0;i<naxes[1];i++) zx[i][j]=zx[i][jmax];
  }
  rename("dx.fits","dx.bak");
  fprintf(stderr,"Write dx.fits ... ");
  status=0;
  if(fits_create_file(&fp, "dx.fits", &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  fpixel=1;
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zx[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
