#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCAL 100
#define NAV 10
#define MAXODR 5
#define WMIN 3
#define W_SIG 2

void printerror( int status);

double fit(int od1, int od2, float x[], float y[], int n1, int n2, double c[MAXODR], double sig)
{
  double sx[MAXODR*2],xn[MAXODR*2],dy=1.,yf[LINEPIX];
  double p[MAXODR][MAXODR],q[MAXODR][MAXODR],xs[MAXODR],xy[MAXODR],ys[MAXODR],f;
  int    i,j,k,n,ns,lp,rj[LINEPIX];

  od2++;
  for(n=n1;n<n2;n++) yf[n]=y[n];
  if(sig<0.001) lp=0; else lp=10;
  for(;lp>=0;lp--){
    xn[0]=1;
    sx[0]=1;
    ns=0;
    for(n=n1;n<n2;n++) rj[n]=0;
    for(n=n1;n<n2;n++){
      if(fabs(y[n]-yf[n])>sig*(1+lp*0.1)*dy){
	for(i=-1;i<=1;i++){
	  if(0<=n+i && n+i<LINEPIX) rj[n+i]=1;
	}
      }
    }
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]){
	if(lp==0){
	  //fprintf(stderr,"%g %g %g %g\n",x[n],y[n],y[n]-yf[n],sig*dy);
	  //y[n]=0.;
	}
	continue;
      }
      ns++;
    }
    if(ns<od2*2) {
      //fprintf(stderr,"ns(%d)<od2*2(%d)\n",ns,od2*2);
      for(k=0;k<MAXODR;k++) c[k]=0./0.;
      return(0.);
    }
    for(i=1;i<od2*2;i++) sx[i]=0;
    for(i=0;i<od2;i++) xy[i]=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      for(i=1;i<od2*2;i++) {
        xn[i]=xn[i-1]*x[n];
        sx[i]+=xn[i]/ns;}
      for(i=0;i<od2;i++) xy[i]+=xn[i]*y[n]/ns;
    }
    for(j=0;j<od2;j++){
      for(i=0;i<od2;i++){
        p[i][j]=sx[i+j];
        if(i==j) q[i][j]=1;
        else q[i][j]=0;
      }
    }
    for(k=od1;k<od2;k++){
      for(j=k+1;j<od2;j++){
        f=p[k][j]/p[k][k];
        for(i=0;i<od2;i++){
          p[i][j]-=f*p[i][k];
          q[i][j]-=f*q[i][k];
        }
      }
      f=p[k][k];
      for(i=od1;i<od2;i++){
        p[i][k]/=f;
        q[i][k]/=f;
      }
    }
    k=od2-1;
    for(j=od1;j<k+1;j++){
      ys[j]=0;
      for(i=od1;i<k+1;i++) ys[j]+=q[i][j]*xy[i];
    }
    for(j=k;j>=od1;j--){
      xs[j]=0;
      for(i=j+1;i<k+1;i++) xs[j]+=p[i][j]*c[i];
      c[j]=ys[j]-xs[j];
    }
    dy=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      yf[n]=c[0];
      xn[0]=1;
      for(i=1;i<=k;i++){
	xn[i]=xn[i-1]*x[n];
	yf[n]+=c[i]*xn[i];
        }
      dy+=(y[n]-yf[n])*(y[n]-yf[n])/ns;
    }
    dy=sqrt(dy);
  }
  return(dy);
}

double airindex(double w)
{
  double   u,n3,t=288,ps=1012.5,pw=0,ds,dw;

  u=1/w;
  ds=ps/t*(1+ps*(57.90e-8-9.3250e-4/t+0.25844/t/t));
  dw=pw/t*(1+pw*(1+3.7e-4*pw)*(-2.37321e-3+2.23366/t-710.792/t/t+7.75141e4/t/t/t));
  n3=(2371.34+683939.7/(130-u*u)+4547.3/(38.9-u*u))*ds+(6487.31+58.058*u*u-0.71150*u*u*u*u+0.08851*u*u*u*u*u*u)*dw;
  return (1+n3*1e-8);
}

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,jc,dj,imin,imax,jmin,jmax,di,status,nfound,bitpix=-32;
  long fpixel,naxis=2,naxes[2]={LINEPIX, LINEPIY};
  static float z[LINEPIY][LINEPIX],zy[LINEPIY][LINEPIX];
  float yc[NCAL][LINEPIX],zc[NCAL][LINEPIX],w[3],wt,x[LINEPIX],y[NCAL],c[NCAL],wcen,dw;
  float cy[NCAL],cy0[NCAL],dy,wc[NCAL],tmpval;
  double cw0[NCAL],cw[NCAL],ycs[NCAL];
  int k,kmax,kk,k1,k2,kt,ns,nl[NCAL],nc,m,n,nmax,nk[NCAL],noi,nof,nrj,ycen,detid;
  double ys,zs,zsmax,y1s,y2s,dysg,co[NCAL][MAXODR],ct[MAXODR][MAXODR];
  char buf[1024],lname[NCAL][8],*ptr,fname[32];
  FILE *fpt;

  if(argc!=2 && argc!=3){
    printf("Usage: %s input [path_to_calfile]\n",argv[0]);
    exit(0);
  }

  fprintf(stderr,"Read %s ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  if(fits_read_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, z[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");

  ycen=naxes[1]/2.;
  for(j=0;j<LINEPIX;j++){
    x[j]=j-LINEPIX/2;
    for(k=0;k<NCAL;k++){
      yc[k][j]=0./0.;
      zc[k][j]=0./0.;
    }
  }

  if(argc==3){
    sprintf(fname,"%s",argv[2]);
    if(detid==0) strcat(fname,"cal_g.dat");
    else if(detid==1) strcat(fname,"cal_r.dat");
    else strcat(fname,"cal_i.dat");
  }
  else sprintf(fname,"caldy.dat");
  fprintf(stderr,"Read %s ... ",fname);
  if((fpt = fopen(fname,"r")) == NULL){
    fprintf(stderr,"Cannot open %s\n",fname);
    exit(0);
  }
  n=0;
  imin=naxes[1]-1;imax=0;
  while(!feof(fpt)){
    if(fgets(buf,sizeof(buf),fpt)==NULL) continue;
    if(buf[0]=='/') continue;
    ptr=buf;
    if(buf[0]=='#') ptr++;
    if(n==0){
      if(sscanf(ptr,"%f %f",&wcen,&dw)!=2){
	fprintf(stderr,"The first line of %s must be wcen & dw (%s)\n",fname,buf);
	exit(0);
      }
    }
    else{
      if(sscanf(ptr,"%f %lf %s",&cy0[n],&cw0[n],lname[n])!=3) continue;
      if(strcmp(lname[n],"HgI")!=0 && strcmp(lname[n],"NeI")!=0 && strcmp(lname[n],"XeI")!=0) continue;
      cw[n]=cw0[n]*airindex(cw0[n]);
      //fprintf(stderr,"%d: %g %g %g\n",n,cy0[n],(cw[n]-wcen)/dw+ycen,cw[n]);
      if(argc==3) cy[n]=cy0[n];
      else cy[n]=(cw[n]-wcen)/dw+ycen;
      if(imin>cy[n]) imin=cy[n]-10;
      if(imax<cy[n]) imax=cy[n]+10;
    }
    n++;
    if(n==NCAL){
      fprintf(stderr,"NCAL must be larger than the lines of %s.\n",fname);
      exit(0);
    }
  }
  fclose(fpt);
  fprintf(stderr,"OK\n");
  cy[n]=LINEPIY;
  nmax=n;
  imin-=4;imax+=4;
  if(imin<0) imin=0;
  if(imax>naxes[1]) imax=naxes[1];
  dy=-1.;zsmax=0.;
  while(1){
    zs=0.;
    for(n=1;n<nmax;n++){
      cy[n]+=dy;
      for(i=cy[n]-2;i<=cy[n]+3;i++){
	zs+=z[i][naxes[0]/2-1]*(3-fabs(cy[n]-i));
      }
    }
    if(zsmax>zs){
      if(fabs(dy)<1e-3) break;
      dy*=-0.1;
    }
    zsmax=zs;
    //fprintf(stderr,"%g %g %g",dy,zsmax,cy[0]);getchar();
  }

  j=naxes[0]/2-1;k=0;
  for(i=imin+4;i<imax-4;i++){
    for(di=-1;di<2;di++) w[di+1]=(z[i+di-1][j]+z[i+di][j]+z[i+di+1][j])/3-(z[i+di-3][j]+z[i+di-2][j]+z[i+di+2][j]+z[i+di+3][j])/4;
    if(w[0]<w[1]&&w[1]>w[2] && w[1]>WMIN){
      //fprintf(stderr,"%d: %d %d %g %g %g %g %g %g %g  %g %g %g\n",k,i,di,z[i-3][j],z[i-2][j],z[i-1][j],z[i][j],z[i+1][j],z[i+2][j],z[i+3][j],w[0],w[1],w[2]);
      zc[k][j]=w[1];
      w[0]=(z[i-3][j]+z[i-2][j])/2;
      w[1]=(z[i-1][j]+z[i][j]+z[i+1][j])/3;
      w[2]=(z[i+2][j]+z[i+3][j])/2;
      if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.) continue;
      if(w[0]*1.1>w[1] || w[1]<w[2]*1.1) continue;
      yc[k][j]=i+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5;
      if(z[i-1][j]<z[i][j] && z[i][j]>z[i+1][j]) di=0;
      else if(z[i-1][j]>z[i][j] && z[i][j]>z[i+1][j]) di=-1;
      else if(z[i-1][j]<z[i][j] && z[i][j]<z[i+1][j]) di=1;
      else di=2;
      if(di!=2){
	w[0]=z[i+di-1][j];
	w[1]=z[i+di][j];
	w[2]=z[i+di+1][j];
	yc[k][j]=i+di+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2]);
      }
      //fprintf(stderr,"%d %d: %g  %d %d %g %g %g %g %g\n",k,j,yc[k][j],i,di,z[i-2][j],z[i-1][j],z[i][j],z[i+1][j],z[i+2][j]);getchar();
      for(dj=1;dj<=20;dj++){
	w[0]=(z[i-3][j-dj]+z[i-2][j-dj])/2;
	w[1]=(z[i-1][j-dj]+z[i][j-dj]+z[i+1][j-dj])/3;
	w[2]=(z[i+2][j-dj]+z[i+3][j-dj])/2;
	if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.) continue;
	if(w[1]-(w[0]+w[2])/2<WMIN) break;
	w[0]=(z[i-3][j+dj]+z[i-2][j+dj])/2;
	w[1]=(z[i-1][j+dj]+z[i][j+dj]+z[i+1][j+dj])/3;
	w[2]=(z[i+2][j+dj]+z[i+3][j+dj])/2;
	if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.) continue;
	if(w[1]-(w[0]+w[2])/2<WMIN) break;
      }
      if(dj!=21) continue;
      k++;
    }
  }
  kmax=k;
  //fprintf(stderr,"kmax=%d\n",kmax);
  dysg=0.;
  for(k=0;k<kmax;k++){
    nc=0;y1s=0.;y2s=0.;
    for(jc=1;jc<naxes[0];jc++){
      if(jc<naxes[0]/2){
	j=naxes[0]/2-1-jc;
	ys=0.;zs=0.;ns=0;
	for(dj=1;dj<=10;dj++){
	  if(isnan(yc[k][j+dj])==0){
	    ys+=yc[k][j+dj];
	    zs+=zc[k][j+dj];
	    ns++;
	  }
	}
	if(ns){ys/=ns;zs/=ns;}
	else{
	  jc=naxes[0]/2-1;
	  continue;
	} 
      }
      else {
	j=jc;
	ys=0.;zs=0.;ns=0;
	for(dj=1;dj<=10;dj++){
	  if(isnan(yc[k][j-dj])==0){
	    ys+=yc[k][j-dj];
	    zs+=zc[k][j-dj];
	    ns++;
	  }
	}
	if(ns){ys/=ns;zs/=ns;}
	else break;  
      }
      i=(ys+0.5);
      for(di=-1;di<2;di++) w[di+1]=(z[i+di-1][j]+z[i+di][j]+z[i+di+1][j])/3-(z[i+di-3][j]+z[i+di-2][j]+z[i+di+2][j]+z[i+di+3][j])/4;
      if(w[0]>w[1]&&w[0]>w[2]){
	if(w[0]>WMIN){
	  zc[k][j]=w[0];
	  i--;
	}
	else continue;
      }
      else if(w[2]>w[1]&&w[2]>w[0]){
	if(w[2]>WMIN){
	  zc[k][j]=w[2];
	  i++;
	}
	else continue;
      }
      else{
	if(w[1]>WMIN) zc[k][j]=w[1];
	else continue;
      }
      //fprintf(stderr,"%d: %d %d %g %g %g %g %g %g %g  %g %g %g\n",k,i,di,z[i-3][j],z[i-2][j],z[i-1][j],z[i][j],z[i+1][j],z[i+2][j],z[i+3][j],w[0],w[1],w[2]);
      w[0]=(z[i-3][j]+z[i-2][j])/2;
      w[1]=(z[i-1][j]+z[i][j]+z[i+1][j])/3;
      w[2]=(z[i+2][j]+z[i+3][j])/2;
      if(fabs((w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5)>1.) continue;
      yc[k][j]=i+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2])*2.5;
      if(z[i-1][j]<z[i][j] && z[i][j]>z[i+1][j]) di=0;
      else if(z[i-1][j]>z[i][j] && z[i][j]>z[i+1][j]) di=-1;
      else if(z[i-1][j]<z[i][j] && z[i][j]<z[i+1][j]) di=1;
      else di=2;
      if(di!=2){
	w[0]=z[i+di-1][j];
	w[1]=z[i+di][j];
	w[2]=z[i+di+1][j];
	yc[k][j]=i+di+(w[2]-w[0])/2/(2*w[1]-w[0]-w[2]);
      }
      //if(k==22){fprintf(stderr,"%d %d %d %g",k,j,i,yc[k][j]);getchar();}
      if(fabs(yc[k][j]-ys)>1.){yc[k][j]=0./0.;zc[k][j]=0./0.;}
      if(zc[k][j]/zs<0.5){yc[k][j]=0./0.;zc[k][j]=0./0.;}
      if(isnan(yc[k][j])==0){
	y1s+=yc[k][j];
	y2s+=(double)yc[k][j]*yc[k][j];
	nc++;
      }
    }
    y1s/=nc;y2s/=nc;
    dysg+=y2s-y1s*y1s;
    //fprintf(stderr,"%2d: %g %g %d %g %g\n",k,yc[k][naxes[0]/2-1],sqrt(y2s-y1s*y1s),nc,y1s,y2s);
  }
  dysg=sqrt(dysg/kmax);
  if(isnan(dysg)) printf("999");
  else printf("%g",dysg);
  fprintf(stderr,"### dysg=%g ###\n",dysg);
  if(isnan(dysg)) exit(0);
  for(k=0;k<kmax;k++){
    //fprintf(stderr,"%d:",k);
    nc=0;
    ns=0;
    for(j=0;j<naxes[0]/2;j++){
      if(isnan(yc[k][j])){
	if(ns==0) continue;
	jc=j+1;
	while(isnan(yc[k][jc])) jc++;
	for(dj=0;dj<jc-j;dj++){
	  wt=(float)(dj+1)/(jc-j+1);
	  yc[k][j+dj]=yc[k][j-1]*(1-wt)+yc[k][jc]*wt;
	  //fprintf(stderr,"%d %d %d %g %g %g\n",j-1,j+dj,jc,yc[k][j-1],yc[k][j+dj],yc[k][jc]);
	  nc++;
	}
      }
      //if(ns==0) fprintf(stderr," %d,%g",j,yc[k][j]);
      ns++;
    }
    nl[k]=ns;
    ns=0;
    for(j=naxes[0]-1;j>=naxes[0]/2;j--){
      if(isnan(yc[k][j])){
	if(ns==0) continue;
	jc=j-1;
	while(isnan(yc[k][jc])) jc--;
	for(dj=0;dj<j-jc;dj++){
	  wt=(float)(dj+1)/(j+1-jc);
	  yc[k][j-dj]=yc[k][j+1]*(1-wt)+yc[k][jc]*wt;
	  //fprintf(stderr,"%d %d %d %g %g %g\n",j+1,j-dj,jc,yc[k][j+1],yc[k][j-dj],yc[k][jc]);
	  nc++;
	}
      }
      //if(ns==0) fprintf(stderr," - %d,%g",j,yc[k][j]);
      ns++;
    }
    nl[k]+=ns;
    //fprintf(stderr,"  %d %d\n",nl[k],nc);
  }
  //for(j=0;j<naxes[0];j++){
    //printf("%3d:",j);
    //for(k=0;k<kmax;k++) printf(" %5.1lf",yc[k][j]);
    //for(k=0;k<10;k++) printf(" %5.1lf",yc[k][j]);
    //for(k=0;k<10;k++) printf(" %5.1lf  %5.1lf",yc[k][j],yc[k][j]-yc[30][j]);
    //printf("\n");
  //}
  for(k=0;k<kmax;k++) ycs[k]=0.;
  for(jc=0;jc<naxes[0];jc++){
    if(jc<naxes[0]/2){
      j=naxes[0]/2-1-jc;
      dj=1;
    }
    else{
      j=jc;
      dj=-1;
    }
    ns=0;
    for(k=0;k<kmax;k++){
      if(isnan(yc[k][j])==0) ns++;
    }
    if(ns<kmax/2){
      if(jc<naxes[0]/2){
	jmin=j+1;
	jc=naxes[0]/2-1;
	for(k=0;k<kmax;k++) ycs[k]=0.;
	continue;
      }
      else{
	jmax=j-1;
	break;
      }
    }
    for(k1=0;k1<kmax;k1++){
      if(isnan(yc[k1][j])==0) break;
    }
    for(k2=k1+1;k2<kmax;k2++){
      if(isnan(yc[k2][j])==0) break;
    }
    for(k=0;k<kmax;k++){      
      if(k2<k){
	for(kt=k;kt<kmax;kt++){
	  if(isnan(yc[kt][j])==0) break;
	}
	if(kt<kmax){
	  k1=k2;
	  k2=kt;
	}
      }
      if(isnan(yc[k][j])){
	wt=(ycs[k]-ycs[k1])/(ycs[k2]-ycs[k1]);
	yc[k][j]=yc[k1][j]*(1-wt)+yc[k2][j]*wt;
      }
      ycs[k]+=yc[k][j];
    }
  }
  fprintf(stderr,"[%d:%d,%d:%d]\n",jmin+1,jmax+1,imin+1,imax+1);
  //for(j=0;j<naxes[0];j++){
    //printf("%3d:",j);
    //for(k=0;k<kmax;k++) printf(" %5.1lf",yc[k][j]);
    //for(k=0;k<10;k++) printf(" %5.1lf",yc[k][j]);
    //for(k=0;k<10;k++) printf(" %5.1lf  %5.1lf",yc[k][j],yc[k][j]-yc[30][j]);
    //printf("\n");
  //}
  
  for(k=0;k<kmax;k++){
    wc[k]=fit(0,MAXODR-1,x,yc[k],0,(int)naxes[0],co[k],0.);
    //printf("%2d %5.3lf",k,wc[k]);
    //for(m=MAXODR-1;m>=0;m--) printf(" %6.2lf",co[k][m]*pow(1000.,(double)m));
    //printf("\n");
  }
  n=1;
  for(k=0;k<kmax;k++){
    y[k]=co[k][0];
    //fprintf(stderr,"%d/%d %d: %g %g\n",k,kmax,n,y[k],cy[n]);
    if(fabs(y[k]-cy[n])<1.){
      nk[k]=n;
      c[k]=y[k]-((cw[nk[k]]-wcen)/dw+ycen);
      //fprintf(stderr,"%d: %g %g %g %s\n",k,y[k],c[k],cw0[nk[k]],lname[nk[k]]);
      n++;
    }
    else if(y[k]<cy[n]){
      nk[k]=0;
      c[k]=0./0.;
      //fprintf(stderr,"%d: %g is not identified.\n",k,y[k]);
    }
    else{
      kmax++;
      if(kmax==NCAL){
	fprintf(stderr,"kmax==NCAL error\n");
	exit(1);
      }
      for(kk=kmax-1;kk>k;kk--){
	y[kk]=y[kk-1];
	c[kk]=c[kk-1];
	nk[kk]=nk[kk-1];
	wc[kk]=wc[kk-1];
	for(m=MAXODR-1;m>=0;m--) co[kk][m]=co[kk-1][m];
      }
      nk[k]=-n;
      c[k]=0./0.;
      y[k]=cy[n];
      n++;
    }
  }
  zs=fit(0,MAXODR-2,y,c,0,kmax,ct[0],W_SIG);
  ct[0][MAXODR-1]=0.;
  fprintf(stderr,"Write caldy.dat ... ");
  fpt=fopen("caldy.dat","w");
  fprintf(fpt,"%g %g\n",wcen,dw);
  noi=0;nof=0;nrj=0;
  for(k=0;k<kmax;k++){
    c[0]=0.;
    for(m=MAXODR-1;m>=0;m--) c[0]=c[0]*y[k]+ct[0][m];
    if(nk[k]==0){
      if(argc==3) fprintf(fpt,"#%8.3f %10.5lf is not identified.\n",y[k],((double)c[0]+y[k]-ycen)*dw+wcen);
      noi++;
    }
    else if(nk[k]<0){
      if(argc==3) fprintf(fpt,"#%8.3f",cy[-nk[k]]);
      else fprintf(fpt,"#%8.3f",cy0[-nk[k]]);
      fprintf(fpt," %10.5lf %s is not found.\n",cw0[-nk[k]],lname[-nk[k]]);
      nof++;
    }
    else{
      if(fabs(y[k]-c[0]-((cw[nk[k]]-wcen)/dw+ycen))>W_SIG*zs){fprintf(fpt,"#");nrj++;}
      else fprintf(fpt," ");
      if(argc==3) fprintf(fpt,"%8.3f",y[k]);
      else fprintf(fpt,"%8.3f",cy0[nk[k]]);
      fprintf(fpt," %10.5lf %s %8.3f %6.3f\n",cw0[nk[k]],lname[nk[k]],y[k]-c[0],y[k]-c[0]-((cw[nk[k]]-wcen)/dw+ycen));
    }
  }
  fclose(fpt);
  fprintf(stderr,"OK\n");
  fprintf(stderr,"Number of Lines (good/rejected/not_found/not_identified): %d/%d/%d/%d\n",kmax-noi-nof-nrj,nrj,nof,noi);
  if(kmax-noi-nof-nrj<10) fprintf(stderr,"#### Number of good Lines is too small ... Check caldy.dat ! ####\n");
  for(m=MAXODR-1;m>=1;m--){
    for(k=0;k<kmax;k++){
      if(wc[k]<1.) c[k]=co[k][m];
      else c[k]=0./0.;
      //fprintf(stderr,"%d: %g %g\n",k,y[k],c[k]);
    } //getchar();
    zs=fit(0,MAXODR-1,y,c,0,kmax,ct[m],0.);
    //fprintf(stderr,"%d %7.3lf",m,zs*pow(1000.,(double)m));
    //for(k=MAXODR-1;k>=0;k--) fprintf(stderr," %7.3lf",ct[m][k]*pow(1000.,(double)(m+k)));
    //fprintf(stderr,"\n");
  }
  for(i=imin;i<=imax;i++){
    for(m=MAXODR-1;m>=0;m--){
      c[m]=0.;
      for(k=MAXODR-1;k>=0;k--) c[m]=c[m]*i+ct[m][k];
      //if(i==964) fprintf(stderr," %6.2lf",c[m]*pow(1000.,(double)m));
    }
    //if(i==964) fprintf(stderr,"\n");
    for(j=jmin;j<=jmax;j++){
      zy[i][j]=0.;
      for(m=MAXODR-1;m>=0;m--) zy[i][j]+=zy[i][j]*x[j]+c[m];
      zy[i][j]*=-1.;
    }
    for(j=jmin-1;j>=0;j--) zy[i][j]=zy[i][jmin];
    for(j=jmax;j<naxes[0];j++) zy[i][j]=zy[i][jmax];
  }
  for(i=imin-1;i>=0;i--){
    for(j=0;j<naxes[0];j++) zy[i][j]=zy[imin][j];
  }
  for(i=imax;i<naxes[1];i++){
    for(j=0;j<naxes[0];j++) zy[i][j]=zy[imax][j];
  }
  rename("dy.fits","dy.bak");
  fprintf(stderr,"Write dy.fits ... ");
  status=0;
  if(fits_create_file(&fp, "dy.fits", &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE2", "WAVELENGTH", NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CUNIT2", "nm", NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL2", &wcen, NULL, &status)) printerror(status);
  ycen++;
  if(fits_write_key(fp, TINT, "CRPIX2", &ycen, NULL, &status)) printerror(status);
  tmpval=1.;
  if(fits_write_key(fp, TFLOAT, "CD1_1", &tmpval, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_2", &dw, NULL, &status)) printerror(status);
  fpixel=1;
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zy[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  return 0;
}


void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
