#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360

void printerror( int status);

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int status;
  FILE *fpr,*fpw;
  char buf[1024],dmy[32];
  float x,y,mag,err,crval1,crval2,cd1_1,cd1_2,cd2_1,cd2_2,no,ar,fl,nmag;
  int crpix1,crpix2,md,detid;

  if(argc != 4){
    fprintf(stderr,"Usage : %s fits phot coo\n",argv[0]);
    exit(0);
  }
  fprintf(stderr,"Read %s  ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  fprintf(stderr,"RA=%9.5f Dec=%9.5f\n",crval1,crval2);
  fprintf(stderr,"Read: %s ...\n",argv[2]);
  if((fpr = fopen(argv[2],"r")) == NULL){
    fprintf(stderr,"Cannot open %s\n",argv[2]);
    exit(0);
  }
  fprintf(stderr,"Write: %s ... ",argv[3]);
  fpw = fopen(argv[3],"w");
  fprintf(fpw,"# %9.5f %9.5f %4d %4d %9.2e %9.2e %9.2e %9.2e %d\n",crval1,crval2,crpix1,crpix2,cd1_1,cd1_2,cd2_1,cd2_2,detid);
  md=-1;
  while(!feof(fpr)){
    if(fgets(buf,sizeof(buf),fpr)==NULL) break;
    if(md<0){
      if(buf[0]=='#'&&buf[1]=='K') md=1;
      else md=0;
      //fprintf(stderr,"%d, %d: %s",n,md,buf);
    }
    if(buf[0]=='#') continue;
    if(md){
      if(fgets(buf,sizeof(buf),fpr)==NULL) break;
      sscanf(buf,"%f %f",&x,&y);
      //fprintf(stderr,"%d: %g %g\n",n,x[n],y[n]);
      if(fgets(buf,sizeof(buf),fpr)==NULL) break;
      sscanf(buf,"%s %f",dmy,&no);
      if(fgets(buf,sizeof(buf),fpr)==NULL) break;
      if(fgets(buf,sizeof(buf),fpr)==NULL) break;
      sscanf(buf,"%s %s %f %f %f %f",dmy,dmy,&ar,&fl,&mag,&err);
      nmag=2.5*log10(fl/(no*sqrt(ar)))+mag;
    }
    else sscanf(buf,"%f %f",&x,&y);
    fprintf(fpw,"%8.3f %8.3f",x,y);
    if(md) fprintf(fpw," %6.3f %5.3f %6.3f\n",mag,err,nmag);
    else fprintf(fpw,"\n");
  }
  fclose(fpw);
  fclose(fpr);
  fprintf(stderr,"OK\n");
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
