#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCUBE 50
#define NX 1
#define NY 1

void printerror( int status);

double tr1(double a,double b,double c,double d,double e,double f,int md)
{
  double s;
  if(md==1) s=(a/(a-c)*(d-b)+a/(a-e)*(b-f))*a/2;
  else if(md==2) s=(b/(b-d)*(a-c)+b/(b-f)*(e-a))*b/2;
  return (fabs(s));
}

double tr2(double a,double b,double c,double d)
{
  a=fabs(a);b=fabs(b);c=fabs(c);d=fabs(d);
  return((a*d-b*c)*(a*d-b*c)/(a+c)/(b+d)/2);
}

double rc1(double a,double b,double c,double d,double e,double f)
{
  return(fabs(a*d-b*c)/fabs(a-c)*fabs(a)/2+fabs(b*e-a*f)/fabs(b-f)*fabs(b)/2);
}

double rc2(double a,double b,double c,double d,double e,double f,double g,double h,int md)
{
  double s;
  if(md==1) s=(a+g)*(b-h)/2+a*a*(d-b)/(a-c)/2+g*g*(h-f)/(g-e)/2;
  else if(md==2) s=(b+d)*(a-c)/2+b*b*(g-a)/(b-h)/2+d*d*(c-e)/(d-f)/2;
  else if(md==3) s=(c+e)*(f-d)/2+c*c*(b-d)/(a-c)/2+e*e*(f-h)/(g-e)/2;
  else if(md==4) s=(f+h)*(e-g)/2+h*h*(a-g)/(b-h)/2+f*f*(e-c)/(d-f)/2;
  if(s<0.){fprintf(stderr,"%g %d\n(%g,%g)\n(%g,%g)\n(%g,%g)\n(%g,%g)\n%g %g %g\n",s,md,a,b,c,d,e,f,g,h,(c+e)*(f-d)/2,c*c*(b-d)/(a-c)/2,e*e*(f-h)/(g-e)/2);getchar();}
  return (s);
}

int main(int argc,char *argv[])
{
  fitsfile *fpr,*fpw;
  int status,nfound,bitpix=-32;
  long fpixelr,fpixelw,naxis=3,naxes[3]={LINEPIX, LINEPIY, 1};
  int i,j,k,ni[3],nj[3],i1,i2,i3,i4,j1,j2,j3,j4,iq,jq,is,js,ic,jc,ii1,ii2,ii3,ii4,jj1,jj2,jj3,jj4,nzi,nzo,nan,nrt;
  float fx1,fx2,fy1,fy2,cx,cy,vx=0.,vy=0.,r,x,y;
  float dis3[3]={4.407,3.592,3.515},dis5[3]={3.651,3.659,3.424};
  static float zx[LINEPIY][LINEPIX],zy[LINEPIY][LINEPIX],zi[LINEPIY][LINEPIX];
  static double zo[2][LINEPIY][LINEPIX];
  double gx[3][3],gy[3][3],x1,x2,x3,x4,y1,y2,y3,y4,ff1,ff2,ff3,ff4,ds,szi,szo,sn,cs,en,rt,mt=1.,zs,z2s,av,sg;
  float crval1,crval2,cd1_1,cd1_2,cd2_1,cd2_2,wcen=-1.,dw,tmpvalf,gain;
  int detid,crpix1,crpix2,px,py,n,ns,tmpvali;
  FILE *fp0;

  if(argc!=5 && argc!=7 && argc!=9 && argc!=10 && argc!=11){
    fprintf(stderr,"Usage: %s input output warp_x warp_y\n",argv[0]);
    fprintf(stderr,"     or\n");
    fprintf(stderr,"Usage: %s input output rotate enlarge center_x center_y [shift_x shift_y [multipl [adj]]]\n",argv[0]);
    exit(1);
  }
  if(strcmp(argv[1],argv[2])==0){
    fprintf(stderr,"input and output must be different.\n");
    exit(0);
  }
  for(i=0;i<LINEPIY;i++){
    for(j=0;j<LINEPIX;j++){
      zi[i][j]=0./0.;
      zx[i][j]=0./0.;
      zy[i][j]=0./0.;
    }
  }
  fprintf(stderr,"Read %s  ... ",argv[1]);
  status=0;
  if(fits_open_file(&fpr, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_key(fpr, TINT, "NAXIS", &naxis, NULL, &status)) printerror(status);
  if(fits_read_keys_lng(fpr, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
  if(naxis==3) fprintf(stderr,"%ld x %ld x %ld format \n",naxes[0],naxes[1],naxes[2]);
  else fprintf(stderr,"%ld x %ld format \n",naxes[0],naxes[1]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY||naxes[2]>NCUBE){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  if(fits_read_key(fpr, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(argc!=5){
    if(fits_read_key(fpr, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
    if(fits_read_key(fpr, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
    if(fits_read_key(fpr, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
    if(fits_read_key(fpr, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
    if(fits_read_key(fpr, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
    if(fits_read_key(fpr, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
    if(fits_read_key(fpr, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
    if(fits_read_key(fpr, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
    naxes[2]=1;
    naxis=2;
  }
  else {
    if(fits_read_key(fpr, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  }
  if(argc!=5){
    sscanf(argv[3],"%lf",&rt);
    fprintf(stderr,"rotate=%g\n",rt);
    sscanf(argv[4],"%lf",&en);
    fprintf(stderr,"enlarge=%g\n",en);
    while(rt>180.) rt-=360.;
    while(rt<-180.) rt+=360.;
    if(rt<-135.){rt+=180.;nrt=2;}
    else if(rt<-45.){rt+=90.;nrt=3;}
    else if(rt<45.){nrt=0;}
    else if(rt<135.){rt-=90.;nrt=1;}
    else {rt-=180.;nrt=2;}
    //fprintf(stderr,"rt=%g nrt=%d\n",rt,nrt);
    sn=sin(rt/180.*M_PI);
    cs=cos(rt/180.*M_PI);
    sscanf(argv[5],"%f",&cx);
    sscanf(argv[6],"%f",&cy);
    cx-=1.;cy-=1.;
    if(argc>=9){
      sscanf(argv[7],"%f",&vx);
      sscanf(argv[8],"%f",&vy);
      if(argc>=10){
	sscanf(argv[9],"%lf",&mt);
	fprintf(stderr,"multiplied by %g\n",mt);
	if(argc==11){
	  fprintf(stderr,"Read: crpix.dat ... ");
	  if((fp0 = fopen("crpix.dat","r")) == NULL){
	    fprintf(stderr,"Cannot open crpix.dat\n");
	    exit(0);
	  }
	  fscanf(fp0,"%d %d",&px,&py);
	  fclose(fp0);
	  fprintf(stderr," (CRPIX1,CRPIX2)=(%d,%d) OK\n",px,py);
	  vx+=px-crpix1;crpix1=px;
	  vy+=py-crpix2;crpix2=py;
	}
      }
    }
    fprintf(stderr,"center=(%g,%g)\n",cx,cy);
    fprintf(stderr,"offset=(%g,%g)\n",vx,vy);
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){
	x=j;y=i;
	if(argc==11){
	  r=sqrt((x-naxes[0]/2)*(x-naxes[0]/2)+(y-naxes[1]/2)*(y-naxes[1]/2))/1000;
	  x-=r*r*(dis3[detid]+r*r*dis5[detid])*(x-naxes[0]/2)/1000;
	  y-=r*r*(dis3[detid]+r*r*dis5[detid])*(y-naxes[1]/2)/1000;
	}
	zx[i][j]=((x-cx)*cs-(y-cy)*sn)*en-(j-cx)+vx;
	zy[i][j]=((x-cx)*sn+(y-cy)*cs)*en-(i-cy)+vy;
	//fprintf(stderr,"%d %d %f %f",j,i,zx[i][j],zy[i][j]);getchar();
      }
    }
  }
  else{
    nrt=0;
    if(strcmp(argv[3],"0")==0){
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++) zx[i][j]=0.;
      }
    }
    else {
      fprintf(stderr,"Read %s  ... ",argv[3]);
      status=0;
      if(fits_open_file(&fpw, argv[3], READONLY, &status)) printerror(status);
      if(fits_read_keys_lng(fpw, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
      fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
      if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
	fprintf(stderr,"\nIllegal image format\n");
	exit(0);
      }
      fpixelr=1;
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fpw, TFLOAT, fpixelr, naxes[0], NULL, zx[i], NULL, &status)) printerror(status);
	fpixelr+=naxes[0];
      }
      if(fits_close_file(fpw, &status)) printerror(status);
      fprintf(stderr,"OK\n");
    }
    if(strcmp(argv[4],"0")==0){
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++) zy[i][j]=0.;
      }
    }
    else {
      fprintf(stderr,"Read %s  ... ",argv[4]);
      status=0;
      if(fits_open_file(&fpw, argv[4], READONLY, &status)) printerror(status);
      if(fits_read_keys_lng(fpw, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
      fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
      if(naxes[0]!=LINEPIX||naxes[1]!=LINEPIY){
	fprintf(stderr,"\nIllegal image format\n");
	exit(0);
      }
      if(fits_read_key(fpw, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
      if(fits_read_key(fpw, TFLOAT, "CRVAL2", &wcen, NULL, &status)) printerror(status);
      if(fits_read_key(fpw, TFLOAT, "CD2_2", &dw, NULL, &status)) printerror(status);
      fpixelr=1;
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fpw, TFLOAT, fpixelr, naxes[0], NULL, zy[i], NULL, &status)) printerror(status);
	fpixelr+=naxes[0];
      }
      if(fits_close_file(fpw, &status)) printerror(status);
      fprintf(stderr,"OK\n");
    }
  }
  remove(argv[2]);
  fprintf(stderr,"Write %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fpw, argv[2], &status)) printerror(status);
  if(fits_create_img(fpw, bitpix, naxis, naxes, &status)) printerror(status);
  if(naxis==3) fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
  else fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  if(fits_write_key(fpw, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(argc!=5){
    if(fits_write_key(fpw, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TSTRING, "CTYPE1", "RA---TAN", NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TSTRING, "CTYPE2", "DEC--TAN", NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
    tmpvalf=1.;
    if(fits_write_key(fpw, TFLOAT, "CD3_3", &tmpvalf, "", &status)) printerror(status);
  }else{
    if(fits_write_key(fpw, TSTRING, "CTYPE1", "pix", "", &status)) printerror(status);
    if(fits_write_key(fpw, TSTRING, "CTYPE2", "WAVELENGTH", "", &status)) printerror(status);
    if(fits_write_key(fpw, TSTRING, "CUNIT2", "nm", "", &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
    tmpvalf=naxes[0]/2+1;
    if(fits_write_key(fpw, TFLOAT, "CRVAL1", &tmpvalf, "", &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CRVAL2", &wcen, "", &status)) printerror(status);
    tmpvali=naxes[0]/2+1;
    if(fits_write_key(fpw, TINT, "CRPIX1", &tmpvali, "", &status)) printerror(status);
    if(fits_write_key(fpw, TINT, "CRPIX2", &crpix2, "", &status)) printerror(status);
    tmpvalf=1.;
    if(fits_write_key(fpw, TFLOAT, "CD1_1", &tmpvalf, "", &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CD2_2", &dw, "", &status)) printerror(status);
    if(fits_write_key(fpw, TFLOAT, "CD3_3", &tmpvalf, "", &status)) printerror(status);
  }
  fpixelr=1; fpixelw=1;
  for(k=0;k<naxes[2];k++){
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fpr, TFLOAT, fpixelr, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
      for(j=0;j<naxes[0];j++) zi[i][j]*=mt;
      fpixelr+=naxes[0];
    }
    if(nrt){
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++){
	  if(nrt==1){ic=cy-(j-cx);jc=cx+(i-cy);}
	  else if(nrt==2){ic=cy-(i-cy);jc=cx-(j-cx);}
	  else{ic=cy+(j-cx);jc=cx-(i-cy);}
	  if(ic<0 || ic>=naxes[1] || jc<0 || jc>=naxes[0]) zo[0][i][j]=0./0.;
	  else zo[0][i][j]=zi[ic][jc];
	}
      }
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++){zi[i][j]=zo[0][i][j];}
      }
    }
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){zo[0][i][j]=0./0.;zo[1][i][j]=0.;}
    }
    for(i=0;i<naxes[1];i++){
      ni[1]=i;
      if(i==0) ni[0]=i; else ni[0]=i-1;
      if(i==naxes[1]-1) ni[2]=i; else ni[2]=i+1;
      for(j=0;j<naxes[0];j++){
	if(isnan(zi[i][j])||isnan(zx[i][j])||isnan(zy[i][j])) continue;
	nj[1]=j;
	if(j==0) nj[0]=j; else nj[0]=j-1;
	if(j==naxes[0]-1) nj[2]=j; else nj[2]=j+1;
	nan=0;
	for(iq=0;iq<3;iq++){
	  for(jq=0;jq<3;jq++){
	    if(isnan(zy[ni[iq]][nj[jq]])) {gy[iq][jq]=zy[i][j]+ni[iq];nan++;} else gy[iq][jq]=zy[ni[iq]][nj[jq]]+ni[iq];
	    if(isnan(zx[ni[iq]][nj[jq]])) {gx[iq][jq]=zx[i][j]+nj[jq];nan++;} else gx[iq][jq]=zx[ni[iq]][nj[jq]]+nj[jq];
	    if(i==0&&iq==0)  gy[iq][jq]-=1.;
	    else if(i==naxes[1]-1&&iq==2)  gy[iq][jq]+=1.;
	    if(j==0&&jq==0)  gx[iq][jq]-=1.;
	    else if(j==naxes[0]-1&&jq==2)  gx[iq][jq]+=1.;
	  }
	}
	if(nan) continue;
	/*
	  if(j==PX&&i==PY){
	  fprintf(stderr,"(%g,%g)(%g,%g)(%g,%g)\n",zx[i+1][j-1],zy[i+1][j-1],zx[i+1][j],zy[i+1][j],zx[i+1][j+1],zy[i+1][j+1]);
	  fprintf(stderr,"(%g,%g)(%g,%g)(%g,%g)\n",zx[i][j-1],zy[i][j-1],zx[i][j],zy[i][j],zx[i][j+1],zy[i][j+1]);
	  fprintf(stderr,"(%g,%g)(%g,%g)(%g,%g)\n",zx[i-1][j-1],zy[i-1][j-1],zx[i-1][j],zy[i-1][j],zx[i-1][j+1],zy[i-1][j+1]);
	  fprintf(stderr,"(%g,%g)(%g,%g)(%g,%g)\n",gx[2][0],gy[2][0],gx[2][1],gy[2][1],gx[2][2],gy[2][2]);
	  fprintf(stderr,"(%g,%g)(%g,%g)(%g,%g)\n",gx[1][0],gy[1][0],gx[1][1],gy[1][1],gx[1][2],gy[1][2]);
	  fprintf(stderr,"(%g,%g)(%g,%g)(%g,%g)\n",gx[0][0],gy[0][0],gx[0][1],gy[0][1],gx[0][2],gy[0][2]);
	  }
	*/
	for(iq=0;iq<3;iq++){
	  for(jq=0;jq<3;jq++){
	    gx[iq][jq]=(gx[iq][jq]+gx[1][1])/2;
	    gy[iq][jq]=(gy[iq][jq]+gy[1][1])/2;
	  }
	}
	for(iq=0;iq<2;iq++){
	  for(jq=0;jq<2;jq++){
	    if((gx[iq][jq]>gx[iq+1][jq+1]&&gy[iq][jq]>gy[iq+1][jq+1])||(gx[iq+1][jq]>gx[iq][jq+1]&&gy[iq][jq+1]>gy[iq+1][jq])){
	      fprintf(stderr,"Crushed grid: %d %d %d %d %d %d\n",nj[0],nj[1],nj[2],ni[0],ni[1],ni[2]);
	      fprintf(stderr,"(%g,%g)(%g,%g)(%g,%g)(%g,%g)\n",gx[iq][jq],gy[iq][jq],gx[iq][jq+1],gy[iq][jq+1],gx[iq+1][jq],gy[iq+1][jq],gx[iq+1][jq+1],gy[iq+1][jq+1]);getchar();
	    }
	    for(is=0;is<NY;is++){
	      fy1=(float)is/NY;
	      fy2=(float)(is+1)/NY;
	      for(js=0;js<NX;js++){
		fx1=(float)js/NX;
		fx2=(float)(js+1)/NX;
		y1=gy[iq][jq]*(1-fy2)*(1-fx2)+gy[iq][jq+1]*(1-fy2)*fx2+gy[iq+1][jq]*fy2*(1-fx2)+gy[iq+1][jq+1]*fy2*fx2+0.5;
		x1=gx[iq][jq]*(1-fy2)*(1-fx2)+gx[iq][jq+1]*(1-fy2)*fx2+gx[iq+1][jq]*fy2*(1-fx2)+gx[iq+1][jq+1]*fy2*fx2+0.5;
		y2=gy[iq][jq]*(1-fy2)*(1-fx1)+gy[iq][jq+1]*(1-fy2)*fx1+gy[iq+1][jq]*fy2*(1-fx1)+gy[iq+1][jq+1]*fy2*fx1+0.5;
		x2=gx[iq][jq]*(1-fy2)*(1-fx1)+gx[iq][jq+1]*(1-fy2)*fx1+gx[iq+1][jq]*fy2*(1-fx1)+gx[iq+1][jq+1]*fy2*fx1+0.5;
		y3=gy[iq][jq]*(1-fy1)*(1-fx1)+gy[iq][jq+1]*(1-fy1)*fx1+gy[iq+1][jq]*fy1*(1-fx1)+gy[iq+1][jq+1]*fy1*fx1+0.5;
		x3=gx[iq][jq]*(1-fy1)*(1-fx1)+gx[iq][jq+1]*(1-fy1)*fx1+gx[iq+1][jq]*fy1*(1-fx1)+gx[iq+1][jq+1]*fy1*fx1+0.5;
		y4=gy[iq][jq]*(1-fy1)*(1-fx2)+gy[iq][jq+1]*(1-fy1)*fx2+gy[iq+1][jq]*fy1*(1-fx2)+gy[iq+1][jq+1]*fy1*fx2+0.5;
		x4=gx[iq][jq]*(1-fy1)*(1-fx2)+gx[iq][jq+1]*(1-fy1)*fx2+gx[iq+1][jq]*fy1*(1-fx2)+gx[iq+1][jq+1]*fy1*fx2+0.5;
		i1=y1;i2=y2;i3=y3;i4=y4;
		if(y1<0) i1--;
		if(y2<0) i2--;
		if(y3<0) i3--;
		if(y4<0) i4--;
		j1=x1;j2=x2;j3=x3;j4=x4;
		if(x1<0) j1--;
		if(x2<0) j2--;
		if(x3<0) j3--;
		if(x4<0) j4--;
		if(i1-i3>1||i2-i4>1){
		  fprintf(stderr,"NY is too small: (%d,%d)(%d,%d)(%d,%d)(%d,%d)\n",j1,i1,j2,i2,j3,i3,j4,i4);
		  exit(0);
		}
		if(j1-j3>1||j4-j2>1){
		  fprintf(stderr,"NX is too small: (%d,%d)(%d,%d)(%d,%d)(%d,%d)\n",j1,i1,j2,i2,j3,i3,j4,i4);
		  exit(0);
		}
		if(x1+x2+x3+x4<-2.) jc=(x1+x2+x3+x4)/4-0.5;else jc=(x1+x2+x3+x4)/4+0.5;
		if(y1+y2+y3+y4<-2.) ic=(y1+y2+y3+y4)/4-0.5;else ic=(y1+y2+y3+y4)/4+0.5;
		x1-=jc;x2-=jc;x3-=jc;x4-=jc;y1-=ic;y2-=ic;y3-=ic;y4-=ic;
		ds=(x1*y2-x2*y1+x2*y3-x3*y2+x3*y4-x4*y3+x4*y1-x1*y4)/2;
		ii1=i1;ii2=i2;ii3=i3;ii4=i4;jj1=j1;jj2=j2;jj3=j3;jj4=j4;ff1=0.;ff2=0.;ff3=0.;ff4=0.;
		//1234
		if(i1==i2&&i2==i3&&i3==i4&&j1==j2&&j2==j3&&j3==j4){ff1=ds;}
		//123-4
		else if(i1==i2&&i2==i3&&i3==i4&&j1==j2&&j2==j3){ff4=tr1(x4,y4,x1,y1,x3,y3,1);ff1=ds-ff4;}
		else if(i1==i2&&i2==i3&&j1==j2&&j2==j3&&j3==j4){ff4=tr1(x4,y4,x3,y3,x1,y1,2);ff1=ds-ff4;}
		else if(i1==i2&&i2==i3&&j1==j2&&j2==j3){
		  if(x3*y4-x4*y3<0.){jj1++;ff1=tr2(x4,y4,x1,y1)-tr2(x3,y3,x4,y4);ff4=tr1(x4,y4,x3,y3,x1,y1,2);ff2=ds-ff1-ff4;}
		  else if(x4*y1-x1*y4<0.){ii3--;ff3=tr2(x3,y3,x4,y4)-tr2(x4,y4,x1,y1);ff4=tr1(x4,y4,x1,y1,x3,y3,1);ff2=ds-ff3-ff4;}
		  else {jj1++;ii3--;ff3=tr2(x3,y3,x4,y4);ff1=tr2(x4,y4,x1,y1);ff4=rc1(x4,y4,x3,y3,x1,y1);ff2=ds-ff1-ff3-ff4;}
		}
		//124-3
		else if(i1==i2&&i2==i3&&i3==i4&&j1==j2&&j2==j4){ff3=tr1(x3,y3,x2,y2,x4,y4,1);ff1=ds-ff3;}
		else if(i1==i2&&i2==i4&&j1==j2&&j2==j3&&j3==j4){ff3=tr1(x3,y3,x4,y4,x2,y2,2);ff1=ds-ff3;}
		else if(i1==i2&&i2==i4&&j1==j2&&j2==j4){
		  if(x2*y3-x3*y2<0.){ii4--;ff4=tr2(x3,y3,x4,y4)-tr2(x2,y2,x3,y3);ff3=tr1(x3,y3,x2,y2,x4,y4,1);ff1=ds-ff3-ff4;}
		  else if(x3*y4-x4*y3<0.){jj2--;ff2=tr2(x2,y2,x3,y3)-tr2(x3,y3,x4,y4);ff3=tr1(x3,y3,x4,y4,x2,y2,2);ff1=ds-ff2-ff3;}
		  else {jj2--;ii4--;ff2=tr2(x2,y2,x3,y3);ff4=tr2(x3,y3,x4,y4);ff3=rc1(x3,y3,x4,y4,x2,y2);ff1=ds-ff2-ff3-ff4;}
		}
		//134-2
		else if(i1==i2&&i2==i3&&i3==i4&&j1==j3&&j3==j4){ff2=tr1(x2,y2,x3,y3,x1,y1,1);ff1=ds-ff2;}
		else if(i1==i3&&i3==i4&&j1==j2&&j2==j3&&j3==j4){ff2=tr1(x2,y2,x1,y1,x3,y3,2);ff1=ds-ff2;}
		else if(i1==i3&&i3==i4&&j1==j3&&j3==j4){
		  if(x1*y2-x2*y1<0.){jj3--;ff3=tr2(x2,y2,x3,y3)-tr2(x1,y1,x2,y2);ff2=tr1(x2,y2,x1,y1,x3,y3,2);ff4=ds-ff2-ff3;}
		  else if(x2*y3-x3*y2<0.){ii1++;ff1=tr2(x1,y1,x2,y2)-tr2(x2,y2,x3,y3);ff2=tr1(x2,y2,x3,y3,x1,y1,1);ff4=ds-ff1-ff2;}
		  else {ii1++;jj3--;ff1=tr2(x1,y1,x2,y2);ff3=tr2(x2,y2,x3,y3);ff2=rc1(x2,y2,x1,y1,x3,y3);ff4=ds-ff1-ff2-ff3;}
		}
		//234-1
		else if(i1==i2&&i2==i3&&i3==i4&&j2==j3&&j3==j4){ff1=tr1(x1,y1,x4,y4,x2,y2,1);ff2=ds-ff1;}
		else if(i2==i3&&i3==i4&&j1==j2&&j2==j3&&j3==j4){ff1=tr1(x1,y1,x2,y2,x4,y4,2);ff2=ds-ff1;}
		else if(i2==i3&&i3==i4&&j2==j3&&j3==j4){
		  if(x4*y1-x1*y4<0.){ii2++;ff2=tr2(x1,y1,x2,y2)-tr2(x4,y4,x1,y1);ff1=tr1(x1,y1,x4,y4,x2,y2,1);ff3=ds-ff1-ff2;}
		  else if(x1*y2-x2*y1<0.){jj4++;ff4=tr2(x4,y4,x1,y1)-tr2(x1,y1,x2,y2);ff1=tr1(x1,y1,x2,y2,x4,y4,2);ff3=ds-ff1-ff4;}
		  else {ii2++;jj4++;ff4=tr2(x4,y4,x1,y1);ff2=tr2(x1,y1,x2,y2);ff1=rc1(x1,y1,x2,y2,x4,y4);ff3=ds-ff1-ff2-ff4;}
		}
		//14-23
		else if(i1==i2&&i2==i3&&i3==i4&&j1==j4&&j2==j3){ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1);ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3);}
		else if(i1==i4&&i2==i3&&i2<i4&&j1==j2&&j2==j3&&j3==j4){ff1=rc2(x4,y4,x1,y1,x2,y2,x3,y3,2);ff3=rc2(x4,y4,x1,y1,x2,y2,x3,y3,4);}
		else if(i1==i4&&i2==i3&&i2<i4&&j1==j4&&j2==j3){
		  if(x1*y2-x2*y1<0.){ii4--;ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3);ff1=rc2(x4,y4,x1,y1,x2,y2,x3,y3,2);ff4=tr2(x3,y3,x4,y4)-tr2(x1,y1,x2,y2);}
		  else if(x3*y4-x4*y3<0.){ii2++;ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1);ff3=rc2(x4,y4,x1,y1,x2,y2,x3,y3,4);ff2=tr2(x1,y1,x2,y2)-tr2(x3,y3,x4,y4);}
		  else {ii2++;ii4--;ff2=tr2(x1,y1,x2,y2);ff4=tr2(x3,y3,x4,y4);ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1)-ff4;ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3)-ff2;}
		}
		else if(i1==i4&&i2==i3&&i1<i3&&j1==j2&&j2==j3&&j3==j4){ff2=rc2(x2,y2,x3,y3,x4,y4,x1,y1,2);ff4=rc2(x2,y2,x3,y3,x4,y4,x1,y1,4);}
		else if(i1==i4&&i2==i3&&i1<i3&&j1==j4&&j2==j3){
		  if(x1*y2-x2*y1<0.){ii3--;ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1);ff2=rc2(x2,y2,x3,y3,x4,y4,x1,y1,2);ff3=tr2(x3,y3,x4,y4)-tr2(x1,y1,x2,y2);}
		  else if(x3*y4-x4*y3<0.){ii1++;ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3);ff4=rc2(x2,y2,x3,y3,x4,y4,x1,y1,4);ff1=tr2(x1,y1,x2,y2)-tr2(x3,y3,x4,y4);}
		  else {ii1++;ii3--;ff1=tr2(x1,y1,x2,y2);ff3=tr2(x3,y3,x4,y4);ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1)-ff1;ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3)-ff3;}
		}
		//12-34
		else if(i1==i2&&i3==i4&&j1==j2&&j2==j3&&j3==j4){ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2);ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4);}
		else if(i1==i2&&i2==i3&&i3==i4&&j1==j2&&j3==j4&&j4<j2){ff1=rc2(x2,y2,x3,y3,x4,y4,x1,y1,1);ff3=rc2(x2,y2,x3,y3,x4,y4,x1,y1,3);}
		else if(i1==i2&&i3==i4&&j1==j2&&j3==j4&&j4<j2){
		  if(x2*y3-x3*y2<0.){jj4++;ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2);ff3=rc2(x2,y2,x3,y3,x4,y4,x1,y1,3);ff4=tr2(x4,y4,x1,y1)-tr2(x2,y2,x3,y3);}
		  else if(x4*y1-x1*y4<0.){jj2--;ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4);ff1=rc2(x2,y2,x3,y3,x4,y4,x1,y1,1);ff2=tr2(x2,y2,x3,y3)-tr2(x4,y4,x1,y1);}
		  else {jj2--;jj4++;ff2=tr2(x2,y2,x3,y3);ff4=tr2(x4,y4,x1,y1);ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2)-ff2;ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4)-ff4;}
		}
		else if(i1==i2&&i2==i3&&i3==i4&&j1==j2&&j3==j4&&j1<j3){ff4=rc2(x4,y4,x1,y1,x2,y2,x3,y3,1);ff2=rc2(x4,y4,x1,y1,x2,y2,x3,y3,3);}
		else if(i1==i2&&i3==i4&&j1==j2&&j3==j4&&j1<j3){
		  if(x2*y3-x3*y2<0.){jj1++;ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4);ff2=rc2(x4,y4,x1,y1,x2,y2,x3,y3,3);ff1=tr2(x4,y4,x1,y1)-tr2(x2,y2,x3,y3);}
		  else if(x4*y1-x1*y4<0.){jj3--;ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2);ff4=rc2(x4,y4,x1,y1,x2,y2,x3,y3,1);ff3=tr2(x2,y2,x3,y3)-tr2(x4,y4,x1,y1);}
		  else {jj1++;jj3--;ff3=tr2(x2,y2,x3,y3);ff1=tr2(x4,y4,x1,y1);ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2)-ff1;ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4)-ff3;}
		}
		//12-3-4
		else if(i1==i2&&i3==i4&&j1==j2&&j2==j4){
		  if(x2*y3-x3*y2>0.) {jj2--;ff2=tr2(x2,y2,x3,y3);ff3=rc1(x3,y3,x4,y4,x2,y2);ff4=rc1(x4,y4,x3,y3,x1,y1);ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2)-ff2;}
		  else {ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2);ff3=tr1(x3,y3,x2,y2,x4,y4,1);ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4)-ff3;}
		}
		else if(i1==i2&&i2==i3&&j1==j2&&j2==j4){
		  if(x3*y4-x4*y3>0.) {jj2--;ii3--;ff3=tr2(x3,y3,x4,y4);ff2=rc1(x3,y3,x2,y2,x4,y4);ff4=rc1(x4,y4,x3,y3,x1,y1);ff1=ds-ff2-ff3-ff4;}
		  else {ff4=tr1(x4,y4,x3,y3,x1,y1,2);ff3=tr1(x3,y3,x2,y2,x4,y4,1);ff1=ds-ff3-ff4;}
		}
		else if(i1==i2&&i2==i3&&j1==j2&&j3==j4&&j4<j2){
		  if(x4*y1-x1*y4>0.) {jj2--;ii3--;jj4++;ff4=tr2(x4,y4,x1,y1);ff2=rc1(x3,y3,x2,y2,x4,y4);ff3=rc1(x4,y4,x1,y1,x3,y3);ff1=rc2(x2,y2,x3,y3,x4,y4,x1,y1,1)-ff4;}
		  else {ff4=tr1(x4,y4,x3,y3,x1,y1,2);ff3=rc1(x3,y3,x2,y2,x4,y4)-tr2(x4,y4,x1,y1);ff1=rc2(x2,y2,x3,y3,x4,y4,x1,y1,1);}
		}
		else if(i1==i2&&i3==i4&&j1==j2&&j2==j3){
		  if(x4*y1-x1*y4>0.) {jj1++;ff1=tr2(x4,y4,x1,y1);ff4=rc1(x4,y4,x3,y3,x1,y1);ff3=rc1(x3,y3,x4,y4,x2,y2);ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2)-ff1;}
		  else {ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,2);ff4=tr1(x4,y4,x1,y1,x3,y3,1);ff3=rc1(x3,y3,x4,y4,x2,y2)-tr2(x4,y4,x1,y1);}
		}
		else if(i1==i2&&i2==i4&&j1==j2&&j2==j3){
		  if(x3*y4-x4*y3>0.) {jj1++;ii4--;ff4=tr2(x3,y3,x4,y4);ff1=rc1(x4,y4,x1,y1,x3,y3);ff3=rc1(x3,y3,x4,y4,x2,y2);ff2=ds-ff1-ff3-ff4;}
		  else {ff3=tr1(x3,y3,x4,y4,x2,y2,2);ff4=tr1(x4,y4,x1,y1,x3,y3,1);ff2=ds-ff3-ff4;}
		}
		else if(i1==i2&&i2==i4&&j1==j2&&j3==j4&&j3>j1){
		  if(x2*y3-x3*y2>0.) {jj1++;ii4--;jj3--;ff3=tr2(x2,y2,x3,y3);ff1=rc1(x4,y4,x1,y1,x3,y3);ff4=rc1(x3,y3,x2,y2,x4,y4);ff2=rc2(x4,y4,x1,y1,x2,y2,x3,y3,3)-ff3;}
		  else {ff3=tr1(x3,y3,x4,y4,x2,y2,2);ff4=rc1(x4,y4,x1,y1,x3,y3)-tr2(x2,y2,x3,y3);ff2=rc2(x4,y4,x1,y1,x2,y2,x3,y3,3);}
		}
		//23-1-4
		else if(i1==i2&&i2==i3&&j1==j4&&j2==j3){
		  if(x3*y4-x4*y3>0.) {ii3--;ff3=tr2(x3,y3,x4,y4);ff4=rc1(x4,y4,x3,y3,x1,y1);ff1=rc1(x1,y1,x2,y2,x4,y4);ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3)-ff3;}
		  else {ff2=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3);ff4=tr1(x4,y4,x3,y3,x1,y1,2);ff1=rc1(x1,y1,x2,y2,x4,y4)-tr2(x3,y3,x4,y4);}
		}
		else if(i1==i2&&i2==i3&&j2==j3&&j3==j4){
		  if(x4*y1-x1*y4>0.) {jj4++;ii3--;ff4=tr2(x4,y4,x1,y1);ff3=rc1(x4,y4,x1,y1,x3,y3);ff1=rc1(x1,y1,x2,y2,x4,y4);ff2=ds-ff1-ff3-ff4;}
		  else {ff1=tr1(x1,y1,x4,y4,x2,y2,1);ff4=tr1(x4,y4,x3,y3,x1,y1,2);ff2=ds-ff1-ff4;}
		}
		else if(i1==i4&&i2==i3&&j2==j3&&j3==j4&&i1<i3){
		  if(x1*y2-x2*y1>0.) {jj4++;ii3--;ii1++;ff1=tr2(x1,y1,x2,y2);ff3=rc1(x4,y4,x1,y1,x3,y3);ff4=rc1(x1,y1,x4,y4,x2,y2);ff2=rc2(x2,y2,x3,y3,x4,y4,x1,y1,2)-ff1;}
		  else {ff1=tr1(x1,y1,x4,y4,x2,y2,1);ff4=rc1(x4,y4,x1,y1,x3,y3)-tr2(x1,y1,x2,y2);ff2=rc2(x2,y2,x3,y3,x4,y4,x1,y1,2);}
		}
		else if(i2==i3&&i3==i4&&j1==j4&&j2==j3){
		  if(x1*y2-x2*y1>0.) {ii2++;ff2=tr2(x1,y1,x2,y2);ff1=rc1(x1,y1,x2,y2,x4,y4);ff4=rc1(x4,y4,x3,y3,x1,y1);ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3)-ff2;}
		  else {ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,3);ff1=tr1(x1,y1,x2,y2,x4,y4,2);ff4=rc1(x4,y4,x3,y3,x1,y1)-tr2(x1,y1,x2,y2);}
		}
		else if(i2==i3&&i3==i4&&j1==j2&&j2==j3){
		  if(x4*y1-x1*y4>0.) {jj1++;ii2++;ff1=tr2(x4,y4,x1,y1);ff2=rc1(x1,y1,x4,y4,x2,y2);ff4=rc1(x4,y4,x3,y3,x1,y1);ff3=ds-ff1-ff2-ff4;}
		  else {ff4=tr1(x4,y4,x1,y1,x3,y3,1);ff1=tr1(x1,y1,x2,y2,x4,y4,2);ff3=ds-ff1-ff4;}
		}
		else if(i1==i4&&i2==i3&&j1==j2&&j2==j3&&i4>i2){
		  if(x3*y4-x4*y3>0.) {jj1++;ii2++;ii4--;ff4=tr2(x3,y3,x4,y4);ff2=rc1(x1,y1,x4,y4,x2,y2);ff1=rc1(x4,y4,x1,y1,x3,y3);ff3=rc2(x4,y4,x1,y1,x2,y2,x3,y3,4)-ff4;}
		  else {ff4=tr1(x4,y4,x1,y1,x3,y3,1);ff1=rc1(x1,y1,x4,y4,x2,y2)-tr2(x3,y3,x4,y4);ff3=rc2(x4,y4,x1,y1,x2,y2,x3,y3,4);}
		}
		//34-1-2
		else if(i1==i2&&i3==i4&&j2==j3&&j3==j4){
		  if(x4*y1-x1*y4>0.) {jj4++;ff4=tr2(x4,y4,x1,y1);ff1=rc1(x1,y1,x2,y2,x4,y4);ff2=rc1(x2,y2,x1,y1,x3,y3);ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4)-ff4;}
		  else {ff3=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4);ff1=tr1(x1,y1,x4,y4,x2,y2,1);ff2=rc1(x2,y2,x1,y1,x3,y3)-tr2(x4,y4,x1,y1);}
		}
		else if(i1==i3&&i3==i4&&j2==j3&&j3==j4){
		  if(x1*y2-x2*y1>0.) {jj4++;ii1++;ff1=tr2(x1,y1,x2,y2);ff4=rc1(x1,y1,x4,y4,x2,y2);ff2=rc1(x2,y2,x1,y1,x3,y3);ff3=ds-ff1-ff2-ff4;}
		  else {ff2=tr1(x2,y2,x1,y1,x3,y3,2);ff1=tr1(x1,y1,x4,y4,x2,y2,1);ff3=ds-ff1-ff2;}
		}
		else if(i1==i3&&i3==i4&&j1==j2&&j3==j4&&j2>j4){
		  if(x2*y3-x3*y2>0.) {jj4++;ii1++;jj2--;ff2=tr2(x2,y2,x3,y3);ff4=rc1(x1,y1,x4,y4,x2,y2);ff1=rc1(x2,y2,x3,y3,x1,y1);ff3=rc2(x2,y2,x3,y3,x4,y4,x1,y1,3)-ff2;}
		  else {ff2=tr1(x2,y2,x1,y1,x3,y3,2);ff1=rc1(x1,y1,x4,y4,x2,y2)-tr2(x2,y2,x3,y3);ff3=rc2(x2,y2,x3,y3,x4,y4,x1,y1,3);}
		}
		else if(i1==i2&&i3==i4&&j1==j3&&j3==j4){
		  if(x2*y3-x3*y2>0.) {jj3--;ff3=tr2(x2,y2,x3,y3);ff2=rc1(x2,y2,x1,y1,x3,y3);ff1=rc1(x1,y1,x2,y2,x4,y4);ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4)-ff3;}
		  else {ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,4);ff2=tr1(x2,y2,x3,y3,x1,y1,1);ff1=rc1(x1,y1,x2,y2,x4,y4)-tr2(x2,y2,x3,y3);}
		}
		else if(i2==i3&&i3==i4&&j1==j3&&j3==j4){
		  if(x1*y2-x2*y1>0.) {jj3--;ii2++;ff2=tr2(x1,y1,x2,y2);ff3=rc1(x2,y2,x3,y3,x1,y1);ff1=rc1(x1,y1,x2,y2,x4,y4);ff4=ds-ff1-ff2-ff3;}
		  else {ff1=tr1(x1,y1,x2,y2,x4,y4,2);ff2=tr1(x2,y2,x3,y3,x1,y1,1);ff4=ds-ff1-ff2;}
		}
		else if(i2==i3&&i3==i4&&j1==j2&&j3==j4&&j1<j3){
		  if(x4*y1-x1*y4>0.) {jj3--;ii2++;jj1++;ff1=tr2(x4,y4,x1,y1);ff3=rc1(x2,y2,x3,y3,x1,y1);ff2=rc1(x1,y1,x4,y4,x2,y2);ff4=rc2(x4,y4,x1,y1,x2,y2,x3,y3,1)-ff1;}
		  else {ff1=tr1(x1,y1,x2,y2,x4,y4,2);ff2=rc1(x2,y2,x3,y3,x1,y1)-tr2(x4,y4,x1,y1);ff4=rc2(x4,y4,x1,y1,x2,y2,x3,y3,1);}
		}
		//14-2-3
		else if(i1==i3&&i3==i4&&j1==j4&&j2==j3){
		  if(x1*y2-x2*y1>0.) {ii1++;ff1=tr2(x1,y1,x2,y2);ff2=rc1(x2,y2,x1,y1,x3,y3);ff3=rc1(x3,y3,x4,y4,x2,y2);ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1)-ff1;}
		  else {ff4=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1);ff2=tr1(x2,y2,x1,y1,x3,y3,2);ff3=rc1(x3,y3,x4,y4,x2,y2)-tr2(x1,y1,x2,y2);}
		}
		else if(i1==i3&&i3==i4&&j1==j2&&j2==j4){
		  if(x2*y3-x3*y2>0.) {jj2--;ii1++;ff2=tr2(x2,y2,x3,y3);ff1=rc1(x2,y2,x3,y3,x1,y1);ff3=rc1(x3,y3,x4,y4,x2,y2);ff4=ds-ff1-ff2-ff3;}
		  else {ff3=tr1(x3,y3,x2,y2,x4,y4,1);ff2=tr1(x2,y2,x1,y1,x3,y3,2);ff4=ds-ff2-ff3;}
		}
		else if(i1==i4&&i2==i3&&j1==j2&&j2==j4&&i3>i1){
		  if(x3*y4-x4*y3>0.) {jj2--;ii1++;i3--;ff3=tr2(x3,y3,x4,y4);ff1=rc1(x2,y2,x3,y3,x1,y1);ff2=rc1(x3,y3,x2,y2,x4,y4);ff4=rc2(x2,y2,x3,y3,x4,y4,x1,y1,4)-ff3;}
		  else {ff3=tr1(x3,y3,x2,y2,x4,y4,1);ff2=rc1(x2,y2,x3,y3,x1,y1)-tr2(x3,y3,x4,y4);ff4=rc2(x2,y2,x3,y3,x4,y4,x1,y1,4);}
		}
		else if(i1==i2&&i2==i4&&j1==j4&&j2==j3){
		  if(x3*y4-x4*y3>0.) {ii4--;ff4=tr2(x3,y3,x4,y4);ff3=rc1(x3,y3,x4,y4,x2,y2);ff2=rc1(x2,y2,x1,y1,x3,y3);ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1)-ff4;}
		  else {ff1=rc2(x1,y1,x2,y2,x3,y3,x4,y4,1);ff3=tr1(x3,y3,x4,y4,x2,y2,2);ff2=rc1(x2,y2,x1,y1,x3,y3)-tr2(x3,y3,x4,y4);}
		}
		else if(i1==i2&&i2==i4&&j1==j3&&j3==j4){
		  if(x2*y3-x3*y2>0.) {jj3--;ii4--;ff3=tr2(x2,y2,x3,y3);ff4=rc1(x3,y3,x2,y2,x4,y4);ff2=rc1(x2,y2,x1,y1,x3,y3);ff1=ds-ff2-ff3-ff4;}
		  else {ff2=tr1(x2,y2,x3,y3,x1,y1,1);ff3=tr1(x3,y3,x4,y4,x2,y2,2);ff1=ds-ff2-ff3;}
		}
		else if(i1==i4&&i2==i3&&j1==j3&&j3==j4&&i2<i4){
		  if(x1*y2-x2*y1>0.) {jj3--;ii4--;ii2++;ff2=tr2(x1,y1,x2,y2);ff4=rc1(x3,y3,x2,y2,x4,y4);ff3=rc1(x2,y2,x3,y3,x1,y1);ff1=rc2(x4,y4,x1,y1,x2,y2,x3,y3,2)-ff2;}
		  else {ff2=tr1(x2,y2,x3,y3,x1,y1,1);ff3=rc1(x3,y3,x2,y2,x4,y4)-tr2(x1,y1,x2,y2);ff1=rc2(x4,y4,x1,y1,x2,y2,x3,y3,2);}
		}
		//13-2-4
		else if(i1==i2&& i2==i3&&j1==j3&&j3==j4){ff2=tr1(x2,y2,x3,y3,x1,y1,1);ff4=tr1(x4,y4,x3,y3,x1,y1,2);ff1=ds-ff2-ff4;}
		else if(i1==i3&& i3==i4&&j1==j2&&j2==j3){ff2=tr1(x2,y2,x1,y1,x3,y3,2);ff4=tr1(x4,y4,x1,y1,x3,y3,1);ff3=ds-ff2-ff4;}
		//24-1-3
		else if(i1==i2&& i2==i4&&j2==j3&&j3==j4){ff1=tr1(x1,y1,x4,y4,x2,y2,1);ff3=tr1(x3,y3,x4,y4,x2,y2,2);ff2=ds-ff1-ff3;}
		else if(i2==i3&& i3==i4&&j1==j2&&j2==j4){ff1=tr1(x1,y1,x2,y2,x4,y4,2);ff3=tr1(x3,y3,x2,y2,x4,y4,1);ff4=ds-ff1-ff3;}
		//1-2-3-4
		else if(i1==i2&&i3==i4&&j1==j4&&j2==j3){
		  ff1=rc1(x1,y1,x2,y2,x4,y4);ff2=rc1(x2,y2,x1,y1,x3,y3);ff3=rc1(x3,y3,x4,y4,x2,y2);ff4=rc1(x4,y4,x3,y3,x1,y1);
		}
		else{
		  fprintf(stderr,"Undefined configuration: %d %d: (%d,%d)(%d,%d)(%d,%d)(%d,%d)(%g,%g)(%g,%g)(%g,%g)(%g,%g)\n",j,i,j1,i1,j2,i2,j3,i3,j4,i4,x1,y1,x2,y2,x3,y3,x4,y4);
		  exit(0);
		}
		if(isinf(ff1)||isinf(ff2)||isinf(ff3)||isinf(ff4)){
		  fprintf(stderr,"Inf factor: %d %d: (%d,%d)(%d,%d)(%d,%d)(%d,%d)(%g,%g)(%g,%g)(%g,%g)(%g,%g) %g %g %g %g\n",j,i,j1,i1,j2,i2,j3,i3,j4,i4,x1,y1,x2,y2,x3,y3,x4,y4,ff1,ff2,ff3,ff4);getchar();
		}
		if(fabs(ff1+ff2+ff3+ff4-ds)>1e-6||ff1<0.||ff2<0.||ff3<0.||ff4<0.){
		  fprintf(stderr,"Something Wrong...: %d %d: (%d,%d)(%d,%d)(%d,%d)(%d,%d)(%g,%g)(%g,%g)(%g,%g)(%g,%g) %g %g %g %g %g %g\n",j,i,j1,i1,j2,i2,j3,i3,j4,i4,x1,y1,x2,y2,x3,y3,x4,y4,ff1,ff2,ff3,ff4,ff1+ff2+ff3+ff4,ds);getchar();
		}
		if(ii1<0) ii1=0; else if(ii1>naxes[1]-1) ii1=naxes[1]-1;
		if(ii2<0) ii2=0; else if(ii2>naxes[1]-1) ii2=naxes[1]-1;
		if(ii3<0) ii3=0; else if(ii3>naxes[1]-1) ii3=naxes[1]-1;
		if(ii4<0) ii4=0; else if(ii4>naxes[1]-1) ii4=naxes[1]-1;
		if(jj1<0) jj1=0; else if(jj1>naxes[0]-1) jj1=naxes[0]-1;
		if(jj2<0) jj2=0; else if(jj2>naxes[0]-1) jj2=naxes[0]-1;
		if(jj3<0) jj3=0; else if(jj3>naxes[0]-1) jj3=naxes[0]-1;
		if(jj4<0) jj4=0; else if(jj4>naxes[0]-1) jj4=naxes[0]-1;
		if(isnan(zo[0][ii1][jj1])) zo[0][ii1][jj1]=0.;
		if(isnan(zo[0][ii2][jj2])) zo[0][ii2][jj2]=0.;
		if(isnan(zo[0][ii3][jj3])) zo[0][ii3][jj3]=0.;
		if(isnan(zo[0][ii4][jj4])) zo[0][ii4][jj4]=0.;
		zo[0][ii1][jj1]+=zi[i][j]/NX/NY/4*ff1/ds;
		zo[0][ii2][jj2]+=zi[i][j]/NX/NY/4*ff2/ds;
		zo[0][ii3][jj3]+=zi[i][j]/NX/NY/4*ff3/ds;
		zo[0][ii4][jj4]+=zi[i][j]/NX/NY/4*ff4/ds;
		zo[1][ii1][jj1]+=1./NX/NY/4*ff1/ds;
		zo[1][ii2][jj2]+=1./NX/NY/4*ff2/ds;
		zo[1][ii3][jj3]+=1./NX/NY/4*ff3/ds;
		zo[1][ii4][jj4]+=1./NX/NY/4*ff4/ds;
		//if(j==PX&&i==PY){
		//if(fabs(ff1)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff1=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj1,ii1,ff1/ds,zi[i][j]/NX/NY/4*ff1/ds,zo[0][ii1][jj1]);
		//if(fabs(ff2)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff2=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj2,ii2,ff2/ds,zi[i][j]/NX/NY/4*ff2/ds,zo[0][ii2][jj2]);
		//if(fabs(ff3)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff3=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj3,ii3,ff3/ds,zi[i][j]/NX/NY/4*ff3/ds,zo[0][ii3][jj3]);
		//if(fabs(ff4)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff4=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj4,ii4,ff4/ds,zi[i][j]/NX/NY/4*ff4/ds,zo[0][ii4][jj4]);
		//}
		//if(jj1==PX&&ii1==PY&&fabs(ff1)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff1=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj1,ii1,ff1/ds,zi[i][j]/NX/NY/4*ff1/ds,zo[0][ii1][jj1]);
		//if(jj2==PX&&ii2==PY&&fabs(ff2)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff2=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj2,ii2,ff2/ds,zi[i][j]/NX/NY/4*ff2/ds,zo[0][ii2][jj2]);
		//if(jj3==PX&&ii3==PY&&fabs(ff3)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff3=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj3,ii3,ff3/ds,zi[i][j]/NX/NY/4*ff3/ds,zo[0][ii3][jj3]);
		//if(jj4==PX&&ii4==PY&&fabs(ff4)>1e-5) fprintf(stderr,"%d %d %d: (%d,%d) ff4=%g => %g/%g\n",j,i,iq*8+jq*4+is*2+js,jj4,ii4,ff4/ds,zi[i][j]/NX/NY/4*ff4/ds,zo[0][ii4][jj4]);
	      }
	    }
	  }
	}
      }
    }
    szi=0.;nzi=0;
    szo=0.;nzo=0;
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){
	if(isnan(zi[i][j])==0){
	  szi+=zi[i][j];
	  nzi++;
	}
	if(isnan(zo[0][i][j])==0){
	  szo+=zo[0][i][j];
	  if(zo[1][i][j]>0.5){zo[0][i][j]/=zo[1][i][j];nzo++;}
	  else zo[0][i][j]=0./0.;
	  if(isnan(szo)){fprintf(stderr,"zo[0][%d][%d]=%g szo=%g nzo=%d\n",i,j,zo[0][i][j],szo,nzo);getchar();}
	}
	zi[i][j]=zo[0][i][j];
      }
    }
    nzi=naxes[1]*naxes[0]-nzi;nzo=naxes[1]*naxes[0]-nzo;
    //fprintf(stderr,"Sum : %lf => %lf\n",szi,szo);
    //fprintf(stderr,"NAN : %d => %d\n",nzi,nzo);
    fprintf(stderr,".");
    if(argc==11){
      av=0.;sg=20000.;
      for(n=10;n>=0;n--){
	zs=0.;z2s=0.;ns=0;
	for(i=0;i<naxes[1];i++){
	  for(j=0;j<naxes[0];j++){
	    if(fabs(zi[i][j]-av)<(n*0.1+2)*sg){
	      zs+=zi[i][j];
	      z2s+=zi[i][j]*zi[i][j];
	      ns++;
	    }
	  }
	}
	av=zs/ns;
	sg=sqrt(z2s/ns-av*av);
      }
      fprintf(stderr,"Subtract 2sigma(%g x2) clipped Average: %g\n",sg,av);
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++) zi[i][j]-=av;
      }
    }
    for(i=0;i<naxes[1];i++){
      if(fits_write_img(fpw, TFLOAT, fpixelw, naxes[0], zi[i], &status)) printerror(status);
      fpixelw+=naxes[0];
    }
  }
  if(fits_close_file(fpr, &status)) printerror(status);
  if(fits_close_file(fpw, &status)) printerror(status);
  fprintf(stderr," OK\n");

  return 0;
}

void printerror( int status)
{
  /*****************************************************/
  /* Print out cfitsio error messages and exit program */
  /*****************************************************/

  char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
  if(status)
    fprintf(stderr, "\n*** Error occurred during program execution ***\n");

  fits_get_errstatus(status, status_str);   /* get the error description */
  fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

  /* get first message; null ifstack is empty */
  if( fits_read_errmsg(errmsg) ) 
    {
      fprintf(stderr, "\nError message stack:\n");
      fprintf(stderr, " %s\n", errmsg);

      while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
	fprintf(stderr, " %s\n", errmsg);
    }

  exit( status );       /* terminate the program, returning error status */
}
