#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCUBE 50
#define THRE 6
#define FLATMINX 0.20
#define FLATMINY 0.05

void printerror( int status);

int main(int argc,char *argv[])
{
  fitsfile *fp,*fp0;
  int i,j,k,i1,i2,j1,j2,ii,jj,status,nfound,bitpix=-32;
  long fpixel,fpixel0,naxis=3,naxes[3]={LINEPIX, LINEPIY, 1};
  static float zi[LINEPIY][LINEPIX],zo[LINEPIY][LINEPIX],mk[LINEPIY][LINEPIX];
  float w,ws,crval1,crval2,cd1_1,cd1_2,cd2_1,cd2_2,cd3_3,gain;
  int ns=0,crpix1,crpix2,detid,nbd,i1f,i2f,j1f,j2f;
  char ctype1[16],ctype2[16];
  double zs;
  
  if(argc!=4&&argc!=5){
    printf("Usage: %s input output mask [flat]\n",argv[0]);
    exit(0);
  }
  if(strcmp(argv[1],argv[2])==0){
    fprintf(stderr,"input and output must be different.\n");
    exit(0);
  }
  fprintf(stderr,"Read %s  ... ",argv[3]);
  status=0;
  if(fits_open_file(&fp, argv[3], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, mk[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  for(j=0;j<naxes[0];j++){
    for(k=0;k<4;k++){
      nbd=0;
      for(i=0;i<naxes[1]/4;i++){
	if(mk[i*4+k][j]>3.) nbd++;
      }
      //if(nbd>100) fprintf(stderr,"%d %d %d\n",j,k,nbd);
      if(nbd>naxes[1]/8){
	fprintf(stderr,"### Bad Column detected @(%d 4n+%d) ###\n",j+1,(k+1)%4);
	for(i=0;i<naxes[1]/4;i++) mk[i*4+k][j]=THRE+1e-5;
      }
    }
  }
  if(argc==5){
    fprintf(stderr,"Read %s  ... ",argv[4]);
    status=0;
    if(fits_open_file(&fp, argv[4], READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    fpixel=1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
    i1f=0;i2f=naxes[1];j1f=0;j2f=naxes[0];
    for(i=naxes[1]/2-1;i>=0;i--){
      zs=0.;
      for(j=naxes[0]/4;j<naxes[0]*3/4;j++) zs+=zi[i][j];
      zs/=naxes[0]/2;
      if(zs<FLATMINY){i1f=i+1;break;}
    }
    for(i=naxes[1]/2;i<naxes[1];i++){
      zs=0.;
      for(j=naxes[0]/4;j<naxes[0]*3/4;j++) zs+=zi[i][j];
      zs/=naxes[0]/2;
      if(zs<FLATMINY){i2f=i;break;}
    }
    for(j=naxes[0]/2-1;j>=0;j--){
      zs=0.;
      for(i=i1f;i<i2f;i++) zs+=zi[i][j];
      zs/=i2f-i1f;
      if(zs<FLATMINX){j1f=j+1;break;}
    }
    for(j=naxes[0]/2;j<naxes[0];j++){
      zs=0.;
      for(i=i1f;i<i2f;i++) zs+=zi[i][j];
      zs/=i2f-i1f;
      if(zs<FLATMINX){j2f=j;break;}
    }
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){
	if(i<i1f || i2f<=i || j<j1f || j2f<=j) mk[i][j]=THRE+1e-5;
	else if(zi[i][j]<0.) mk[i][j]=THRE+1e-5;
      }
    }
  }

  fprintf(stderr,"Read %s  ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "NAXIS", &naxis, NULL, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, naxis, naxes, &nfound, &status)) printerror(status);
  if(naxis==3) fprintf(stderr,"%ld x %ld x %ld format \n",naxes[0],naxes[1],naxes[2]);
  else fprintf(stderr,"%ld x %ld format \n",naxes[0],naxes[1]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY||naxes[2]>NCUBE){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  if(fits_read_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
  remove(argv[2]);
  fprintf(stderr,"Write %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fp0, argv[2], &status)) printerror(status);
  if(fits_create_img(fp0, bitpix, naxis, naxes, &status)) printerror(status);
  if(naxis==3) fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
  else fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  if(fits_write_key(fp0, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
  if(strcmp(ctype2,"WAVELENGTH")==0) strcpy(ctype1,"pixel   ");
  if(fits_write_key(fp0, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
  if(fits_write_key(fp0, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
  j1--;i1--;
  j2+=j1;i2+=i1;
  fpixel=1;fpixel0=1;
  for(k=0;k<naxes[2];k++){
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
    }
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){
	if(i<i1 || i2<=i || j<j1 || j2<=j) zo[i][j]=0./0.;
	else{
	  if(mk[i][j]<THRE) zo[i][j]=zi[i][j];
	  else {
	    zo[i][j]=0.;ws=0.;
	    ii=i;jj=j;
	    while(jj>=j1){
	      if(mk[ii][jj]>=THRE) jj--;
	      else break;
	    }
	    if(jj>=j1){
	      w=1./(j-jj);
	      zo[i][j]+=w*zi[ii][jj];
	      ws+=w;
	      //fprintf(stderr,"%d %d %g %d %d %g %g\n",j,i,mk[i][j],jj,ii,mk[ii][jj],zi[ii][jj]);
	    }
	    else{zo[i][j]=0./0.;continue;}
	    ii=i;jj=j;
	    while(jj<j2){
	      if(mk[ii][jj]>=THRE) jj++;
	      else break;
	    }
	    if(jj<j2){
	      w=1./(jj-j);
	      zo[i][j]+=w*zi[ii][jj];
	      ws+=w;
	      //fprintf(stderr,"%d %d %g %d %d %g %g\n",j,i,mk[i][j],jj,ii,mk[ii][jj],zi[ii][jj]);
	    }
	    else{zo[i][j]=0./0.;continue;}
	    ii=i;jj=j;
	    while(ii>=i1){
	      if(mk[ii][jj]>=THRE) ii--;
	      else break;
	    }
	    if(ii>=i1){
	      w=1./(i-ii);
	      zo[i][j]+=w*zi[ii][jj];
	      ws+=w;
	      //fprintf(stderr,"%d %d %g %d %d %g %g\n",j,i,mk[i][j],jj,ii,mk[ii][jj],zi[ii][jj]);
	    }
	    else{zo[i][j]=0./0.;continue;}
	    ii=i;jj=j;
	    while(ii<i2){
	      if(mk[ii][jj]>=THRE) ii++;
	      else break;
	    }
	    if(ii<i2){
	      w=1./(ii-i);
	      zo[i][j]+=w*zi[ii][jj];
	      ws+=w;
	      //fprintf(stderr,"%d %d %g %d %d %g %g\n",j,i,mk[i][j],jj,ii,mk[ii][jj],zi[ii][jj]);
	    }
	    else{zo[i][j]=0./0.;continue;}
	    zo[i][j]/=ws;
	    ns++;
	    //fprintf(stderr,"\n%d: %d %d %g %g %g ",ns,j,i,zi[i][j],zo[i][j],mk[i][j]);//getchar();
	  }
	}
      }
    }    
    for(i=0;i<naxes[1];i++){
      if(fits_write_img(fp0, TFLOAT, fpixel0, naxes[0], zo[i], &status)) printerror(status);
      fpixel0+=naxes[0];
    }
    fprintf(stderr,".");
  }
  if(fits_close_file(fp0, &status)) printerror(status);
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr," OK\n");
  fprintf(stderr,"Replaced pixels: %ld pix/image\n",ns/naxes[2]);
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
