#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fitsio.h"
#define LINEPIX 2220
#define LINEPIY 1360

int printerror( int status);

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int nfound,status, ag, nsat,i,j;
  char fname[128], direc[128], com[256], value[64];
  long fpixel,naxes[3]={LINEPIX, LINEPIY, 1};
  unsigned short z[LINEPIY][LINEPIX];
  FILE *in_pipe;
  double zs;

  if(argc == 1){
    printf("Usage: %s directory [KeyWord1] [KeyWord2] ...\n", argv[0]);
    exit(1);
  }

  sscanf(argv[1],"%s",direc);
  if(strcmp(direc,"./")==0) direc[0]='\0';

  sprintf(com,"ls %sTRCS????????.fits",direc);
  if((in_pipe = popen(com,"r")) == NULL){
    fprintf(stderr,"Command (%s) returns NULL.\n",com);
    exit(1);
  }
  
  printf("#File_name        Target_name  Nframe  Exptime  Gain Average >15000pix\n");
  while(fgets(fname,128,in_pipe)!=NULL){
    fname[strlen(fname)-1]='\0';
    status=0;
    if(fits_open_file(&fp, fname, READONLY, &status)) printerror(status);
    if(status) continue;
    if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
    //fname[strlen(fname)-5]='\0';strcat(fname,"[0]");
    printf("%s ", fname);
    if(fits_read_keyword(fp, "OBJECT", value, NULL, &status)) {printf(" # Not found # ");status=0;} else printf("%-16s", value);
    if(fits_read_keyword(fp, "NAXIS3", value, NULL, &status)) {printf(" # Not found # ");status=0;} else printf("%3s", value);
    if(fits_read_keyword(fp, "EXPTIME1", value, NULL, &status)) {printf(" # Not found # ");status=0;} else printf("%10s", value);
    if(fits_read_keyword(fp, "GAIN", value, NULL, &status)) {printf(" # Not found # ");status=0;} else printf("%5s", value);
    //fprintf(stderr,"OK\n");
    for(ag=2;ag<argc;ag++){
	if(fits_read_keyword(fp, argv[ag], value, NULL, &status)) {printf(" # Not found # ");status=0;} else printf(" %s", value);
    }
    if((naxes[0]>LINEPIX)||(naxes[1]>LINEPIY)){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    fpixel = 1;
    zs=0.;nsat=0;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TUSHORT, fpixel, naxes[0], NULL, z[i], NULL, &status)) break;
      fpixel+=naxes[0];
      for(j=0;j<naxes[0];j++){
	zs+=z[i][j];
	if(z[i][j]>15000) nsat++;
      }
    }
    if(status){
      printf("\n");
      continue;
    }
    zs/=naxes[0]*naxes[1];
    printf(" %g %d", zs,nsat);
    printf("\n");
    if(fits_close_file(fp, &status)) printerror(status);
  }
  pclose(in_pipe);
  return 0;
}


int printerror( int status )
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    return( status );       /* terminate the program, returning error status */
}
