#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 3220
#define LINEPIY 2360
#define BUFFER 512
#define BUF 100

void printerror( int status);

void radec2xy(float ra0,float dec0,double ra,double dec,float *x,float *y){
  ra0*=M_PI/180;
  dec0*=M_PI/180;
  ra*=M_PI/180;
  dec*=M_PI/180;
  *y=asin(sin(dec)*cos(dec0)-cos(dec)*cos(ra-ra0)*sin(dec0));
  *x=asin(cos(dec)*sin(ra-ra0)/cos(*y));
  *x*=180/M_PI;
  *y*=180/M_PI;
}

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,k,status,nfound,bitpix=-32;
  long fpixel,naxis=2,naxes[2],naxes_[2];
  static float z[LINEPIY+BUFFER*2][LINEPIX+BUFFER*2],zt[LINEPIY][LINEPIX];
  static int n[LINEPIY+BUFFER*2][LINEPIX+BUFFER*2];
  float crval1,crval2,cd1_1,cd1_2,cd2_1,cd2_2,crval1_,crval2_,cd1_1_,cd1_2_,cd2_1_,cd2_2_,xt,yt;
  int crpix1,crpix2,crpix1_,crpix2_,imin,imax,jmin,jmax,is[BUF],js[BUF],ns;
  FILE *fpr;
  char buf[256],fname[64];
  int cl[10][9]={
    {1,1,1,1,0,0,1,0,0},{1,1,1,0,0,0,0,0,0},{1,1,1,0,0,1,0,0,1},
    {1,0,0,1,0,0,1,0,0},{0,0,0,0,0,0,0,0,0},{0,0,1,0,0,1,0,0,1},
    {1,0,0,1,0,0,1,1,1},{0,0,0,0,0,0,1,1,1},{0,0,1,0,0,1,1,1,1},
    {1,1,1,1,1,1,1,1,1}},md,md0,ddi,ddj;
  double dz2s,dz2smin,r2s,rs,zrs,z2s,zs,a,b;

  if(argc!=3 && argc!=4){
    printf("Usage: %s input_list output [+]\n",argv[0]);
    exit(1);
  }
  
  for(i=0;i<LINEPIY+BUFFER*2;i++){
    for(j=0;j<LINEPIX+BUFFER*2;j++){
      z[i][j]=0.;
      n[i][j]=0;
    }
  }
  fprintf(stderr,"Read: %s ...\n",argv[1]);
  if((fpr = fopen(argv[1],"r")) == NULL){
    fprintf(stderr,"Cannot open %s\n",argv[1]);
    exit(0);
  }
  k=0;
  while(!feof(fpr)){
    if(fgets(buf,sizeof(buf),fpr)==NULL) continue;
    if(buf[0]=='#') continue;
    sscanf(buf,"%s",fname);
    fprintf(stderr,"Read %s  ... ",fname);
    status=0;
    if(fits_open_file(&fp, fname, READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if((naxes[0]>LINEPIX)||(naxes[1]>LINEPIY)){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
    if(k==0){
      crval1_=crval1;crval2_=crval2;
      crpix1_=crpix1;crpix2_=crpix2;
      cd1_1_=cd1_1,cd1_2_=cd1_2;
      cd2_1_=cd2_1,cd2_2_=cd2_2;
    }
    else{
      if(fabs(cd1_1-cd1_1_)>1e-5){
	fprintf(stderr,"CD1_1 is not the same value (%g != %g)\n",cd1_1,cd1_1_);exit(0);
      }
      if(fabs(cd1_2-cd1_2_)>1e-5){
	fprintf(stderr,"CD1_2 is not the same value (%g != %g)\n",cd1_2,cd1_2_);exit(0);
      }
      if(fabs(cd2_1-cd2_1_)>1e-5){
	fprintf(stderr,"CD2_1 is not the same value (%g != %g)\n",cd2_1,cd2_1_);exit(0);
      }
      if(fabs(cd2_2-cd2_2_)>1e-5){
	fprintf(stderr,"CD2_2 is not the same value (%g != %g)\n",cd2_2,cd2_2_);exit(0);
      }
    }
    radec2xy(crval1_,crval2_,crval1,crval2,&xt,&yt);
    js[k]=( cd2_2_*xt-cd1_2_*yt)/(cd1_1_*cd2_2_-cd1_2_*cd2_1_)+crpix1_-crpix1+0.5+BUFFER;
    is[k]=(-cd2_1_*xt+cd1_1_*yt)/(cd1_1_*cd2_2_-cd1_2_*cd2_1_)+crpix2_-crpix2+0.5+BUFFER;
    fpixel = 1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zt[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    if(k==0){
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++){
	  if(isnan(zt[i][j])==0){
	    z[i+is[k]][j+js[k]]+=zt[i][j];
	    n[i+is[k]][j+js[k]]++;
	  }
	}
      }
      fprintf(stderr,"%+d %+d offset, OK\n",js[k]-BUFFER,is[k]-BUFFER);
      imin=is[k];imax=is[k];
      jmin=js[k];jmax=js[k];
      k++;
      continue;
    }
    dz2smin=1e99;md0=9;
    while(md0!=4){
      md=4;
      for(ddi=-1;ddi<=1;ddi++){
	for(ddj=-1;ddj<=1;ddj++){
	  if(cl[md0][3*(ddi+1)+(ddj+1)]){
	    r2s=0.;rs=0.;zrs=0.;z2s=0.;zs=0.;ns=0;
	    for(i=BUF;i<naxes[1]-BUF;i++){
	      for(j=BUF;j<naxes[0]-BUF;j++){
		if(n[i+is[k]+ddi][j+js[k]+ddj]>0){
		  r2s+=(double)z[i+is[k]+ddi][j+js[k]+ddj]*z[i+is[k]+ddi][j+js[k]+ddj]/n[i+is[k]+ddi][j+js[k]+ddj]/n[i+is[k]+ddi][j+js[k]+ddj];
		  rs+=(double)z[i+is[k]+ddi][j+js[k]+ddj]/n[i+is[k]+ddi][j+js[k]+ddj];
		  zrs+=(double)zt[i][j]*z[i+is[k]+ddi][j+js[k]+ddj]/n[i+is[k]+ddi][j+js[k]+ddj];
		  z2s+=(double)zt[i][j]*zt[i][j];
		  zs+=zt[i][j];
		  ns++;
		}
	      }
	    }
	    r2s/=ns;
	    rs/=ns;
	    zrs/=ns;
	    z2s/=ns;
	    zs/=ns;
	    a=(zrs-zs*rs)/(r2s-rs*rs);
	    b=(zs*r2s-zrs*rs)/(r2s-rs*rs);
	    dz2s=a*a*r2s+b*b+z2s+2*a*b*rs-2*a*zrs-2*b*zs;
	    if(dz2smin>dz2s){
	      dz2smin=dz2s;
	      md=3*(ddi+1)+(ddj+1);
	    }
	    //fprintf(stderr,"%d: %d %d %d %g %g %g %g %g %g %d\n",k,md,ddj,ddi,dz2s,r2s,rs,zrs,z2s,zs,ns);
	  }
	}
      }
      md0=md;
      js[k]+=md%3-1;
      is[k]+=md/3-1;
      //fprintf(stderr,"%d: %d %d %d %g",k,md0,js,is,dz2smin);
    }
    //fprintf(stderr,"%2d:%3d%3d %g\n",k,js,is,dz2smin);
    if(jmin>js[k]) jmin=js[k];
    if(jmax<js[k]) jmax=js[k];
    if(imin>is[k]) imin=is[k];
    if(imax<is[k]) imax=is[k];
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){
	if(isnan(zt[i][j])==0){
	  z[i+is[k]][j+js[k]]+=zt[i][j];
	  n[i+is[k]][j+js[k]]++;
	}
      }
    }
    fprintf(stderr,"%+d %+d offset, OK\n",js[k]-BUFFER,is[k]-BUFFER);
    if(abs(js[k]-BUFFER)>BUFFER || abs(is[k]-BUFFER)>BUFFER){
      fprintf(stderr,"BUFFER must be larger than %d and %d\n",abs(js[k]-BUFFER),abs(is[k]-BUFFER));
      exit(0);
    }
    k++;
  }
  fclose(fpr);
  for(i=imin;i<imax+LINEPIY;i++){
    for(j=jmin;j<jmax+LINEPIX;j++){
      if(n[i][j]==0) z[i-imin][j-jmin]=0./0.;
      else z[i-imin][j-jmin]=z[i][j]/n[i][j];
    }
  }
  naxes_[0]=naxes[0]+jmax-jmin;
  naxes_[1]=naxes[1]+imax-imin;
  crpix1_+=BUFFER-jmin;
  crpix2_+=BUFFER-imin;
  remove(argv[2]);
  fprintf(stderr,"Write %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fp, argv[2], &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes_, &status)) printerror(status);
  fpixel=1;
  fprintf(stderr,"%ld x %ld format ",naxes_[0],naxes_[1]);
  if(fits_write_key(fp, TFLOAT, "CRVAL1", &crval1_, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL2", &crval2_, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX1", &crpix1_, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX2", &crpix2_, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE1", "RA---TAN", NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE2", "DEC--TAN", NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD1_1", &cd1_1_, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD1_2", &cd1_2_, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_1", &cd2_1_, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_2", &cd2_2_, NULL, &status)) printerror(status);
  for(i=0;i<naxes_[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes_[0], z[i], &status)) printerror(status);
    fpixel+=naxes_[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  if(argc==4){
    fprintf(stderr,"Read: %s ...\n",argv[1]);
    if((fpr = fopen(argv[1],"r")) == NULL){
      fprintf(stderr,"Cannot open %s\n",argv[1]);
      exit(0);
    }
    k=0;
    while(!feof(fpr)){
      if(fgets(buf,sizeof(buf),fpr)==NULL) continue;
      if(buf[0]=='#') continue;
      sscanf(buf,"%s",fname);
      fprintf(stderr,"Read %s  ... ",fname);
      status=0;
      if(fits_open_file(&fp, fname, READONLY, &status)) printerror(status);
      if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
      fpixel = 1;
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zt[i], NULL, &status)) printerror(status);
	fpixel+=naxes[0];
      }
      if(fits_close_file(fp, &status)) printerror(status);
      for(i=0;i<naxes_[1];i++){
	for(j=0;j<naxes_[0];j++) z[i][j]=0./0.;
      }
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++) z[i+is[k]-imin][j+js[k]-jmin]=zt[i][j];
      }
      sscanf(argv[2],"%s",fname);
      sprintf(fname+strlen(fname)-5,"_%d.fits",k);
      remove(fname);
      fprintf(stderr,"Write %s ... ",fname);
      status=0;
      if(fits_create_file(&fp, fname, &status)) printerror(status);
      if(fits_create_img(fp, bitpix, naxis, naxes_, &status)) printerror(status);
      fpixel=1;
      if(fits_write_key(fp, TFLOAT, "CRVAL1", &crval1_, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CRVAL2", &crval2_, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TINT, "CRPIX1", &crpix1_, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TINT, "CRPIX2", &crpix2_, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TSTRING, "CTYPE1", "RA---TAN", NULL, &status)) printerror(status);
      if(fits_write_key(fp, TSTRING, "CTYPE2", "DEC--TAN", NULL, &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CD1_1", &cd1_1_, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CD1_2", &cd1_2_, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CD2_1", &cd2_1_, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CD2_2", &cd2_2_, NULL, &status)) printerror(status);
      for(i=0;i<naxes_[1];i++){
	if(fits_write_img(fp, TFLOAT, fpixel, naxes_[0], z[i], &status)) printerror(status);
	fpixel+=naxes_[0];
      }
      if(fits_close_file(fp, &status)) printerror(status);
      fprintf(stderr,"OK\n");
      k++;
    }
    fclose(fpr);
  }
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
