#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define MAXODR 30
#define CA 200

void printerror( int status);

double fit(int od1, int od2, float x[], float y[], int n1, int n2, double c[MAXODR], double sig)
{
  double sx[MAXODR*2],xn[MAXODR*2],dy=1.,yf[LINEPIX];
  double p[MAXODR][MAXODR],q[MAXODR][MAXODR],xs[MAXODR],xy[MAXODR],ys[MAXODR],f;
  int    i,j,k,n,ns,lp,rj[LINEPIX];

  od2++;
  for(n=n1;n<n2;n++) yf[n]=y[n];
  if(sig<0.001) lp=0; else lp=10;
  for(;lp>=0;lp--){
    xn[0]=1;
    sx[0]=1;
    ns=0;
    for(n=n1;n<n2;n++) rj[n]=0;
    for(n=n1;n<n2;n++){
      if(fabs(y[n]-yf[n])>sig*(1+lp*0.1)*dy){
	for(i=-1;i<=1;i++){
	  if(0<=n+i && n+i<LINEPIX) rj[n+i]=1;
	}
      }
    }
    for(n=n1;n<n2;n++){
      if(isnan(x[n])) continue;
      if(rj[n]){
	if(lp==0){
	  //fprintf(stderr,"%g %g %g %g\n",x[n],y[n],y[n]-yf[n],sig*dy);
	  //y[n]=0.;
	}
	continue;
      }
      ns++;
    }
    if(ns<od2*2) {
      //fprintf(stderr,"ns(%d)<od2*2(%d)\n",ns,od2*2);
      for(k=0;k<MAXODR;k++) c[k]=0./0.;
      return(0.);
    }
    for(i=1;i<od2*2;i++) sx[i]=0;
    for(i=0;i<od2;i++) xy[i]=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n])) continue;
      if(rj[n]) continue;
      for(i=1;i<od2*2;i++) {
        xn[i]=xn[i-1]*x[n];
        sx[i]+=xn[i]/ns;}
      for(i=0;i<od2;i++){
	if(isnan(y[n])) xy[i]+=xn[i]/ns;
	else xy[i]+=xn[i]*y[n]/ns;
      }
    }
    for(j=0;j<od2;j++){
      for(i=0;i<od2;i++){
        p[i][j]=sx[i+j];
        if(i==j) q[i][j]=1;
        else q[i][j]=0;
      }
    }
    for(k=od1;k<od2;k++){
      for(j=k+1;j<od2;j++){
        f=p[k][j]/p[k][k];
        for(i=0;i<od2;i++){
          p[i][j]-=f*p[i][k];
          q[i][j]-=f*q[i][k];
        }
      }
      f=p[k][k];
      for(i=od1;i<od2;i++){
        p[i][k]/=f;
        q[i][k]/=f;
      }
    }
    k=od2-1;
    for(j=od1;j<k+1;j++){
      ys[j]=0;
      for(i=od1;i<k+1;i++) ys[j]+=q[i][j]*xy[i];
    }
    for(j=k;j>=od1;j--){
      xs[j]=0;
      for(i=j+1;i<k+1;i++) xs[j]+=p[i][j]*c[i];
      c[j]=ys[j]-xs[j];
    }
    dy=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      yf[n]=c[0];
      xn[0]=1;
      for(i=1;i<=k;i++){
	xn[i]=xn[i-1]*x[n];
	yf[n]+=c[i]*xn[i];
      }
      dy+=(y[n]-yf[n])*(y[n]-yf[n])/ns;
    }
    dy=sqrt(dy);
  }
  return(dy);
}

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,i1,i2,j1,j2,status,nfound,bitpix=-32;
  long fpixel,naxis=2,naxes[2]={LINEPIX, LINEPIY};
  static float z[LINEPIY][LINEPIX];
  static double zx[LINEPIY][LINEPIX],zy[LINEPIY][LINEPIX];
  double zs0,zs,zs1,co[MAXODR];
  int detid,crpix1,crpix2,md,od,m;
  float crval1,crval2,cd1_1,cd1_2,cd2_1,cd2_2,x[LINEPIX],y[LINEPIX],gain;
  
  if(argc!=4 && argc!=5){
    printf("Usage: %s input output x/y/xy/xr/yr/xyr/xn/yn/xyn [order]\n",argv[0]);
    exit(0);
  }
  if(strcmp(argv[3],"x")==0) md=0;
  else if(strcmp(argv[3],"xr")==0) md=1;
  else if(strcmp(argv[3],"xd")==0) md=2;
  else if(strcmp(argv[3],"y")==0) md=3;
  else if(strcmp(argv[3],"yr")==0) md=4;
  else if(strcmp(argv[3],"yd")==0) md=5;
  else if(strcmp(argv[3],"xy")==0) md=6;
  else if(strcmp(argv[3],"xyr")==0) md=7;
  else if(strcmp(argv[3],"xyd")==0) md=8;
  else{
    fprintf(stderr,"Mode (%s) must be x/y/xy/xr/yr/xyr/xd/yd/xyd.\n",argv[3]);
    exit(0);
  }
  if(argc==5) sscanf(argv[4],"%d",&od);
  else od=0;
  if(od>=MAXODR){
    fprintf(stderr,"order must be less than %d\n",MAXODR);
    exit(0);
  }
  fprintf(stderr,"Read %s  ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  if(fits_read_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, z[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");

  for(i=0;i<naxes[1];i++){
    for(j=0;j<naxes[0];j++){
      if(isnan(z[i][j])==0) break;
    }
    if(j!=naxes[0]) break;
  }
  i1=i;
  for(i=naxes[1]-1;i>=0;i--){
    for(j=0;j<naxes[0];j++){
      if(isnan(z[i][j])==0) break;
    }
    if(j!=naxes[0]) break;
  }
  i2=i+1;
  for(j=0;j<naxes[0];j++){
    for(i=0;i<naxes[1];i++){
      if(isnan(z[i][j])==0) break;
    }
    if(i!=naxes[1]) break;
  }
  j1=j;
  for(j=naxes[0]-1;j>=0;j--){
    for(i=0;i<naxes[1];i++){
      if(isnan(z[i][j])==0) break;
    }
    if(i!=naxes[1]) break;
  }
  j2=j+1;
  //fprintf(stderr,"%d %d %d %d %d\n",j1,j2,i1,i2,od);
  if(md/3!=1){
    for(j=j1;j<j2;j++) x[j]=j-naxes[0]/2;
    for(i=i1;i<i2;i++){
      for(j=j1;j<j2;j++) y[j]=z[i][j];
      fit(0,od,x,y,j1,j2,co,3);
      //for(m=od;m>=0;m--) fprintf(stderr," %g",co[m]*pow(10.,m*3));getchar();
      for(j=j1;j<j2;j++){
	zy[i][j]=0.;
	for(m=od;m>=0;m--) zy[i][j]=zy[i][j]*x[j]+co[m];
	//if(i==naxes[1]/2){fprintf(stderr,"X: %d %d %g %g",j,i,x[j],zy[i][j]);getchar();}
      }
    }
  }
  else{
    for(i=i1;i<i2;i++){
      for(j=j1;j<j2;j++) zy[i][j]=1.;
    }
  }
  if(md/3!=0){
    for(i=i1;i<i2;i++) x[i]=i-naxes[1]/2;
    for(j=j1;j<j2;j++){
      for(i=i1;i<i2;i++) y[i]=z[i][j];
      fit(0,od,x,y,i1,i2,co,3);
      //if(j==989||j==990){for(m=od;m>=0;m--) fprintf(stderr," %g",co[m]*pow(10.,m*3));getchar();}
      for(i=i1;i<i2;i++){
	zx[i][j]=0.;
	for(m=od;m>=0;m--) zx[i][j]=zx[i][j]*x[i]+co[m];
	//if(j==naxes[0]/2){fprintf(stderr,"Y: %d %d %g %g",j,i,x[j],zx[i][j]);getchar();}
      }
    }
  }
  else{
    for(i=i1;i<i2;i++){
      for(j=j1;j<j2;j++) zx[i][j]=1.;
    }
  }
  zs0=0.;zs=0.;zs1=0;
  for(i=i1;i<i2;i++){
    for(j=j1;j<j2;j++){
      if(isnan(z[i][j])==0 && abs(i1+i2-2*i)<CA && abs(j1+j2-2*j)<CA){
	zs0+=z[i][j];
	zs+=zy[i][j]*zx[i][j];
	zs1+=z[i][j]/(zy[i][j]*zx[i][j]);
      }
    }
  }
  if(md%3==1) zs0=zs1/(CA-1)/(CA-1);
  else zs0/=zs;
  for(i=i1;i<i2;i++){
    for(j=j1;j<j2;j++){
      //if(j==990&&i==760){fprintf(stderr,"%d %d %g %g %g\n",j,i,zx[i][j],zy[i][j],z[i][j]);}
      if(md%3==1) z[i][j]/=zy[i][j]*zx[i][j]*zs0;
      else if(md%3==2) z[i][j]-=zy[i][j]*zx[i][j]*zs0;
      else z[i][j]=zy[i][j]*zx[i][j]*zs0;
    }
  }
  //fprintf(stderr,"[%d:%d,%d:%d] %g %g\n",(j1+j2-CA)/2,(j1+j2+CA)/2,(i1+i2-CA)/2,(i1+i2+CA)/2,zs0,zs);
  remove(argv[2]);
  fprintf(stderr,"Write %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fp, argv[2], &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  cd1_2=0.;cd2_1=0.;
  if(fits_write_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
  fpixel=1;
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], z[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
