#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCUBE 50
#define BUF 100
#define NOBJ 10
#define MAXODR 7
#define HW 5
#define SIGMA 30
#define RNO 7.1

void printerror( int status);

double fit(int od1, int od2, float x[], float y[], int n1, int n2, double c[MAXODR], double sig)
{
  double sx[MAXODR*2],xn[MAXODR*2],dy=1.,yf[LINEPIX];
  double p[MAXODR][MAXODR],q[MAXODR][MAXODR],xs[MAXODR],xy[MAXODR],ys[MAXODR],f;
  int    i,j,k,n,ns,lp,rj[LINEPIX];

  od2++;
  for(n=n1;n<n2;n++) yf[n]=y[n];
  if(sig<0.001) lp=0; else lp=10;
  for(;lp>=0;lp--){
    xn[0]=1;
    sx[0]=1;
    ns=0;
    for(n=n1;n<n2;n++) rj[n]=0;
    for(n=n1;n<n2;n++){
      if(fabs(y[n]-yf[n])>sig*(1+lp*0.1)*dy){
	for(i=-1;i<=1;i++){
	  if(0<=n+i && n+i<LINEPIX) rj[n+i]=1;
	}
      }
    }
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]){
	if(lp==0){
	  //fprintf(stderr,"%g %g %g %g\n",x[n],y[n],y[n]-yf[n],sig*dy);
	  //y[n]=0.;
	}
	continue;
      }
      ns++;
    }
    if(ns<od2*2) {
      //fprintf(stderr,"ns(%d)<od2*2(%d)\n",ns,od2*2);
      for(k=0;k<MAXODR;k++) c[k]=0./0.;
      return(0.);
    }
    for(i=1;i<od2*2;i++) sx[i]=0;
    for(i=0;i<od2;i++) xy[i]=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      for(i=1;i<od2*2;i++) {
        xn[i]=xn[i-1]*x[n];
        sx[i]+=xn[i]/ns;}
      for(i=0;i<od2;i++) xy[i]+=xn[i]*y[n]/ns;
    }
    for(j=0;j<od2;j++){
      for(i=0;i<od2;i++){
        p[i][j]=sx[i+j];
        if(i==j) q[i][j]=1;
        else q[i][j]=0;
      }
    }
    for(k=od1;k<od2;k++){
      for(j=k+1;j<od2;j++){
        f=p[k][j]/p[k][k];
        for(i=0;i<od2;i++){
          p[i][j]-=f*p[i][k];
          q[i][j]-=f*q[i][k];
        }
      }
      f=p[k][k];
      for(i=od1;i<od2;i++){
        p[i][k]/=f;
        q[i][k]/=f;
      }
    }
    k=od2-1;
    for(j=od1;j<k+1;j++){
      ys[j]=0;
      for(i=od1;i<k+1;i++) ys[j]+=q[i][j]*xy[i];
    }
    for(j=k;j>=od1;j--){
      xs[j]=0;
      for(i=j+1;i<k+1;i++) xs[j]+=p[i][j]*c[i];
      c[j]=ys[j]-xs[j];
    }
    dy=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      yf[n]=c[0];
      xn[0]=1;
      for(i=1;i<=k;i++){
	xn[i]=xn[i-1]*x[n];
	yf[n]+=c[i]*xn[i];
      }
      dy+=(y[n]-yf[n])*(y[n]-yf[n])/ns;
    }
    dy=sqrt(dy);
  }
  return(dy);
}

int main(int argc,char *argv[])
{
  fitsfile *fp,*fp0;
  int i,j,k,i1,i2,j1,j2,status,nfound,bitpix=-32;
  long fpixel,fpixel0,naxis=3,naxes[3]={LINEPIX, LINEPIY, NCUBE};
  static float zi[LINEPIY][LINEPIX],zo[LINEPIY][LINEPIX],zr[LINEPIY][LINEPIX],zn[LINEPIY][LINEPIX],zd2s[LINEPIY][LINEPIX],za[LINEPIY][LINEPIX],zb[LINEPIY][LINEPIX],zf[LINEPIY][LINEPIX],na[LINEPIY][LINEPIX];
  static int ns[LINEPIY][LINEPIX];
  float zt[LINEPIX],x[LINEPIX],crval1,crval2,cd1_1,cd1_2,cd2_1,cd2_2,cd3_3=1.,gain;
  int is,js,di,dj,ddi,ddj,n,nr,nr0=-2,argc_ref,crpix1,crpix2,detid,om,nob,m,ddj1,ddj2;
  int cl[10][9]={
    {1,1,1,1,0,0,1,0,0},{1,1,1,0,0,0,0,0,0},{1,1,1,0,0,1,0,0,1},
    {1,0,0,1,0,0,1,0,0},{0,0,0,0,0,0,0,0,0},{0,0,1,0,0,1,0,0,1},
    {1,0,0,1,0,0,1,1,1},{0,0,0,0,0,0,1,1,1},{0,0,1,0,0,1,1,1,1},
    {1,1,1,1,1,1,1,1,1}},md,md0;
  double dz2s,dz2smin,r2s,rs,zrs,z2s,zs,a,b,amin=0.,bmin,sig,sigfac=0.90;
  double zx[6][LINEPIX],zy[NOBJ][LINEPIY],xc[2][NOBJ],co[MAXODR],zm,dxs,w[NOBJ];
  char ctype1[16],ctype2[16],cunit2[16];

  if(argc!=3&&argc!=4){
    printf("Usage: %s input output [reference/auto]\n",argv[0]);
    exit(0);
  }
  if(argc==4){
    if(strcmp(argv[3],"auto")==0||strcmp(argv[3],"AUTO")==0) argc_ref=0;
    else argc_ref=3;
  }
  else argc_ref=0;
  while(1){
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){
	zo[i][j]=0.;
	zr[i][j]=0.;
	zn[i][j]=20000.;
	zd2s[i][j]=0.;
	za[i][j]=0.;
	ns[i][j]=0;
	na[i][j]=0.;
	zb[i][j]=0.;
	zf[i][j]=0.;
      }
    }
    if(argc_ref){
      fprintf(stderr,"Read %s  ... ",argv[argc_ref]);
      status=0;
      if(fits_open_file(&fp, argv[argc_ref], READONLY, &status)) printerror(status);
      if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
      fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
      if(naxes[0]>LINEPIX||naxes[1]>LINEPIY||naxes[2]>NCUBE){
	fprintf(stderr,"\nIllegal image format\n");
	exit(0);
      }
      if(naxes[2]!=2 && naxes[2]!=3){
	fprintf(stderr,"NAXIS3 of the reference image must be 2 or 3. (%ld)\n",naxes[2]);
	exit(0);
      }
      fpixel=1;
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zr[i], NULL, &status)) printerror(status);
	fpixel+=naxes[0];
      }
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zn[i], NULL, &status)) printerror(status);
	fpixel+=naxes[0];
      }
      if(naxes[2]==3){
	om=0;
	for(i=0;i<naxes[1];i++){
	  if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zf[i], NULL, &status)) printerror(status);
	  fpixel+=naxes[0];
	}
      }
      else om=1;
      if(fits_close_file(fp, &status)) printerror(status);
      fprintf(stderr,"OK\n");
    }
    fprintf(stderr,"Read %s  ... ",argv[1]);
    status=0;
    if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld x %ld format \n",naxes[0],naxes[1],naxes[2]);
    if(naxes[0]>LINEPIX||naxes[1]>LINEPIY||naxes[2]>NCUBE){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    if(fits_read_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
    if(strcmp(ctype2,"WAVELENGTH")) om=1; 
    else{
      if(om){
	fprintf(stderr,"NAXIS3 of the reference image must be 3.\n");
	exit(0);
      }
      om=0;
    }
    //fprintf(stderr,"%s %d\n",ctype2,om);getchar();
    if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
    if(om){
      if(fits_read_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
      if(fits_read_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
      if(fits_read_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
      if(fits_read_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
      if(fits_read_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
      if(fits_read_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
      if(fits_read_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
      j1--;i1--;
      j2+=j1;i2+=i1;
      js=(naxes[0]-j1-j2)/2;
      is=(naxes[1]-i1-i2)/2;
    }
    else{
      if(fits_read_key(fp, TSTRING, "CUNIT2", cunit2, NULL, &status)) printerror(status);
      js=0;is=0;
    }
    if(argc_ref!=0){
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++){
	  if(isnan(zr[i][j])){
	    zr[i][j]=0.;
	    zn[i][j]=20000.;
	  }
	}
      }
    }
    if(argc_ref==3){
      remove("shift_ave.fits");
      fprintf(stderr,"Write %s ... ","shift_ave.fits");
      status=0;
      if(fits_create_file(&fp0, "shift_ave.fits", &status)) printerror(status);
      if(fits_create_img(fp0, bitpix, naxis, naxes, &status)) printerror(status);
      fprintf(stderr,"%ld x %ld x %ld format \n",naxes[0],naxes[1],naxes[2]);
      fpixel0=1;
    }
    dj=0;di=0;
    fpixel=1;nr=0;
    sig=sqrt(naxes[2]-1);
    if(sig>3.5) sig=3.5;
    else if(sig<2) sig=2;
    sig*=sigfac;
    for(k=0;k<naxes[2];k++){
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
	fpixel+=naxes[0];
      }
      //fprintf(stderr,"%d %d\n",k,om);
      if(k==0 && om==0){
	for(i=0;i<naxes[1];i++){
	  for(j=0;j<naxes[0];j++){
	    if(fabs(zi[i][j])>1e-5) break;
	  }
	  if(j!=naxes[0]) break;
	}
	i1=i;
	for(i=naxes[1]-1;i>=0;i--){
	  for(j=0;j<naxes[0];j++){
	    if(fabs(zi[i][j])>1e-5) break;
	  }
	  if(j!=naxes[0]) break;
	}
	i2=i+1;
	for(j=0;j<naxes[0];j++){
	  for(i=0;i<naxes[1];i++){
	    if(fabs(zi[i][j])>1e-5) break;
	  }
	  if(i!=naxes[1]) break;
	}
	j1=j;
	for(j=naxes[0]-1;j>=0;j--){
	  for(i=0;i<naxes[1];i++){
	    if(fabs(zi[i][j])>1e-5) break;
	  }
	  if(i!=naxes[1]) break;
	}
	j2=j+1;
	//fprintf(stderr,"%d %d %d %d\n",j1,j2,i1,i2);getchar();
	for(i=i1;i<i2;i++){
	  for(j=j1;j<j2;j++){
	    if(isnan(zi[i][j])) continue;
	    if(fabs(zi[i][j])<1e-5) zi[i][j]=0./0.;
	    else break;
	  }
	  for(j=j2-1;j>=j1;j--){
	    if(isnan(zi[i][j])) continue;
	    if(fabs(zi[i][j])<1e-5) zi[i][j]=0./0.;
	    else break;
	  }
	}
	for(j=j1;j<j2;j++){
	  for(i=i1;i<i2;i++){
	    if(isnan(zi[i][j])) continue;
	    if(fabs(zi[i][j])<1e-5) zi[i][j]=0./0.;
	    else break;
	  }
	  for(i=i2-1;i>=i1;i--){
	    if(isnan(zi[i][j])) continue;
	    if(fabs(zi[i][j])<1e-5) zi[i][j]=0./0.;
	    else break;
	  }
	}
      }
      if(k!=0||argc_ref){
	if(argc_ref==0){
	  for(i=0;i<naxes[1];i++){
	    for(j=0;j<naxes[0];j++){
	      zr[i][j]=zo[i][j]/k;
	      if(isnan(zr[i][j])){
		zr[i][j]=0.;
		zn[i][j]=20000.;
	      }
	    }
	  }
	}
	if(om){
	  dz2smin=1e99;md0=9;
	  while(md0!=4){
	    md=4;
	    for(ddi=-1;ddi<=1;ddi++){
	      for(ddj=-1;ddj<=1;ddj++){
		if(cl[md0][3*(ddi+1)+(ddj+1)]){
		  r2s=0.;rs=0.;zrs=0.;z2s=0.;zs=0.;n=0;
		  for(i=i1+BUF;i<i2-BUF;i++){
		    for(j=j1+BUF;j<j2-BUF;j++){
		      if(isnan(zr[i+is+di+ddi][j+js+dj+ddj])==0 && isnan(zi[i][j])==0){
			r2s+=(double)zr[i+is+di+ddi][j+js+dj+ddj]*zr[i+is+di+ddi][j+js+dj+ddj];
			rs+=zr[i+is+di+ddi][j+js+dj+ddj];
			zrs+=(double)zi[i][j]*zr[i+is+di+ddi][j+js+dj+ddj];
			z2s+=(double)zi[i][j]*zi[i][j];
			zs+=zi[i][j];
			n++;
		      }
		    }
		  }
		  r2s/=n;
		  rs/=n;
		  zrs/=n;
		  z2s/=n;
		  zs/=n;
		  a=(zrs-zs*rs)/(r2s-rs*rs);
		  b=(zs*r2s-zrs*rs)/(r2s-rs*rs);
		  dz2s=a*a*r2s+b*b+z2s+2*a*b*rs-2*a*zrs-2*b*zs;
		  if(dz2smin>dz2s){
		    dz2smin=dz2s;
		    amin=a;bmin=b;
		    md=3*(ddi+1)+(ddj+1);
		  }
		}
	      }
	    }
	    md0=md;
	    dj+=md%3-1;
	    di+=md/3-1;
	    //fprintf(stderr,"%d: %d %d %d %g\n",k,md0,dj,di,dz2smin);
	  }
	  fprintf(stderr,"%2d:%3d%3d %g %g %g\n",k,dj,di,dz2smin,amin,bmin);
	  for(i=0;i<naxes[1];i++){
	    for(j=0;j<naxes[0];j++) zb[i][j]=bmin;
	  }
	}
	else{
	  di=0;
	  if(k==0||argc_ref==0){
	    for(j=0;j<naxes[0];j++){
	      x[j]=j-naxes[0]/2;
	      zx[0][j]=0.;
	    }
	    for(j=j1;j<j2;j++){
	      for(i=i1;i<i2;i++) zx[0][j]+=zr[i][j];
	    }
	    for(j=0;j<naxes[0];j++) zt[j]=zx[0][j];
	    zs=fit(0,MAXODR-1,x,zt,j1,j2,co,3);
	    //fprintf(stderr,"zs=%g\n",zs);
	    a=0.;
	    for(j=j1;j<j2;j++){
	      zm=0.;
	      for(m=MAXODR-1;m>=0;m--) zm=zm*x[j]+co[m];
	      zx[2][j]=zm;
	      zx[4][j]=zx[0][j]-zm;
	      a+=zm;
	      //printf("%d %g %g\n",j,zx[0][j],zm);
	    }
	    n=0;
	    for(j=j1+BUF;j<j2-BUF;j++){
	      if(zx[4][j]>SIGMA*zs){
		//fprintf(stderr,"%d: %g %g",j,zx[4][j],3*zs);getchar();
		if(zx[4][j-1]<zx[4][j] && zx[4][j]>zx[4][j+1] && zx[4][j-4]+zx[4][j-3]+zx[4][j-2]<zx[4][j-1]+zx[4][j]+zx[4][j+1] && zx[4][j-1]+zx[4][j]+zx[4][j+1]>zx[4][j+2]+zx[4][j+3]+zx[4][j+4]){
		  xc[0][n]=j+(zx[4][j+1]-zx[4][j-1])/2/(2*zx[4][j]-zx[4][j-1]-zx[4][j+1]);
		  //fprintf(stderr,"%d: %g %g %g %g\n",n,xc[0][n],zx[4][j-1],zx[4][j],zx[4][j+1]);
		  n++;
		}
	      }
	    }
	    nob=n;
	    if(argc_ref==0){
	      for(i=i1;i<i2;i++){
		for(j=0;j<naxes[0];j++) zt[j]=zr[i][j];
		for(n=0;n<nob;n++){
		  j=xc[0][n]+0.5;
		  for(ddj=-HW*5;ddj<=HW*5;ddj++) zt[j+ddj]=0./0.;
		}
		fit(0,MAXODR-1,x,zt,j1,j2,co,3);
		/*
		  if(k==4 && i==900){
		  for(j=j1;j<j2;j++){
		  zm=0.;
		  for(m=MAXODR-1;m>=0;m--) zm=zm*x[j]+co[m];
		  if(isnan(zt[j])) zt[j]=0.;
		  printf("%d %g %g %g\n",j,zr[i][j],zt[j],zm);
		  }
		  }
		*/
		for(j=j1;j<j2;j++){
		  zm=0.;
		  for(m=MAXODR-1;m>=0;m--) zm=zm*x[j]+co[m];
		  zf[i][j]=zr[i][j]-zm;
		}
	      }
	    }
	    for(n=0;n<nob;n++){
	      ddj1=HW;
	      ddj2=HW;
	      if(n>0){
		if(xc[0][n]-xc[0][n-1]<HW*2) ddj1=(xc[0][n]-xc[0][n-1])/2;
	      }
	      if(n<nob-1){
		if(xc[0][n+1]-xc[0][n]<HW*2) ddj2=(xc[0][n+1]-xc[0][n])/2+1.;
	      }
	      w[n]=0.;
	      for(i=i1;i<i2;i++){
		j=xc[0][n]+0.5;
		zy[n][i]=0.;
		for(ddj=-ddj1;ddj<=ddj2;ddj++){
		  if(isnan(zf[i][j+ddj])==0) zy[n][i]+=zf[i][j+ddj];
		}
		w[n]+=zy[n][i];
	      }
	      //fprintf(stderr,"w=%g\n",w[n]);
	      for(i=i1;i<i2;i++) zy[n][i]/=w[n];
	    }
	  }
	  for(j=0;j<naxes[0];j++) zx[1][j]=0.;
	  for(j=j1;j<j2;j++){
	    for(i=i1;i<i2;i++){
	      if(isnan(zi[i][j])==0){
		zx[1][j]+=zi[i][j];
		zb[i][j]=0.;
	      }
	    }
	  }
	  for(j=0;j<naxes[0];j++) zt[j]=zx[1][j];
	  zs=fit(0,MAXODR-1,x,zt,j1,j2,co,3);
	  //fprintf(stderr,"zs=%g\n",zs);
	  amin=0.;
	  for(j=j1;j<j2;j++){
	    zm=0.;
	    for(m=MAXODR-1;m>=0;m--) zm=zm*x[j]+co[m];
	    zx[3][j]=zm;
	    zx[5][j]=zx[1][j]-zm;
	    amin+=zm;
	  }
	  amin/=a;
	  if(amin<0.5) amin=0.5;
	  else if(amin>2.) amin=2.;
	  for(n=0;n<nob;n++){
	    j=xc[0][n]+0.5;
	    xc[1][n]=xc[0][n];
	    for(ddj=0;ddj<5;ddj++){
	      dj=(ddj+1)/2*(1-ddj%2*2);
	      if(zx[5][j+dj-1]<zx[5][j+dj] && zx[5][j+dj]>zx[5][j+dj+1] && zx[5][j+dj-4]+zx[5][j+dj-3]+zx[5][j+dj-2]<zx[5][j+dj-1]+zx[5][j+dj]+zx[5][j+dj+1] && zx[5][j+dj-1]+zx[5][j+dj]+zx[5][j+dj+1]>zx[5][j+dj+2]+zx[5][j+dj+3]+zx[5][j+dj+4]){
		xc[1][n]=j+(zx[5][j+dj+1]-zx[5][j+dj-1])/2/(2*zx[5][j+dj]-zx[5][j+dj-1]-zx[5][j+dj+1]);
		break;
	      }
	    }
	    //fprintf(stderr,"%d: %g %g %g %g\n",n,xc[0][n],zx[4][j-1],zx[4][j],zx[4][j+1]);
	    //fprintf(stderr,"%d: %g %g %g %g\n",n,xc[1][n],zx[5][j+dj-1],zx[5][j+dj],zx[5][j+dj+1]);
	  }
	  fprintf(stderr,"Number of bright spectra(offset): %d @x=",nob);
	  dxs=0.;zs=0.;
	  for(n=0;n<nob;n++){
	    fprintf(stderr," %7.2lf(%+5.2lf)",xc[0][n]+1,xc[0][n]-xc[1][n]);
	    dxs+=(xc[0][n]-xc[1][n])*w[n];
	    zs+=w[n];
	  }
	  if(nob>0){
	    dj=dxs/zs;
	    //dj=floor(dxs/zs+0.5);
	    //if(dxs>0.) dj=dxs/zs+0.25;
	    //else dj=dxs/zs-0.25;
	  }
	  else dj=0;
	  fprintf(stderr," dx=%d\n",dj);
	  for(n=0;n<nob;n++){
	    j=xc[0][n]+0.5;
	    ddj1=HW*10;
	    ddj2=HW*10;
	    if(n>0){
	      if(xc[0][n]-xc[0][n-1]<HW*20) ddj1=(xc[0][n]-xc[0][n-1])/2;
	    }
	    if(n<nob-1){
	      if(xc[0][n+1]-xc[0][n]<HW*20) ddj2=(xc[0][n+1]-xc[0][n])/2+1.;
	    }
	    //fprintf(stderr,"%d %d %d\n",n,ddj1,ddj2);
	    for(ddj=-ddj1;ddj<=ddj2;ddj++){
	      b=zx[1][j-dj+ddj]-amin*zx[0][j+ddj];
	      //fprintf(stderr,"%d %g %g %g",j+ddj,zx[1][j-dj+ddj],amin*zx[0][j+ddj],b);getchar();
	      for(i=i1;i<i2;i++) zb[i][j+ddj]=b*zy[n][i];
	    }
	  }
	}
      }
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++){
	  zt[j]=0.;
	  if(i1<=i-(is+di)&&i-(is+di)<i2&&j1<=j-(js+dj)&&j-(js+dj)<j2){
	    if(isnan(zi[i-(is+di)][j-(js+dj)])==0){
	      zt[j]=zi[i-(is+di)][j-(js+dj)]-(amin*zr[i][j]+zb[i][j]);
	      if(argc_ref==2 && nr0==-2){
		if(za[i][j]<zi[i-(is+di)][j-(js+dj)]){
		  za[i][j]=zi[i-(is+di)][j-(js+dj)];
		  na[i][j]=zt[j];
		}
	      }
	      else{
		za[i][j]+=zi[i-(is+di)][j-(js+dj)];
		na[i][j]+=1.;
	      }
	      if(fabs(zt[j])<sig*zn[i][j]+1e-5){
		zd2s[i][j]+=(double)zt[j]*zt[j];
		zo[i][j]+=zi[i-(is+di)][j-(js+dj)];
		ns[i][j]++;
	      }
	      else nr++;
	      //if(j==1264 && i==616){fprintf(stderr,"%g %g %g %g %g %g %g %g %d %g %g",zi[i-(is+di)][j-(js+dj)],amin*zr[i][j]+zb[i][j],amin,zr[i][j],zb[i][j],zt[j],zn[i][j],sig,ns[i][j],zd2s[i][j],za[i][j]);getchar();}
	    }
	  }
	}
	if(argc_ref==3){
	  if(fits_write_img(fp0, TFLOAT, fpixel0, naxes[0], zt, &status)) printerror(status);
	  fpixel0+=naxes[0];
	}
      }
    }
    if(argc_ref==3){
      if(fits_close_file(fp0, &status)) printerror(status);
    }
    if(fits_close_file(fp, &status)) printerror(status);
    
    if(argc_ref==3 || nr0>=0){
      if(om){
	rs=0.;r2s=20000.;
	for(k=0;k<3;k++){
	  zs=0.;z2s=0.;n=0;
	  for(i=i1+is+HW*2;i<i2+is-HW*2;i++){
	    for(j=j1+is+HW*2;j<j2+is-HW*2;j++){
	      if(fabs(zr[i][j]-rs)<3*r2s){
		zs+=zr[i][j];
		z2s+=(double)zr[i][j]*zr[i][j];
		n++;
	      }
	    }
	  }
	  rs=zs/n;
	  r2s=sqrt(z2s/n-rs*rs);
	}
	for(i=0;i<naxes[1];i++){
	  for(j=0;j<naxes[0];j++) zf[i][j]=za[i][j]/na[i][j]-zo[i][j]/ns[i][j];
	}
	zm=0.;
	for(ddi=-HW*2;ddi<=HW*2;ddi++){
	  for(ddj=-HW*2;ddj<=HW*2;ddj++){
	    zs=0.;
	    for(i=i1+is+HW*2;i<i2+is-HW*2;i++){
	      for(j=j1+is+HW*2;j<j2+is-HW*2;j++){
		if(fabs(zf[i][j])>1. && isnan(zr[i+ddi][j+ddj])==0){
		  zs+=zr[i+ddi][j+ddj];
		}
	      }
	    }
	    if(zm<zs){
	      zm=zs;
	      di=ddi;
	      dj=ddj;
	    }
	    //fprintf(stderr," %g",zs/1000);
	  }
	  //fprintf(stderr,"\n");
	}
	//fprintf(stderr,"(dj,di)=(%d,%d)\n",dj,di);
	n=0;
	for(i=i1+is+HW*2;i<i2+is-HW*2;i++){
	  for(j=j1+is+HW*2;j<j2+is-HW*2;j++){
	    if(fabs(zf[i][j])>1. && zr[i+di][j+dj]-rs>r2s*100) n++;
	    else zf[i][j]=0.;
	  }
	}
	//fprintf(stderr,"n=%d av=%g sg=%g\n",n,rs,r2s);
	if(n>10){
	  fprintf(stderr,"PSF warp is detected in %d pixels.\n",n);
	  for(i=i1+is+HW*2;i<i2+is-HW*2;i++){
	    for(j=j1+is+HW*2;j<j2+is-HW*2;j++){
	      if(fabs(zf[i][j])>1. && zr[i+di][j+dj]-rs>r2s*100){
		zo[i][j]=za[i][j];
		ns[i][j]=na[i][j]+0.5;
		zd2s[i][j]+=zn[i][j]*zn[i][j]*ns[i][j];
		nr--;
	      }
	    }
	  }
	}
      }
      else{
	for(n=0;n<nob;n++){
	  j=xc[0][n]+0.5;
	  for(ddj=-HW*5;ddj<=HW*5;ddj++){
	    for(i=i1;i<i2;i++){
	      if(fabs(za[i][j+ddj]/na[i][j+ddj]-zo[i][j+ddj]/ns[i][j+ddj])>1.){
		m=0;
		for(ddi=0;ddi<BUF;ddi++){
		  if(fabs(za[i+ddi][j+ddj]/na[i+ddi][j+ddj]-zo[i+ddi][j+ddj]/ns[i+ddi][j+ddj])>1.) m++;
		}
		if(i+BUF>i2) b=i2-i;
		else b=BUF;
		if(m>b*(abs(ddj)*0.01+0.05)){
		  zo[i][j+ddj]=za[i][j+ddj];
		  ns[i][j+ddj]=na[i][j+ddj]+0.5;
		  zd2s[i][j+ddj]+=zn[i][j+ddj]*zn[i][j+ddj]*ns[i][j+ddj];
		  nr--;
		}
	      }
	    }
	  }
	}
      }
    }
    fprintf(stderr,"Rejected pixels: %d\n",nr);
    m=0;n=0;
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++){
	if(i-is<i1 || i2<=i-is || j-js<j1 || j2<=j-js) zo[i][j]=0./0.;
	else{
	  if(argc_ref==2 && nr0==-2 && naxes[2]!=1){
	    if(ns[i][j]){
	      //if(j==506 && i==87) fprintf(stderr,"%g %g %d\n",zo[i][j],za[i][j],ns[i][j]);
	      zo[i][j]-=za[i][j];
	      ns[i][j]--;
	      zd2s[i][j]-=na[i][j];
	    }
	  }
	  if(ns[i][j]>0) zo[i][j]/=ns[i][j];
	  else {
	    //fprintf(stderr,"(%d %d)",j,i);
	    zo[i][j]=0./0.;
	    if(i1+BUF<=i-is&&i-is<i2-BUF&&j1+BUF<=j-js&&j-js<j2-BUF){
	      m++;
	    }
	  }
	  if(i-is!=i1 && i2-1!=i-is && j-js!=j1 && j2-1!=j-js){
	    if(ns[i][j]!=naxes[2]&&ns[i-1][j]==naxes[2]&&ns[i][j-1]==naxes[2]&&ns[i+1][j]==naxes[2]&&ns[i][j+1]==naxes[2]) n++;
	  }
	}
      }
    }
    if(m){
      if(m==1) fprintf(stderr,"### All data has been rejected in a pixel ... ###\n");
      else fprintf(stderr,"### All data has been rejected in %d pixels ... ###\n",m);
      sigfac*=1.1;nr0=-1;nr=0;argc_ref=2;
      fprintf(stderr,"Increase threshold & Retry ... (sigfac=%.1lf)\n",sigfac);
    }
    else if(n>1000){
      fprintf(stderr,"### Number of single corrected pixels is too large ... (%d) ###\n",n);
      sigfac*=1.1;nr0=-1;nr=0;argc_ref=2;
      fprintf(stderr,"Increase threshold & Retry ... (sigfac=%.1lf)\n",sigfac);
    }
    if(argc_ref){
      zs=RNO/gain*RNO/gain;
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++){
	  if(ns[i][j]>1){
	    if(zo[i][j]>0.) zn[i][j]=sqrt(zd2s[i][j]/(ns[i][j]-1)+zo[i][j]+zs);
	    else zn[i][j]=sqrt(zd2s[i][j]/(ns[i][j]-1)+zs);
	    //if(j==729 && i==605) fprintf(stderr," %g %d %g %g\n",zn[i][j],ns[i][j],zd2s[i][j],zo[i][j]);
	  }
	  else zn[i][j]=20000.;
	}
      }
    }
    if(om) naxes[2]=2;
    else {
      naxes[2]=3;
      //fprintf(stderr,"1dfit ... ");
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++) zf[i][j]=zo[i][j];
      }
      for(i=i1;i<i2;i++){
	for(j=0;j<naxes[0];j++) zt[j]=zo[i][j];
	for(n=0;n<nob;n++){
	  j=xc[0][n]+0.5;
	  for(ddj=-HW*5;ddj<=HW*5;ddj++) zt[j+ddj]=0./0.;
	}
	zs=fit(0,MAXODR-1,x,zt,j1,j2,co,3);
	/*
	if(k==0 && i==900){
	  for(j=j1;j<j2;j++){
	    zm=0.;
	    for(m=MAXODR-1;m>=0;m--) zm=zm*x[j]+co[m];
	    if(isnan(zt[j])) zt[j]=0.;
	    printf("%d %g %g %g\n",j,zr[i][j],zt[j],zm);
	  }
	}
	*/
	if(zs>1e-5){
	  if(zs<RNO/gain) zs=RNO/gain;
	  for(j=j1;j<j2;j++){
	    zm=0.;
	    for(m=MAXODR-1;m>=0;m--) zm=zm*x[j]+co[m];
	    zf[i][j]-=zm;
	    //fprintf(stderr,"%d %d %g %g %g",j,i,zf[i][j],zm,zs);
	  }
	  for(j=j1;j<j2;j++){
	    if(nr<=nr0){
	      if(zf[i][j]<-6*zs){
		for(dj=j;dj>=j1;dj--){
		  if(zf[i][dj]<-3*zs){
		    zo[i][dj]-=zf[i][dj];
		    zf[i][dj]=0.;
		    if(zf[i-1][dj]<-3*zs){
		      zo[i-1][dj]-=zf[i-1][dj];
		      zf[i-1][dj]=0.;
		    }
		  }
		  else break;
		}
		for(dj=j+1;dj<j2;dj++){
		  if(zf[i][dj]<-3*zs){
		    zo[i][dj]-=zf[i][dj];
		    zf[i][dj]=0.;
		    if(zf[i-1][dj]<-3*zs){
		      zo[i-1][dj]-=zf[i-1][dj];
		      zf[i-1][dj]=0.;
		    }
		  }
		  else break;
		}
	      }
	      else if(zf[i][j]<-3*zs){
		if(fabs(zf[i-1][j])<1e-5){
		    zo[i][j]-=zf[i][j];
		    zf[i][j]=0.;
		}
	      }
	    }
	  }
	}	  
      }
      //fprintf(stderr,"OK\n");
    }
    remove(argv[2]);
    fprintf(stderr,"Write %s ... ",argv[2]);
    status=0;
    if(fits_create_file(&fp, argv[2], &status)) printerror(status);
    if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
    if(om){
      if(fits_write_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
    }
    else{
      if(fits_write_key(fp, TSTRING, "CUNIT2", cunit2, NULL, &status)) printerror(status);
    }
    if(fits_write_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
    j2-=j1;i2-=i1;
    j1++;i1++;
    j1+=js;crpix1+=js;
    i1+=is;crpix2+=is;
    if(fits_write_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
    if(om){
      if(fits_write_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
    }
    if(fits_write_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
    if(fits_write_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
    fpixel=1;
    for(i=0;i<naxes[1];i++){
      if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zo[i], &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(naxes[2]!=1){
      for(i=0;i<naxes[1];i++){
	if(nr<=nr0){
	  for(j=0;j<naxes[0];j++) za[i][j]=za[i][j]/na[i][j]-zo[i][j];
	  if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], za[i], &status)) printerror(status);
	}
	else{
	  if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zn[i], &status)) printerror(status);
	}
	fpixel+=naxes[0];
      }
      if(om==0){
	for(i=0;i<naxes[1];i++){
	  if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zf[i], &status)) printerror(status);
	  fpixel+=naxes[0];
	}
      }
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
    if(argc==3||argc_ref==3) break;
    if(nr<=nr0) break;
    if(argc_ref==2) nr0=nr;
    argc_ref=2;
  }
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
