#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCUBE 1000
#define SATURATE 15000
#define CA 200

void printerror( int status);

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,k,imin,jmin,i1,i2,j1,j2,i1t,i2t,j1t,j2t,ns,status,nfound,bitpix=-32;
  long fpixel,naxis=3,naxes[3]={LINEPIX, LINEPIY, NCUBE};
  float sig=3.,dk[LINEPIX],av0,sg0;
  unsigned short zi[NCUBE][LINEPIX],nsmin,argc_dark=0;
  unsigned long nsat;
  double zs,z2s,ks,k2s,kzs;
  static float a[LINEPIY][LINEPIX],b[LINEPIY][LINEPIX],sg[LINEPIY][LINEPIX];
  FILE *fpw;
  float crval1,crval2,cd1_1,cd1_2,cd2_1,cd2_2,cd3_3,gain;
  int crpix1,crpix2,detid,lp,lpmax;
  char ctype1[16],ctype2[16];
 
  if(argc!=3&&argc!=4&&argc!=5){
    printf("Usage: %s input output [sigma] [darkimage]\n",argv[0]);
    printf("  or\n");
    printf("     : %s input_list output [sigma] [darkimage]\n",argv[0]);
    exit(1);
  }
  if(argc>=4){
    if(strcmp(argv[3]+strlen(argv[3])-4,"fits")==0){
      argc_dark=3;
      if(argc==5) sscanf(argv[4],"%f",&sig);
    }
    else{
      sscanf(argv[3],"%f",&sig);
      if(argc==5) argc_dark=4;
    }
  }
  fprintf(stderr,"Clipping threshold: %g sigma\n",sig);
  if(argc_dark) fprintf(stderr,"Dark image: %s\n",argv[argc_dark]);

  fprintf(stderr,"Read %s  ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY||naxes[2]>NCUBE){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  lpmax=naxes[2]-1;
  if(lpmax>3) lpmax=3;
  if(fits_read_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
  j1--;i1--;
  j2+=j1;i2+=i1;
  for(i=0;i<naxes[1];i++){
    for(j=0;j<naxes[0];j++){
      a[i][j]=0.;
      b[i][j]=0.;
      sg[i][j]=20000.;
    }
  }
  nsmin=naxes[2];
  nsat=0;
  for(i=0;i<naxes[1];i++){
    fpixel=i*naxes[0]+1;
    for(k=0;k<naxes[2];k++){
      if(fits_read_img(fp, TUSHORT, fpixel, naxes[0], NULL, zi[k], NULL, &status)) printerror(status);
      fpixel+=naxes[0]*naxes[1];
    }
    for(j=0;j<naxes[0];j++){
      for(lp=lpmax;lp>=0;lp--){
	zs=0.;
	z2s=0.;
	ks=0.;
	k2s=0.;
	kzs=0.;
	ns=0;
	for(k=0;k<naxes[2];k++){
	  //if(j+1==539 && i+1==717) fprintf(stderr,"%d/%d: %d %d %g %g\n",lp,lpmax,k,zi[k][j],sg[i][j],a[i][j]*(k-naxes[2]/2)+b[i][j]-zi[k][j]);
	  if(fabs(a[i][j]*(k-naxes[2]/2)+b[i][j]-zi[k][j])<sig*sg[i][j]+1e-5){
	    zs+=zi[k][j];
	    z2s+=(double)zi[k][j]*zi[k][j];
	    ks+=(k-naxes[2]/2);
	    k2s+=(k-naxes[2]/2)*(k-naxes[2]/2);
	    kzs+=(double)(k-naxes[2]/2)*zi[k][j];
	    ns++;
	    if(lp==0&&zi[k][j]>SATURATE&&i1<=i&&i<i2&&j1<=j&&j<j2) nsat++;
	  }
	  //if(j+1==955 && i+1==1279) fprintf(stderr,"%ld %lf %lf %d %lf %d\n",k-naxes[2]/2,a[i][j],b[i][j],zi[k][j],sg[i][j],ns);
	}
	zs/=ns;
	z2s/=ns;
	ks/=ns;
	k2s/=ns;
	kzs/=ns;
	if(lpmax){
	  a[i][j]=(kzs-ks*zs)/(k2s-ks*ks);
	  b[i][j]=(k2s*zs-kzs*ks)/(k2s-ks*ks);
	  sg[i][j]=sqrt(k2s*a[i][j]*a[i][j]+(double)b[i][j]*b[i][j]+z2s-2*kzs*a[i][j]-2*zs*b[i][j]+2*ks*a[i][j]*b[i][j]+1e-6);
	  //if(j+1==539 && i+1==717) fprintf(stderr,"%d %g %g %g %g %g %g %g %g\n",ns,zs,z2s,ks,k2s,kzs,a[i][j],b[i][j],sg[i][j]);
	}
	else{
	  a[i][j]=0.;
	  b[i][j]=zs;
	  sg[i][j]=0.;
	}
	if(isnan(a[i][j])||isnan(b[i][j])||isnan(sg[i][j])) fprintf(stderr,"NaN: (%d,%d) %g %g %g\n",j+1,i+1,a[i][j],b[i][j],sg[i][j]);
	//if(j+1==955 && i+1==1279){
	  //fprintf(stderr,"%lf %lf %lf %lf %lf %lf %lf %lf %lf\n",a[i][j],b[i][j],sg[i][j],k2s,z2s,kzs,zs,ks,a[i][j]*a[i][j]*k2s+(double)b[i][j]*b[i][j]+z2s-2*a[i][j]*kzs-2*b[i][j]*zs+2*a[i][j]*b[i][j]*ks);
	  //fprintf(stderr,"%lf %lf %lf %lf %lf %lf\n",a[i][j]*a[i][j]*k2s,(double)b[i][j]*b[i][j],z2s,-2*a[i][j]*kzs,-2*b[i][j]*zs,2*a[i][j]*b[i][j]*ks);
	//}
	//if(lp==0 && isnan(sg[i][j])) fprintf(stderr,"%d %d %g %g\n",j,i,a[i][j],b[i][j]);
      }
      //if(i==naxes[1]/2 && j==naxes[0]/2) fprintf(stderr,"y=%g*x%+g  %g %g %g %g %g %g\n",a[i][j],b[i][j],zs,z2s,ks,k2s,kzs,sg[i][j]);
      if(nsmin>=ns&&i1<=i&&i<i2&&j1<=j&&j<j2){
	nsmin=ns;
	jmin=j;
	imin=i;
      } 
      //if(i==naxes[1]/2 && j==naxes[0]/2) fprintf(stderr,"\n %d %g %g %g %g %g\n",ns,zs,z2s,a[i][j],b[i][j],sg[i][j]);
    }
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  fprintf(stderr,"Effective Area [%d:%d,%d:%d]\n",j1+1,j2,i1+1,i2);
  if(lpmax) fprintf(stderr,"Maximum rejected pixel: [%d,%d] %ld/%ld rejected\n",jmin+1,imin+1,naxes[2]-nsmin,naxes[2]);
  fprintf(stderr,"Saturated (>%d ADU) pixels: %g pix/frame\n",SATURATE,(double)nsat/naxes[2]);
  //zs=0.;
  //for(i=0;i<naxes[1];i++){
  //for(j=0;j<naxes[0];j++) zs+=b[i][j];
  //}
  //fprintf(stderr,"average=%g\n",zs/naxes[1]/naxes[0]);

  if(argc_dark){
    fprintf(stderr,"Read %s  ... ",argv[argc_dark]);
    status=0;
    if(fits_open_file(&fp, argv[argc_dark], READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    fpixel=1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, dk, NULL, &status)) printerror(status);
      fpixel+=naxes[0];
      for(j=0;j<naxes[0];j++) b[i][j]-=dk[j];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"Subtraction OK\n"); 
    i1t=(i1+i2-CA)/2;
    i2t=(i1+i2+CA)/2;
    zs=0.;
    for(j=j1;j<j2;j++){
      dk[j]=0.;
      for(i=i1t;i<i2t;i++) dk[j]+=b[i][j]/CA;
      zs+=dk[j];
      //fprintf(stderr,"%d %g %g",j,dk[j],zs);getchar();
    }
    av0=zs/(j2-j1);
    //fprintf(stderr,"av0=%g\n",av0);
    zs=0.;kzs=0.;
    for(j=j1;j<j2;j++){
      if(dk[j]>av0){
	zs+=dk[j];
	kzs+=j*dk[j];
      }
    }
    //fprintf(stderr,"cen=%g\n",kzs/zs);
    j1t=kzs/zs-CA/2;
    j2t=kzs/zs+CA/2;
    fprintf(stderr,"Center : %dx%d [%d:%d,%d:%d] pix\n",CA,CA,j1t+1,j2t+1,i1t+1,i2t+1);
  }
  else{
    i1t=i1;i2t=i2;
    j1t=j1;j2t=j2;
  }
  av0=0.;sg0=20000.;
  for(lp=5;lp>=0;lp--){
    zs=0.;
    z2s=0.;
    ns=0;
    for(i=i1t;i<i2t;i++){
      for(j=j1t;j<j2t;j++){
	if(sg[i][j]<av0+3*sg0+1e-5){
	  zs+=sg[i][j];
	  z2s+=(double)sg[i][j]*sg[i][j];
	  ns++;
	}
      }
    }
    zs/=ns;
    sg0=sqrt(z2s/ns-zs*zs);
    av0=zs;
    //fprintf(stderr,"%d: Sigma  : %g ± %g (Clipped: %d pix)\n",lp,av0,sg0,(i2t-i1t)*(j2t-j1t)-ns);
  }
  fprintf(stderr,"Sigma  : %g ± %g (Clipped: %d pix)\n",av0,sg0,(i2t-i1t)*(j2t-j1t)-ns);
  for(i=0;i<naxes[1];i++){
    for(j=0;j<naxes[0];j++){
      if(i1t<=i&&i<i2t&&j1t<=j&&j<j2t){
	if(isnan(sg[i][j])) sg[i][j]=1000;
	else sg[i][j]=(sg[i][j]-av0)/(sg0+1e-5);
      }
      else sg[i][j]=0.;
    }
  }
  zs=0.;
  z2s=0.;
  ns=0;
  for(i=i1t;i<i2t;i++){
    for(j=j1t;j<j2t;j++){
      if(sg[i][j]<3.){
	zs+=b[i][j];
	z2s+=(double)b[i][j]*b[i][j];
	ns++;
      }
    }
  }
  zs/=ns;
  sg0=sqrt(z2s/ns-zs*zs);
  av0=zs;
  fprintf(stderr,"Average: %g ± %g (Clipped: %d pix)\n",av0,sg0,(i2t-i1t)*(j2t-j1t)-ns);
  if(argc_dark){
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++) b[i][j]/=av0;
    }
    fprintf(stderr,"Normalization OK\n");
    fprintf(stderr,"Open: sigclip.log ... ");
    fpw=fopen("sigclip.log","a");
    fprintf(fpw,"%s %s %g %g %g %d\n",argv[1],argv[2],(double)nsat/naxes[2],av0,sg0,(i2-i1)*(j2-j1)-ns);
    fclose(fpw);
    fprintf(stderr,"OK\n");
  }

  naxis=2;
  remove(argv[2]);
  fprintf(stderr,"Write %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fp, argv[2], &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  j2-=j1;i2-=i1;
  j1++;i1++;
  if(fits_write_key(fp, TINT, "DETID", &detid, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL2", &crval2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX2", &crpix2, NULL, &status)) printerror(status);
  if(strcmp(ctype2,"WAVELENGTH")==0) strcpy(ctype1,"pixel   ");
  if(fits_write_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD1_2", &cd1_2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_1", &cd2_1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_2", &cd2_2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
  fpixel=1;
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], b[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  remove("sigclip.fits");
  fprintf(stderr,"Write %s ... ","sigclip.fits");
  status=0;
  if(fits_create_file(&fp, "sigclip.fits", &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  fpixel=1;
  fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], sg[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
