#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCUBE 3
#define BUF 50
#define SFT 5
#define RNO 7.1
#define GF 3.
#define MAXODR 25
#define CA 200

void printerror( int status);

double fit(int od1, int od2, float x[], float y[], int n1, int n2, double c[MAXODR], double sig)
{
  double sx[MAXODR*2],xn[MAXODR*2],dy=1.,yf[LINEPIX];
  double p[MAXODR][MAXODR],q[MAXODR][MAXODR],xs[MAXODR],xy[MAXODR],ys[MAXODR],f;
  int    i,j,k,n,ns,ns0,lp,rj[LINEPIX];

  od2++;
  for(n=n1;n<n2;n++) yf[n]=y[n];
  if(sig<0.001) lp=0; else lp=10;
  for(;lp>=0;lp--){
    xn[0]=1;
    sx[0]=1;
    for(n=n1;n<n2;n++) rj[n]=0;
    for(n=n1;n<n2;n++){
      if(fabs(y[n]-yf[n])>sig*(1+lp*0.1)*dy){
	for(i=-1;i<=1;i++){
	  if(0<=n+i && n+i<LINEPIX) rj[n+i]=1;
	}
      }
    }
    ns=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]){
	if(lp==0){
	  //fprintf(stderr,"%g %g %g %g\n",x[n],y[n],y[n]-yf[n],sig*dy);
	  //y[n]=0.;
	}
	continue;
      }
      ns++;
    }
    if(lp==10) ns0=ns;
    if(ns<ns0/2){
      //for(n=n1;n<n2;n++){
      //if(isnan(y[n])==0) fprintf(stderr,"%d %g %g\n",n,x[n],y[n]);
      //}
      //fprintf(stderr,"lp=%d %d/%d %g\n",lp,ns,n2-n1,dy);
      //exit(0);
      fprintf(stderr,"x");
    }
    if(ns<od2*2) {
      //fprintf(stderr,"ns(%d)<od2*2(%d)\n",ns,od2*2);
      for(k=0;k<MAXODR;k++) c[k]=0./0.;
      return(0.);
    }
    for(i=1;i<od2*2;i++) sx[i]=0;
    for(i=0;i<od2;i++) xy[i]=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      for(i=1;i<od2*2;i++) {
        xn[i]=xn[i-1]*x[n];
        sx[i]+=xn[i]/ns;}
      for(i=0;i<od2;i++) xy[i]+=xn[i]*y[n]/ns;
    }
    for(j=0;j<od2;j++){
      for(i=0;i<od2;i++){
        p[i][j]=sx[i+j];
        if(i==j) q[i][j]=1;
        else q[i][j]=0;
      }
    }
    for(k=od1;k<od2;k++){
      for(j=k+1;j<od2;j++){
        f=p[k][j]/p[k][k];
        for(i=0;i<od2;i++){
          p[i][j]-=f*p[i][k];
          q[i][j]-=f*q[i][k];
        }
      }
      f=p[k][k];
      for(i=od1;i<od2;i++){
        p[i][k]/=f;
        q[i][k]/=f;
      }
    }
    k=od2-1;
    for(j=od1;j<k+1;j++){
      ys[j]=0;
      for(i=od1;i<k+1;i++) ys[j]+=q[i][j]*xy[i];
    }
    for(j=k;j>=od1;j--){
      xs[j]=0;
      for(i=j+1;i<k+1;i++) xs[j]+=p[i][j]*c[i];
      c[j]=ys[j]-xs[j];
    }
    dy=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      yf[n]=c[0];
      xn[0]=1;
      for(i=1;i<=k;i++){
	xn[i]=xn[i-1]*x[n];
	yf[n]+=c[i]*xn[i];
      }
      dy+=(y[n]-yf[n])*(y[n]-yf[n])/ns;
    }
    dy=sqrt(dy);
  }
  return(dy);
}

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,i1,i2,j1,j2,i1t,i2t,j1t,j2t,k,status,nfound,bitpix=-32;
  long fpixel,naxis=3,naxes[3]={LINEPIX, LINEPIY, 1};
  static float zi[LINEPIY][LINEPIX],zo[LINEPIY][LINEPIX],zf[LINEPIY][LINEPIX];
  float x[LINEPIX],zx[LINEPIX],zxo[LINEPIX],zy[LINEPIY],zy0[LINEPIY],zyn[LINEPIY],zynb[LINEPIY],vf[2][LINEPIX];
  float wcen,dw,crval1,cd1_1,cd3_3=1.,gain,zt,bk,bk0,wt,rn2,wt0;
  int jmax,jmax0,jw1,jw2,di,dj,crpix1,ycen,ddi,dimin,nav,jj,odr,m,di0,dj0,nvf;
  double zs,dz2s,dz2smin,zmax,sg,sg0,sgs,sgws,z2x[LINEPIX],no,no0,no2s,bks,co[MAXODR],v2s,vzs;
  FILE *fpr;
  char buf[256],fname[64],ctype1[16],ctype2[16],cunit2[16];

  if(argc!=3&&argc!=4&&argc!=5){
    printf("Usage: %s input_list output [order]\n",argv[0]);
    printf(" default order: 6, order<0: no weight mode.\n");
    exit(0);
  }
  if(argc==4){
    sscanf(argv[3],"%d",&odr);
    if(odr>MAXODR-1){
      fprintf(stderr,"oder must be less than %d\n",MAXODR);
      exit(0);
    }
  }
  else odr=6;

  fprintf(stderr,"Read _vflat.fits ... ");
  status=0;
  if(fits_open_file(&fp, "_vflat.fits", READONLY, &status)==0){
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if((naxes[0]>LINEPIX)||(naxes[1]>LINEPIY)){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    fpixel = naxes[0]*naxes[1]/2+1;
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, vf[0], NULL, &status)) printerror(status);
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
    for(j=0;j<naxes[0];j++){
      if(isnan(vf[0][j])==0) break;
    }
    j1=j;
    for(j=naxes[0]-1;j>=0;j--){
      if(isnan(vf[0][j])==0) break;
    }
    j2=j+1;
    zs=0;
    for(j=(j1+j2-CA)/2;j<(j1+j2+CA)/2;j++) zs+=vf[0][j];
    zs/=CA;
    for(j=0;j<naxes[0];j++){
      vf[0][j]/=zs;
      x[j]=j-naxes[0]/2;
    }
    fit(0,6,x,vf[0],j1,j2,co,3);
    for(j=j1;j<j2;j++){
      zs=0.;
      for(m=6;m>=0;m--) zs=zs*x[j]+co[m];
      vf[1][j]=vf[0][j]-zs;
      //printf("%d %g %g %g\n",j,vf[0][j],zs,vf[1][j]);
    }
    nvf=1;
  }
  else{
    fprintf(stderr,"\n### Cannot find_vflat.fits... No slit correction will be applied. ###\n");
    nvf=0;
  }
  for(i=0;i<LINEPIY;i++){
    for(j=0;j<LINEPIX;j++){
      zo[i][j]=0./0.;
      zf[i][j]=0./0.;
    }
  }
  fprintf(stderr,"Read: %s ...\n",argv[1]);
  if((fpr = fopen(argv[1],"r")) == NULL){
    fprintf(stderr,"Cannot open %s\n",argv[1]);
    exit(0);
  }
  k=-1;
  sgs=0.;sgws=0.;no2s=0.;bks=0.;
  while(!feof(fpr)){
    k++;
    if(fgets(buf,sizeof(buf),fpr)==NULL) continue;
    if(buf[0]=='#') continue;
    di0=9999;dj0=9999;wt0=0./0.;
    sscanf(buf,"%s %d %d %f",fname,&dj0,&di0,&wt0);
    nav=strlen(fname);
    sprintf(buf,"TRCS%s",fname+nav-13);
    status=0;
    if(fits_open_file(&fp, buf, READONLY, &status)){
      fprintf(stderr,"### Connot open %s. Using N_frame=1 instead ... ###\n",buf);
      nav=1;
    }
    else {
      if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
      nav=naxes[2];
      if(fits_close_file(fp, &status)) printerror(status);
    }
    fprintf(stderr,"Read %s (x%2d) ... ",fname,nav);
    status=0;
    if(fits_open_file(&fp, fname, READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
    if((naxes[0]!=LINEPIX)||(naxes[1]!=LINEPIY)||(naxes[2]!=NCUBE)){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    if(fits_read_key(fp, TSTRING, "CUNIT2", cunit2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CRVAL2", &wcen, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "CRPIX2", &ycen, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD2_2", &dw, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
    
    rn2=RNO/gain*RNO/gain;
    j1--;i1--;
    j2+=j1;i2+=i1;
    j1t=j1+BUF;
    j2t=j2-BUF;
    if(wcen>720) i1t=i1+BUF*12;
    else i1t=i1+BUF;
    i2t=i2-BUF;
    if(k==0){
      if(di0==9999) di=0;else di=di0;
      if(dj0==9999) dj=0;else dj=dj0;
      if(isnan(wt0)) wt=1.;else wt=wt0;
      fpixel = 1;
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
	fpixel+=naxes[0];
	if(nvf){
	  for(j=j1;j<j2;j++) zi[i][j]/=vf[0][j]*vf[0][j];
	}
	for(j=j1;j<j2;j++){
	  zi[i][j]/=gain*GF;
	  zi[i][j]+=rn2;
	  zi[i][j]*=nav;
	}
      }
      for(i=i1;i<i2;i++){
	for(j=j1;j<j2;j++){
	  if(isnan(zi[i][j])==0){
	    if(isnan(zo[i+di][j+dj])) zo[i+di][j+dj]=zi[i][j]*wt*wt;
	    else zo[i+di][j+dj]+=zi[i][j]*wt*wt;
	  }
	}
      }
      fpixel+=naxes[0]*naxes[1];
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
	fpixel+=naxes[0];
	for(j=j1;j<j2;j++) zi[i][j]*=nav;
      }
    }
    else{
      fpixel = naxes[0]*naxes[1]*2+1;
      for(i=0;i<naxes[1];i++){
	if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
	fpixel+=naxes[0];
	for(j=j1;j<j2;j++) zi[i][j]*=nav;
      }
    }
    if(nvf){
      fprintf(stderr,"vcor ");
      for(i=i1;i<i2;i++){
	//for(j=j1;j<j2;j++) zi[i][j]/=vf[0][j];
      }
      for(jj=1;jj>=0;jj--){
	for(i=i1;i<i2;i++){
	  sg=fit(0,6,x,zi[i],j1,j2,co,3);
	  //fprintf(stderr,"sg=%g\n",sg);
	  for(j=j1;j<j2;j++){
	    zs=0.;
	    for(m=6;m>=0;m--) zs=zs*x[j]+co[m];
	    zi[i][j]-=zs;
	  }
	  v2s=0.;vzs=0.;
	  for(j=j1+BUF;j<j2-BUF;j++){
	    if(isnan(zi[i][j])==0 && isnan(vf[1][j])==0){
	      if(fabs(zi[i][j])<sg*3){
		v2s+=(double)vf[1][j]*vf[1][j];
		vzs+=(double)vf[1][j]*zi[i][j];
	      }
	    }
	  }
	  for(j=j1;j<j2;j++) zi[i][j]-=vzs/v2s*vf[1][j];
	}
      }
      fprintf(stderr,"OK ");
    }
    if(k==0){
      for(i=i1;i<i2;i++){
	for(j=j1;j<j2;j++){
	  if(isnan(zi[i][j])==0){
	    if(isnan(zf[i+di][j+dj])) zf[i+di][j+dj]=zi[i][j]*wt;
	    else zf[i+di][j+dj]+=zi[i][j]*wt;
	  }
	}
      }
      for(i=0;i<naxes[1];i++){
	for(j=0;j<naxes[0];j++) zi[i][j]=zf[i][j];
      }
    }
    for(j=j1t;j<j2t;j++){
      zx[j]=0.;z2x[j]=0.;
      for(i=i1t;i<i2t;i++){
	if(isnan(zi[i][j])==0){
	  zx[j]+=zi[i][j];
	  z2x[j]+=(double)zi[i][j]*zi[i][j];
	}
      }
    }
    zmax=0.;
    for(j=j1t;j<j2t;j++){
      zt=(zx[j-4]+zx[j-3]+zx[j-2])/3+100;
      if(zt<zx[j-1] && zt<zx[j] && zt<zx[j+1]){
	zt=(zx[j+4]+zx[j+3]+zx[j+2])/3+100;
	if(zt<zx[j-1] && zt<zx[j] && zt<zx[j+1]){
	  if(zmax<zx[j]){
	    zmax=zx[j];
	    jmax=j;
	  }
	}
      }
      z2x[j]/=i2t-i1t;
    }
    fprintf(stderr,"zmax=%g @x=%d",zmax,jmax+1);
    sg=0.;
    for(j=j1t;j<j2t;j++) zx[j]/=zmax;
    for(j=j1t+1;j<j2t;j++){
      for(jj=j;jj>j1t;jj--){
	if(z2x[jj]>z2x[jj-1]){
	  zt=z2x[jj-1];
	  z2x[jj-1]=z2x[jj];
	  z2x[jj]=zt;
	}
	else break;
      }
    }
    //for(j=j1t+1;j<j2t;j++){fprintf(stderr,"%d %g",j,z2x[j]);getchar();}
    no=sqrt(z2x[(j1t+j2t)/2]);
    for(j=jmax;j>=0;j--){
      if(zx[j]<0.1) break;
      else sg+=zx[j];
    }
    jw1=j;
    for(j=jmax+1;j<naxes[0];j++){
      if(zx[j]<0.1) break;
      else sg+=zx[j];
    }
    jw2=j;
    sg*=zmax/(i2t-i1t);
    zs=0.;
    for(i=0;i<naxes[1];i++) zy[i]=0.;
    for(i=i1t;i<i2t;i++){
      for(j=jw1;j<=jw2;j++){
	if(isnan(zi[i][j])==0) zy[i]+=zi[i][j]*zx[j];
      }
      zs+=zy[i];
    }
    zs/=i2t-i1t;
    for(i=i1t;i<i2t;i++) zy[i]/=zs;
    if(k==0){
      jmax0=jmax;
      sg0=sg;
      no0=no;
      for(i=0;i<naxes[1];i++){
	zy0[i]=zy[i];
      }
      for(j=j1t;j<j2t;j++){
	zxo[j]=0.;
	for(i=i1t;i<i2t;i++){
	  if(isnan(zo[i][j])==0) zxo[j]+=zo[i][j];
	}
	zxo[j]/=i2t-i1t;
      }
      for(j=j1t+1;j<j2t;j++){
	for(jj=j;jj>j1t;jj--){
	  if(zxo[jj]>zxo[jj-1]){
	    zt=zxo[jj-1];
	    zxo[jj-1]=zxo[jj];
	    zxo[jj]=zt;
	  }
	  else break;
	}
      }
      bk0=zxo[(j1t+j2t)/2];
      sgs=sg0;
      sgws=sg0;
      no2s=no0*no0;
      bks=bk0;
      fprintf(stderr," %+d %+d offset, sig=%g noi=%g wei=%g ",dj,di,sg0,no0,wt);
      //fprintf(stderr,"bak=%g noi2/bak=%g ",bk0,no0*no0/bk0);
      if(fits_close_file(fp, &status)) printerror(status);
      fprintf(stderr,"OK\n");
      continue;
    }
    if(dj0==9999) dj=jmax0-jmax;
    else dj=dj0;
    dz2smin=1e99;
    for(di=-SFT;di<=SFT;di++){
      for(i=i1t;i<i2t;i++) zyn[i]=zy[i]-zy0[i+di];
      for(i=i1t;i<i2t;i++){
	zynb[i]=0.;
	for(ddi=-BUF;ddi<=BUF;ddi++) zynb[i]+=zyn[i+ddi];
	zynb[i]/=2*BUF+1;
      }
      dz2s=0.;
      for(i=i1t;i<i2t;i++){
	dz2s+=(zyn[i]-zynb[i])*(zyn[i]-zynb[i]);
      }
      if(dz2smin>dz2s){
	dz2smin=dz2s;
	dimin=di;
      }
    }
    if(di0==9999) di=dimin;
    else di=di0;
    if(odr<0) wt=1.;
    else{
      if(isnan(wt0)) wt=sg/sg0*(no0/no)*(no0/no);
      else wt=wt0;
    }
    for(i=i1;i<i2;i++){
      for(j=j1;j<j2;j++){
	if(isnan(zi[i][j])==0){
	  if(isnan(zf[i+di][j+dj])) zf[i+di][j+dj]=zi[i][j]*wt;
	  else zf[i+di][j+dj]+=zi[i][j]*wt;
	}
      }
    }
    fpixel = 1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zi[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
      for(j=0;j<naxes[0];j++){
	//if(j==1000 && i==700) fprintf(stderr,"\n%g %g %g %d %g =>",gain,GF,rn2,nav,zi[i][j]);
	zi[i][j]/=gain*GF;
	zi[i][j]+=rn2;
	zi[i][j]*=nav;
	//if(j==1000 && i==700) fprintf(stderr,"%g\n",zi[i][j]);
      }
    }
    for(j=j1t;j<j2t;j++){
      zxo[j]=0.;
      for(i=i1t;i<i2t;i++){
	if(isnan(zi[i][j])==0) zxo[j]+=zi[i][j];
      }
      zxo[j]/=i2t-i1t;
    }
    for(j=j1t+1;j<j2t;j++){
      for(jj=j;jj>j1t;jj--){
	if(zxo[jj]>zxo[jj-1]){
	  zt=zxo[jj-1];
	  zxo[jj-1]=zxo[jj];
	  zxo[jj]=zt;
	}
	else break;
      }
    }
    bk=zxo[(j1t+j2t)/2];
    sgs+=sg;
    sgws+=sg*wt;
    no2s+=no*wt*no*wt;
    bks+=bk*wt*wt;
    for(i=i1;i<i2;i++){
      for(j=j1;j<j2;j++){
	if(isnan(zi[i][j])==0){
	  if(isnan(zo[i+di][j+dj])) zo[i+di][j+dj]=zi[i][j]*wt*wt;
	  else zo[i+di][j+dj]+=zi[i][j]*wt*wt;
	}
      }
    }
    fprintf(stderr," %+d %+d offset, sig=%g noi=%g wei=%g ",dj,di,sg,no,wt);
    //fprintf(stderr,"bak=%g noi2/bak=%g ",bk,no*no/bk);
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
  }
  fclose(fpr);

  for(i=i1;i<i2;i++){
    for(j=j1;j<j2;j++){
      zf[i][j]*=sgs/sgws;
      zo[i][j]*=sgs/sgws*sgs/sgws*no2s/bks;
    }
  }
  fprintf(stderr,"Fit & subtract background (%dth order) ... ",odr);
  for(i=i1;i<i2;i++){
    for(j=j1;j<j2;j++) zxo[j]=zf[i][j];
    for(j=jmax-10;j<=jmax+10;j++) zxo[j]=0./0.;
    fit(0,abs(odr),x,zxo,j1,j2,co,3);
    for(j=j1;j<j2;j++){
      zs=0.;
      for(m=abs(odr);m>=0;m--) zs=zs*x[j]+co[m];
      zf[i][j]-=zs;
    }
  }
  fprintf(stderr,"OK\n");
  j2-=j1;i2-=i1;
  j1++;i1++;
  naxes[2]=2;
  remove(argv[2]);
  fprintf(stderr,"Write %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fp, argv[2], &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
  if(fits_write_key(fp, TSTRING, "CUNIT2", cunit2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL2", &wcen, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX2", &ycen, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_2", &dw, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zf[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zo[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");

  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
