#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360
#define NCUBE 3
#define MEDHW 2
#define BUF 50
#define SFT 5
#define MAXODR 2

void printerror( int status);

double fit(int od1, int od2, float x[], float y[], int n1, int n2, double c[MAXODR], double sig)
{
  double sx[MAXODR*2],xn[MAXODR*2],dy=1.,yf[LINEPIX];
  double p[MAXODR][MAXODR],q[MAXODR][MAXODR],xs[MAXODR],xy[MAXODR],ys[MAXODR],f;
  int    i,j,k,n,ns,lp,rj[LINEPIX];

  od2++;
  for(n=n1;n<n2;n++) yf[n]=y[n];
  if(sig<0.001) lp=0; else lp=10;
  for(;lp>=0;lp--){
    xn[0]=1;
    sx[0]=1;
    ns=0;
    for(n=n1;n<n2;n++) rj[n]=0;
    for(n=n1;n<n2;n++){
      if(fabs(y[n]-yf[n])>sig*(1+lp*0.1)*dy){
        for(i=-1;i<=1;i++){
          if(0<=n+i && n+i<LINEPIX) rj[n+i]=1;
        }
      }
    }
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]){
        if(lp==0){
          //fprintf(stderr,"%g %g %g %g\n",x[n],y[n],y[n]-yf[n],sig*dy);
          //y[n]=0.;
        }
        continue;
      }
      ns++;
    }
    if(ns<od2*2) {
      //fprintf(stderr,"ns(%d)<od2*2(%d)\n",ns,od2*2);
      for(k=0;k<MAXODR;k++) c[k]=0./0.;
      return(0.);
    }
    for(i=1;i<od2*2;i++) sx[i]=0;
    for(i=0;i<od2;i++) xy[i]=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      //fprintf(stderr,"%d %g %g",n,x[n],y[n]);getchar();
      for(i=1;i<od2*2;i++) {
        xn[i]=xn[i-1]*x[n];
        sx[i]+=xn[i]/ns;}
      for(i=0;i<od2;i++) xy[i]+=xn[i]*y[n]/ns;
    }
    for(j=0;j<od2;j++){
      for(i=0;i<od2;i++){
        p[i][j]=sx[i+j];
        if(i==j) q[i][j]=1;
        else q[i][j]=0;
      }
    }
    for(k=od1;k<od2;k++){
      for(j=k+1;j<od2;j++){
        f=p[k][j]/p[k][k];
        for(i=0;i<od2;i++){
          p[i][j]-=f*p[i][k];
          q[i][j]-=f*q[i][k];
        }
      }
      f=p[k][k];
      for(i=od1;i<od2;i++){
        p[i][k]/=f;
        q[i][k]/=f;
      }
    }
    k=od2-1;
    for(j=od1;j<k+1;j++){
      ys[j]=0;
      for(i=od1;i<k+1;i++) ys[j]+=q[i][j]*xy[i];
    }
    for(j=k;j>=od1;j--){
      xs[j]=0;
      for(i=j+1;i<k+1;i++) xs[j]+=p[i][j]*c[i];
      c[j]=ys[j]-xs[j];
    }
    dy=0;
    for(n=n1;n<n2;n++){
      if(isnan(x[n]) || isnan(y[n])) continue;
      if(rj[n]) continue;
      yf[n]=c[0];
      xn[0]=1;
      for(i=1;i<=k;i++){
        xn[i]=xn[i-1]*x[n];
        yf[n]+=c[i]*xn[i];
      }
      dy+=(y[n]-yf[n])*(y[n]-yf[n])/ns;
    }
    dy=sqrt(dy);
  }
  return(dy);
}

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,i1,i2,j1,j2,i1t,i2t,j1t,j2t,status,nfound,bitpix=-32;
  long fpixel,naxis=3,naxes[3]={LINEPIX, LINEPIY,NCUBE};
  static float z[LINEPIY][LINEPIX],zn[LINEPIY][LINEPIX];
  FILE *fpr,*fpw;
  char buf[1024],ctype1[16],ctype2[16],cunit2[16];
  float wl[LINEPIY],rf[LINEPIY],zx[2][LINEPIX],zy[2][LINEPIY],zyn[LINEPIY],zb[2*MEDHW*(MEDHW+1)+1],y[LINEPIY];
  float wcen,dw,crval1,cd1_1,cd3_3=1.,gain;
  int jmax,jw1,jw2,di,dj,im,jm,ddi,dimin=0,k,n,diw,i1s,i2s,nbuf,crpix1,ycen,jc;
  double zmax,wlt0,spt0,wlt,spt,wf,ws,wlt1,wlt2,dwl,dwl1,dwl2,dsp,zynb[LINEPIY],dz2s,dz2smin,dx,co[2];

  if(argc!=5 && argc!=6 && argc!=7){
    printf("Usage: %s input output reference flux_data [column] [psf_file]\n",argv[0]);
    exit(0);
  }
  if(argc>=6){sscanf(argv[5],"%d",&jc);jc--;}
  else jc=-1;
  fprintf(stderr,"Read %s  ... ",argv[3]);
  status=0;
  if(fits_open_file(&fp, argv[3], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY||naxes[2]>NCUBE){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  if(fits_read_key(fp, TSTRING, "CUNIT2", cunit2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CRVAL2", &wcen, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TINT, "CRPIX2", &ycen, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD2_2", &dw, NULL, &status)) printerror(status);
  if(fits_read_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
  ycen--;
  j1--;i1--;
  j2+=j1;i2+=i1;
  j1t=j1+BUF;
  j2t=j2-BUF;
  if(wcen>720) i1t=i1+BUF*12;
  else i1t=i1+BUF;
  i2t=i2-BUF;
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, z[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");
  zmax=0.;
  for(j=j1;j<j2;j++){zx[0][j]=0.;zx[1][j]=0.;}
  for(j=j1t;j<j2t;j++){
    for(i=i1t;i<i2t;i++){
      if(isnan(z[i][j])) z[i][j]=0.;
      zx[0][j]+=z[i][j];
    }
    //fprintf(stderr,"%d %g\n",j,zx[0][j]);getchar();
    if(zmax<zx[0][j]){
      zmax=zx[0][j];
      jmax=j;
    }
  }
  for(i=0;i<naxes[1];i++){zy[0][i]=0.;zy[1][i]=0.;y[i]=i-naxes[1]/2;}
  for(i=i1t;i<i2t;i++){
    zmax=0.;
    for(j=jmax-SFT;j<=jmax+SFT;j++){
      if(zmax<z[i][j]){
	zmax=z[i][j];
	jm=j;
      }
    }
    zy[1][i]=jm+(z[i][jm+1]-z[i][jm-1])/2/(2*z[i][jm]-z[i][jm-1]-z[i][jm+1])-jmax;
  }
  dx=fit(0,MAXODR-1,y,zy[1],i1t,i2t,co,3);
  fprintf(stderr,"peak: x=%+g(y-%ld)%+g ± %g\n",co[1],naxes[1]/2+1,co[0]+jmax+1,dx);
  if(dx<1.){
    for(i=i1;i<i2;i++) zy[1][i]=co[1]*y[i]+co[0];
  }
  else fprintf(stderr,"### No tilt correction will be applied. ###\n");
  for(j=j1t;j<j2t;j++) zx[0][j]=0.;
  for(i=i1t;i<i2t;i++){
    dj=floor(zy[1][i]);
    dx=zy[1][i]-dj;
    for(j=j1t;j<j2t;j++){
      zx[0][j]+=z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx;
      //if(j==jmax){fprintf(stderr,"%d %d %d %g %g %g %g %g",i,j,dj,dx,z[i][j+dj],(z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx),z[i][j+dj+1],zx[1][j]);getchar();}
    }
  }
  //fprintf(stderr,"%g %g %g %g %g\n",zx[0][jmax-2],zx[0][jmax-1],zx[0][jmax],zx[0][jmax+1],zx[0][jmax+2]);
  zmax=zx[0][jmax];
  for(j=j1t;j<j2t;j++) zx[0][j]/=zmax;
  for(j=jmax;j>=0;j--){
    if(zx[0][j]<0.1) break;
  }
  jw1=j-(jmax-j);
  for(;j>=j1t;j--) zx[0][j]=0.;
  for(j=jmax;j<naxes[0];j++){
    if(zx[0][j]<0.1) break;
  }
  jw2=j+(j-jmax);
  for(;j<j2t;j++) zx[0][j]=0.;
  for(j=jw1;j<jmax;j++) zx[1][j]=(zx[0][jmax-(jmax-j)/2]+zx[0][jmax-(jmax-j+1)/2])/2;
  for(j=jmax;j<=jw2;j++) zx[1][j]=(zx[0][jmax+(j-jmax)/2]+zx[0][jmax+(j-jmax+1)/2])/2;
  fprintf(stderr,"Write trcalib_psf0.dat ... ");
  fpw = fopen("trcalib_psf0.dat","w");
  for(j=jw1;j<=jw2;j++) fprintf(fpw,"%d %g %g\n",j-jmax,zx[1][j],zx[0][j]);
  fclose(fpw);
  fprintf(stderr,"OK\n");
  for(i=i1;i<i2;i++){
    dj=floor(zy[1][i]);
    dx=zy[1][i]-dj;
    for(j=jw1;j<=jw2;j++){
      if(isnan(z[i][j+dj])==0 && isnan(z[i][j+dj+1])==0){
	zy[0][i]+=(z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx)*zx[1][j];
	//if(j==jmax){fprintf(stderr,"%d %d %d %g %g %g %g %g",i,j,dj,dx,z[i][j+dj],(z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx),z[i][j+dj+1],zx[1][j]);getchar();}
      }
    }
  }
  //for(j=jw1;j<=jw2;j++) zy[0][i]+=z[i][j]*zx[0][j];
  //for(i=i1;i<i2;i++) fprintf(stderr,"%d %g\n",i,zy[0][i]);getchar();
  fprintf(stderr,"Read %s ...\n",argv[4]);
  if((fpr = fopen(argv[4],"r")) == NULL){
    fprintf(stderr,"Cannot open %s\n",argv[4]);
    exit(0);
  }
  i1s=i1;i2s=i2;
  for(k=1;k>=0;k--){
    i1=i1s;i2=i2s;ycen-=dimin;
    if(k) diw=SFT; else diw=0;
    if(i1<diw) i1=diw;
    if(i2>naxes[1]-diw) i2=naxes[1]-diw;
    for(i=i1-diw;i<i2+diw;i++){
      wl[i]=(i-ycen)*dw+wcen;
      rf[i]=0.;
    }
    fseek(fpr, 0L, SEEK_SET);
    while(!feof(fpr)){
      if(fgets(buf,sizeof(buf),fpr)==NULL) continue;
      if(buf[0]=='#' || buf[0]=='/') continue;
      break;
    }
    sscanf(buf,"%lf %lf",&wlt0,&spt0);
    if(k){
      if(wlt0<0.9){
	fprintf(stderr,"Unit of wavelength: um\n");
	wf=100.;
      }
      else if(wlt0>900.){
	fprintf(stderr,"Unit of wavelength: A\n");
	wf=0.1;
      }
      else if(wlt0>90.){
	fprintf(stderr,"Unit of wavelength: nm\n");
	wf=1.;
      }
      else{
	fprintf(stderr,"Cannot deduce the unit of wavelength in %s\n",argv[3]);
	exit(0);
      }
    }
    wlt0*=wf;
    ws=0.;
    if(dw<0.){
      di=-1;
      i=i2-1+diw;
    }
    else{
      di=1;
      i=i1-diw;
    }
    wlt1=wl[i]-dw*di/2;
    wlt2=wl[i]+dw*di/2;
    im=-1;
    while(!feof(fpr)){
      if(fgets(buf,sizeof(buf),fpr)==NULL) continue;
      if(buf[0]=='#'){
	if(k) sscanf(buf+1,"%lf %lf",&wlt,&spt);
	else continue;
      }
      else if(buf[0]=='/'){
	if(k) sscanf(buf+1,"%lf %lf",&wlt,&spt);
	else{
	  if(im==-1) im=i;
	  continue;
	}
      }
      else sscanf(buf,"%lf %lf",&wlt,&spt);
      wlt*=wf;
      //fprintf(stderr,"%d %g %g %g %g %g\n",i,wlt,spt,wlt0,wlt1,wlt2);
      if(wlt>wlt1){
	while(wlt0>=wlt2){
	  if(ws>1e-8) rf[i]/=ws;
	  //fprintf(stderr,"%d %g",i,rf[i]);getchar();
	  i+=di;ws=0.;
	  if(i<i1-diw || i>=i2+diw) break;
	  wlt1=wl[i]-dw*di/2;
	  wlt2=wl[i]+dw*di/2;
	}
	if(wlt0>wlt1 && wlt>wlt2){
	  dwl=wlt2-wlt0;
	  dwl1=dwl/2;
	  dwl2=wlt-(wlt2+wlt0)/2;
	  rf[i]+=(dwl1*spt+dwl2*spt0)/(dwl1+dwl2)*dwl;
	  ws+=dwl;
	  rf[i]/=ws;
	  //fprintf(stderr,"%d %g\n",i,rf[i]);
	  i+=di;ws=0.;
	  if(i<i1-diw || i>=i2+diw) break;
	  wlt1=wl[i]-dw*di/2;
	  wlt2=wl[i]+dw*di/2;
	}
	while(wlt0<=wlt1 && wlt>=wlt2){
	  dwl1=(wlt2+wlt1)/2-wlt0;
	  dwl2=wlt-(wlt2+wlt1)/2;
	  rf[i]=(dwl1*spt+dwl2*spt0)/(dwl1+dwl2);
	  //fprintf(stderr,"%d %g\n",i,rf[i]);
	  i+=di;ws=0.;
	  if(i<i1-diw || i>=i2+diw) break;
	  wlt1=wl[i]+dw*di/2;
	  wlt2=wl[i]-dw*di/2;
	}
	if(i<i1-diw || i>=i2+diw) break;
	if(wlt0<wlt1 && wlt<wlt2){
	  dwl=wlt-wlt1;
	  dwl1=(wlt+wlt1)/2-wlt0;
	  dwl2=dwl/2;
	  rf[i]+=(dwl1*spt+dwl2*spt0)/(dwl1+dwl2)*dwl;
	  ws+=dwl;
	}
	if(wlt0>=wlt1 && wlt<=wlt2){
	  dwl=wlt-wlt0;
	  rf[i]+=(spt+spt0)/2*dwl;
	  ws+=dwl;
	}
      }
      wlt0=wlt;
      spt0=spt;
      if(im>0){
	dsp=(zy[0][i]-zy[0][im])/(i-im)*di;
	while(im!=i){
	  zy[0][im+di]=zy[0][im]+dsp;
	  im+=di;
	}
	im=-1;
      }
    }
    for(i=i1-diw;i<i2+diw;i++){
      if(rf[i]>1e-5) break;
    }
    if(i!=i1-diw){
      if(k==0) fprintf(stderr,"### No reference data between %gnm-%gnm. ###\n",wl[i-1],wl[i1-diw]);
      i1=i+diw;
    }
    for(i=i2+diw-1;i>=i1-diw;i--){
      if(rf[i]>1e-5) break;
    }
    if(i!=i2+diw-1){
      if(k==0) fprintf(stderr,"### No reference data between %gnm-%gnm.\n ###",wl[i+1],wl[i2+diw-1]);
      i2=i-diw+1;
    }
    dimin=0;
    if(k){
      dz2smin=1e99;
      for(di=-diw;di<=diw;di++){
	for(i=i1;i<i2;i++) zyn[i]=zy[0][i]/rf[i+di];
	for(i=i1t;i<i2t;i++){
	  zynb[i]=0.;
	  for(ddi=-BUF;ddi<=BUF;ddi++) zynb[i]+=zyn[i+ddi];
	  zynb[i]/=2*BUF+1;
	}
	dz2s=0.;
	for(i=i1t;i<i2t;i++){
	  dz2s+=(zyn[i]-zynb[i])*(zyn[i]-zynb[i]);
	}
	if(dz2smin>dz2s){
	  dz2smin=dz2s;
	  dimin=di;
	}
      }
      fprintf(stderr,"y offset=%+d\n",dimin);
    }
    if(k==1){
      fprintf(stderr,"Write trcalib.dat ... ");
      fpw = fopen("trcalib.dat","w");
      for(i=i1;i<i2;i++) fprintf(fpw,"%g %g %g %g\n",wl[i+dimin],zy[0][i]/rf[i+dimin],zy[0][i],rf[i+dimin]);
      fclose(fpw);
      fprintf(stderr,"OK\n");
    }
  }
  fclose(fpr);
  fprintf(stderr,"Read %s  ... ",argv[1]);
  status=0;
  if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
  if(fits_read_keys_lng(fp, "NAXIS", 1, 3, naxes, &nfound, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
  if(naxes[0]>LINEPIX||naxes[1]>LINEPIY||naxes[2]>NCUBE){
    fprintf(stderr,"\nIllegal image format\n");
    exit(0);
  }
  fpixel=naxes[0]*naxes[1]+1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, z[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
    for(j=0;j<naxes[0];j++) zn[i][j]=z[i][j];
  }
  nbuf=2*MEDHW*(MEDHW+1)+1;
  for(i=i1+MEDHW;i<i2-MEDHW;i++){
    for(j=j1+MEDHW;j<j2-MEDHW;j++){
      for(n=0;n<nbuf;n++) zb[n]=0.;
      for(di=-MEDHW;di<=MEDHW;di++){
	for(dj=-MEDHW;dj<=MEDHW;dj++){
	  for(n=nbuf-1;n>=0;n--){
	    if(zb[n]<z[i+di][j+dj]) break;
	  }
	  for(k=0;k<n;k++){
	    zb[k]=zb[k+1];
	  }
	  if(n>=0) zb[n]=z[i+di][j+dj];
	}
      }
      zn[i][j]=zb[0];
    }
  }
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, z[i], NULL, &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");

  zmax=0.;
  for(j=j1;j<j2;j++){zx[0][j]=0.;zx[1][j]=0.;}
  for(j=j1t;j<j2t;j++){
    for(i=i1t;i<i2t;i++){
      if(isnan(z[i][j])) z[i][j]=0.;
      zx[0][j]+=z[i][j];
    }
    if(zmax<zx[0][j]){
      zmax=zx[0][j];
      jmax=j;
    }
    //printf("%d %g\n",j,zx[0][j]);
  }
  if(jc>=0 && abs(jmax-jc)>2){
    zmax=0.;
    for(j=jc-2;j<=jc+2;j++){
      if(zmax<zx[0][j]){
	zmax=zx[0][j];
	jmax=j;
      }
    }
  }
  //fprintf(stderr,"zmax=%g @x=%d\n",zmax,jmax+1);
  for(i=0;i<naxes[1];i++) zy[1][i]=0.;
  for(i=i1t;i<i2t;i++){
    zmax=0.;
    for(j=jmax-SFT;j<=jmax+SFT;j++){
      if(zmax<z[i][j]){
        zmax=z[i][j];
        jm=j;
      }
    }
    zy[1][i]=jm+(z[i][jm+1]-z[i][jm-1])/2/(2*z[i][jm]-z[i][jm-1]-z[i][jm+1])-jmax;
  }
  dx=fit(0,MAXODR-1,y,zy[1],i1t,i2t,co,3);
  fprintf(stderr,"peak: x=%+g(y-%ld)%+g ± %g\n",co[1],naxes[1]/2+1,co[0]+jmax+1,dx);
  if(dx<1.){
    for(i=i1;i<i2;i++) zy[1][i]=co[1]*y[i]+co[0];
  }
  else fprintf(stderr,"### No tilt correction will be applied. ###\n");
  for(j=j1t;j<j2t;j++) zx[0][j]=0.;
  for(i=i1t;i<i2t;i++){
    dj=floor(zy[1][i]);
    dx=zy[1][i]-dj;
    for(j=j1t;j<j2t;j++){
      zx[0][j]+=z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx;
      //if(j==jmax){fprintf(stderr,"%d %d %d %g %g %g %g %g",i,j,dj,dx,z[i][j+dj],(z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx),z[i][j+dj+1],zx[1][j]);getchar();}
    }
  }
  //fprintf(stderr,"%g %g %g %g %g\n",zx[0][jmax-2],zx[0][jmax-1],zx[0][jmax],zx[0][jmax+1],zx[0][jmax+2]);
  zmax=zx[0][jmax];
  for(j=j1t;j<j2t;j++) zx[0][j]/=zmax;
  for(j=jmax;j>=j1;j--){
    if(zx[0][j]<0.1) break;
  }
  jw1=j-(jmax-j);
  for(;j>=j1t;j--) zx[0][j]=0.;
  for(j=jmax;j<j2;j++){
    if(zx[0][j]<0.1) break;
  }
  jw2=j+(j-jmax);
  for(;j<j2t;j++) zx[0][j]=0.;
  for(j=jw1;j<jmax;j++) zx[1][j]=(zx[0][jmax-(jmax-j)/2]+zx[0][jmax-(jmax-j+1)/2])/2;
  for(j=jmax;j<=jw2;j++) zx[1][j]=(zx[0][jmax+(j-jmax)/2]+zx[0][jmax+(j-jmax+1)/2])/2;
  if(argc==7){
    fprintf(stderr,"Read %s ... ",argv[6]);
    if((fpr = fopen(argv[6],"r")) == NULL){
      fprintf(stderr,"Cannot open %s\n",argv[6]);
      fprintf(stderr,"### Default aperture willbe applied. ###\n");
    }
    else{
      for(j=jw1;j<=jw2;j++) zx[1][j]=0.;
      while(!feof(fpr)){
	if(fgets(buf,sizeof(buf),fpr)==NULL) continue;
	sscanf(buf,"%d",&j);
	j+=jmax;
	sscanf(buf,"%d %f",&dj,&zx[1][j]);
      }
      fprintf(stderr,"OK\n");
    }
  }
  fprintf(stderr,"Write trcalib_psf.dat ... ");
  fpw = fopen("trcalib_psf.dat","w");
  for(j=jw1;j<=jw2;j++) fprintf(fpw,"%d %g %g\n",j-jmax,zx[1][j],zx[0][j]);
  fclose(fpw);
  fprintf(stderr,"OK\n");
  for(i=i1;i<i2;i++){
    zyn[i]=zy[0][i]/rf[i];
    if(zy[0][i]<1e-5) ws=0.;
    else ws=rf[i]/zy[0][i];
    for(j=j1;j<j2;j++){
      if(isnan(z[i][j])==0) z[i][j]*=ws;
      else z[i][j]=0.;
      if(isnan(zn[i][j])==0) zn[i][j]*=ws*ws;
      else zn[i][j]=0.;
    }
  }
  for(i=i1;i<i2;i++){
    zy[0][i]=0.;
    zynb[i]=0.;
    dj=floor(zy[1][i]);
    dx=zy[1][i]-dj;
    for(j=jw1;j<=jw2;j++){
      if(isnan(z[i][j+dj])==0 && isnan(z[i][j+dj+1])==0){
	zy[0][i]+=(z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx)*zx[1][j];
	zynb[i]+=(zn[i][j+dj]*(1-dx)+zn[i][j+dj+1]*dx)*zx[1][j];
	//if(j==jmax){fprintf(stderr,"%d %d %d %g %g %g %g %g",i,j,dj,dx,z[i][j+dj],(z[i][j+dj]*(1-dx)+z[i][j+dj+1]*dx),z[i][j+dj+1],zx[1][j]);getchar();}
      }
    }
    zynb[i]=sqrt(zynb[i]);
  }
  for(i=i1;i<i2;i++){
    z[i][0]=wl[i];
    z[i][1]=zy[0][i];
    z[i][2]=zynb[i];
    z[i][3]=zyn[i];
    z[i][4]=zyn[i]*rf[i];
    z[i][5]=rf[i];
    printf("%g %g %g %g %g %g\n",wl[i],zy[0][i],zynb[i],zyn[i],zyn[i]*rf[i],rf[i]);
    for(j=0;j<naxes[0];j++){
      z[i-i1][j]=z[i][j];
      zn[i-i1][j]=zn[i][j];
    }
  }
  for(j=0;j<naxes[0];j++) z[i2-i1][j]=0./0.;
  for(j=jw1;j<=jw2;j++) z[i2-i1][j]=zx[0][j];

  naxes[1]=i2-i1+1;
  naxes[2]=2;
  remove(argv[2]);
  fprintf(stderr,"Write %s ... ",argv[2]);
  status=0;
  if(fits_create_file(&fp, argv[2], &status)) printerror(status);
  if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
  fprintf(stderr,"%ld x %ld x %ld format ",naxes[0],naxes[1],naxes[2]);
  if(fits_write_key(fp, TSTRING, "CUNIT2", cunit2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "GAIN", &gain, NULL, &status)) printerror(status);
  ycen-=i1;ycen++;
  j2-=j1;i2=naxes[1];
  j1++;i1=1;
  if(fits_write_key(fp, TINT, "EFP-MIN1", &j1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-MIN2", &i1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG1", &j2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "EFP-RNG2", &i2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL1", &crval1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CRVAL2", &wcen, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX1", &crpix1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TINT, "CRPIX2", &ycen, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE1", ctype1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TSTRING, "CTYPE2", ctype2, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD1_1", &cd1_1, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD2_2", &dw, NULL, &status)) printerror(status);
  if(fits_write_key(fp, TFLOAT, "CD3_3", &cd3_3, NULL, &status)) printerror(status);
  fpixel=1;
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], z[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  for(i=0;i<naxes[1];i++){
    if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zn[i], &status)) printerror(status);
    fpixel+=naxes[0];
  }
  if(fits_close_file(fp, &status)) printerror(status);
  fprintf(stderr,"OK\n");

  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if (status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null if stack is empty */
    if ( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
