#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fitsio.h"

#define LINEPIX 2220
#define LINEPIY 1360

void printerror( int status);

int main(int argc,char *argv[])
{
  fitsfile *fp;
  int i,j,i1,i2,j1,j2,status,nfound,bitpix=-32;
  long fpixel,naxis=2,naxes[2]={LINEPIX, LINEPIY};
  static float zx[LINEPIY][LINEPIX],zy[LINEPIY][LINEPIX],dzx[LINEPIY][LINEPIX],dzy[LINEPIY][LINEPIX];
  float wcen=-1.,dw,tmpval,fac;
  double x,y,wx,wy;
  char fname[32],*dot;
  int ycen;

  if(argc!=5 && argc!=6){
    printf("Usage : %s dx0.fits dy0.fits dx.fits dy.fits [fac]\n",argv[0]);
    printf("  or  : %s dx0.fits dy0.fits 0 dy.fits [fac]\n",argv[0]);
    printf("  or  : %s dx0.fits dy0.fits dx.fits 0 [fac]\n",argv[0]);
    printf("  or  : %s 0 dy0.fits 0 dy.fits [fac]\n",argv[0]);
    printf("  or  : %s dx0.fits 0 dx.fits 0 [fac]\n",argv[0]);
    exit(0);
  }
  if(argc==6) sscanf(argv[5],"%f",&fac);
  else fac=1.;

  if(strcmp(argv[1],"0")==0){
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++) zx[i][j]=0.;
    }
  }
  else {
    fprintf(stderr,"Read %s  ... ",argv[1]);
    status=0;
    if(fits_open_file(&fp, argv[1], READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    fpixel=1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zx[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
  }
  if(strcmp(argv[2],"0")==0){
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++) zy[i][j]=0.;
    }
  }
  else {
    fprintf(stderr,"Read %s  ... ",argv[2]);
    status=0;
    if(fits_open_file(&fp, argv[2], READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    if(fits_read_key(fp, TFLOAT, "CRVAL2", &wcen, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TINT, "CRPIX2", &ycen, NULL, &status)) printerror(status);
    if(fits_read_key(fp, TFLOAT, "CD2_2", &dw, NULL, &status)) printerror(status);
    fpixel=1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, zy[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
  }
  if(strcmp(argv[3],"0")==0){
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++) dzx[i][j]=0.;
    }
  }
  else {
    fprintf(stderr,"Read %s  ... ",argv[3]);
    status=0;
    if(fits_open_file(&fp, argv[3], READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    fpixel=1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, dzx[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
  }
  if(strcmp(argv[4],"0")==0){
    for(i=0;i<naxes[1];i++){
      for(j=0;j<naxes[0];j++) dzy[i][j]=0.;
    }
  }
  else {
    fprintf(stderr,"Read %s  ... ",argv[4]);
    status=0;
    if(fits_open_file(&fp, argv[4], READONLY, &status)) printerror(status);
    if(fits_read_keys_lng(fp, "NAXIS", 1, 2, naxes, &nfound, &status)) printerror(status);
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if(naxes[0]>LINEPIX||naxes[1]>LINEPIY){
      fprintf(stderr,"\nIllegal image format\n");
      exit(0);
    }
    fpixel=1;
    for(i=0;i<naxes[1];i++){
      if(fits_read_img(fp, TFLOAT, fpixel, naxes[0], NULL, dzy[i], NULL, &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
  }
  for(i=0;i<naxes[1];i++){
    for(j=0;j<naxes[0];j++){
      x=j+zx[i][j];y=i+zy[i][j];
      j1=x;i1=y;
      j2=x+1;i2=y+1;
      wx=x-j1;wy=y-i1;
      if(j1<0) j1=0;
      if(j1>naxes[0]-1) j1=naxes[0]-1;
      if(j2<0) j2=0;
      if(j2>naxes[0]-1) j2=naxes[0]-1;
      if(i1<0) i1=0;
      if(i1>naxes[1]-1) i1=naxes[1]-1;
      if(i2<0) i2=0;
      if(i2>naxes[1]-1) i2=naxes[1]-1;
      zx[i][j]+=(dzx[i1][j1]*(1-wx)*(1-wy)+dzx[i1][j2]*wx*(1-wy)+dzx[i2][j1]*(1-wx)*wy+dzx[i2][j2]*wx*wy)*fac;
      zy[i][j]+=(dzy[i1][j1]*(1-wx)*(1-wy)+dzy[i1][j2]*wx*(1-wy)+dzy[i2][j1]*(1-wx)*wy+dzy[i2][j2]*wx*wy)*fac;
    }
  }
  if(strcmp(argv[3],"0")){
    sprintf(fname,"%s",argv[1]);
    dot=strrchr(fname, '.');
    sprintf(dot+1,"bak");
    fprintf(stderr,"Rename %s => %s\n",argv[1],fname);
    rename(argv[1],fname);
    fprintf(stderr,"Write %s ... ",argv[1]);
    status=0;
    if(fits_create_file(&fp, argv[1], &status)) printerror(status);
    if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
    fpixel=1;
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    for(i=0;i<naxes[1];i++){
      if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zx[i], &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
  }
  if(strcmp(argv[4],"0")){
    sprintf(fname,"%s",argv[2]);
    dot=strrchr(fname, '.');
    sprintf(dot+1,"bak");
    fprintf(stderr,"Rename %s => %s\n",argv[2],fname);
    rename(argv[2],fname);
    fprintf(stderr,"Write %s ... ",argv[2]);
    status=0;
    if(fits_create_file(&fp, argv[2], &status)) printerror(status);
    if(fits_create_img(fp, bitpix, naxis, naxes, &status)) printerror(status);
    fpixel=1;
    fprintf(stderr,"%ld x %ld format ",naxes[0],naxes[1]);
    if(wcen>0){
      if(fits_write_key(fp, TSTRING, "CTYPE2", "WAVELENGTH", "", &status)) printerror(status);
      if(fits_write_key(fp, TSTRING, "CUNIT2", "nm", "", &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CRVAL2", &wcen, "", &status)) printerror(status);
      if(fits_write_key(fp, TINT, "CRPIX2", &ycen, "", &status)) printerror(status);
      tmpval=1.;
      if(fits_write_key(fp, TFLOAT, "CD1_1", &tmpval, "", &status)) printerror(status);
      if(fits_write_key(fp, TFLOAT, "CD2_2", &dw, "", &status)) printerror(status);
    }
    for(i=0;i<naxes[1];i++){
      if(fits_write_img(fp, TFLOAT, fpixel, naxes[0], zy[i], &status)) printerror(status);
      fpixel+=naxes[0];
    }
    if(fits_close_file(fp, &status)) printerror(status);
    fprintf(stderr,"OK\n");
  }
  return 0;
}

void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
    if(status)
      fprintf(stderr, "\n*** Error occurred during program execution ***\n");

    fits_get_errstatus(status, status_str);   /* get the error description */
    fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);

    /* get first message; null ifstack is empty */
    if( fits_read_errmsg(errmsg) ) 
    {
         fprintf(stderr, "\nError message stack:\n");
         fprintf(stderr, " %s\n", errmsg);

         while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
             fprintf(stderr, " %s\n", errmsg);
    }

    exit( status );       /* terminate the program, returning error status */
}
